from typing import List


class Callback(object):
    """The class representing a callback"""

    def on_train_begin(self, logs=None):
        """Method called at the start of the training process.

        Args:
            logs: (Optional) The logged data before this step.

        Returns:
            None.
        """
        pass

    def on_train_end(self, logs=None):
        """Method called at the end of the training process."""
        pass

    def on_test_begin(self, logs=None):
        """Method called at the start of the testing process.

        Args:
            logs: (Optional) The logged data before this step.

        Returns:
            None.
        """
        pass

    def on_test_end(self, logs=None):
        """Method called at the end of the testing process."""
        pass

    def on_step_begin(self, step, logs=None):
        """Method called before a time step has started.

        Args:
            step: The current step.
            logs: (Optional) The logged data before this step.

        Returns:
            None.
        """
        pass

    def on_step_end(self, step, logs=None):
        """Method called at the end of a time step."""
        pass

    def on_episode_begin(self, step, episode, logs=None):
        """Method called before the start of an episode.

        Args:
            step: The step at which the episode ended.
            episode: The current episode.
            logs: (Optional) The logged data before this step.

        Returns:
            None.
        """
        pass

    def on_episode_end(self, step, episode, logs=None):
        """Method called at the end of an episode."""
        pass


class CallbackGroup(Callback):
    """The class representing a group of callbacks.

    Attributes:
        callbacks: A list of callbacks in the group.
    """
    def __init__(self, callbacks: List[Callback]):
        # Store the callbacks
        self.callbacks = callbacks

    def on_train_begin(self, logs=None):
        for callback in self.callbacks:
            callback.on_train_begin(logs)

    def on_train_end(self, logs=None):
        for callback in self.callbacks:
            callback.on_train_end(logs)

    def on_test_begin(self, logs=None):
        for callback in self.callbacks:
            callback.on_test_begin(logs)

    def on_test_end(self, logs=None):
        for callback in self.callbacks:
            callback.on_test_end(logs)

    def on_step_begin(self, step, logs=None):
        for callback in self.callbacks:
            callback.on_step_begin(step, logs)

    def on_step_end(self, step, logs=None):
        for callback in self.callbacks:
            callback.on_step_end(step, logs)

    def on_episode_begin(self, step, episode, logs=None):
        for callback in self.callbacks:
            callback.on_episode_begin(step, episode, logs)

    def on_episode_end(self, step, episode, logs=None):
        for callback in self.callbacks:
            callback.on_episode_end(step, episode, logs)
