import os
from collections import deque

import gym
import numpy as np
import csv

from core.callbacks.callback import Callback
from core.models.dqn.dqn_agent import DQNAgent
from core.paths.dqn_path import DQNPaths


class AgentCallback(Callback):
    """The class representing a callback for a DQN agent.

    Attributes:
        agent: The agent the callback is meant for.
    """
    def __init__(self, agent: DQNAgent):
        # Store the agent
        self.agent = agent


class SaveAgentCallback(AgentCallback):
    """A save callback for a DQN agent.

    Attributes:
        agent: The agent the callback is meant for.
        step_interval: The number of steps between each save.
        average_interval: The interval of episodes to calculate the average for.
    """
    def __init__(self, agent: DQNAgent, step_interval, average_interval=10):
        # Super call
        super(SaveAgentCallback, self).__init__(agent)
        # Store the interval
        self.step_interval = step_interval
        # Save the best agent based on performance
        self.average_interval = average_interval
        self.last_rewards = deque(maxlen=average_interval)
        self.previous_best = -np.inf
        # Initialize the log file
        d = DQNPaths.log_directory(self.agent.experiment_config, agent.training, best_save=True)
        os.makedirs(d, exist_ok=True)
        from core.callbacks.loggers.dqn_logger import DQNEpisodeLog
        self.log = DQNEpisodeLog()
        self.log_file = DQNPaths.save_best_log(self.agent.experiment_config, agent.training)
        self.header = ["Step", "Episode", "Interval", "Average Reward"]
        with open(self.log_file, "w", newline="") as file_writer:
            writer = csv.DictWriter(file_writer, fieldnames=self.header)
            writer.writeheader()

    def on_step_end(self, step, logs=None):
        # Save the agent
        if step % self.step_interval == 0:
            gym.logger.info(f"Saving agent at time step {step}.")
            path = DQNPaths.save_path(self.agent.experiment_config, step_save=step)
            self.agent.save(path)

    def on_episode_end(self, step, episode, logs=None):
        # Add the episode reward to the last rewards
        self.last_rewards.append(logs[self.log.episode_reward])
        new_average = sum(self.last_rewards) / len(self.last_rewards)
        # If the average reward is better, update the best model
        if new_average > self.previous_best:
            gym.logger.info(f"Saving new best agent... New average reward: {new_average}")
            self.previous_best = new_average
            path = DQNPaths.save_best_path(self.agent.experiment_config)
            self.agent.save(path)
            with open(self.log_file, "a", newline="") as file_writer:
                writer = csv.DictWriter(file_writer, fieldnames=self.header)
                entry = {"Step": step, "Episode": episode, "Interval": self.average_interval,
                         "Average Reward": new_average}
                writer.writerow(entry)
