import gym
import numpy as np

from core.callbacks.dqn.agent_callback import AgentCallback
from core.callbacks.loggers.dqn_logger import DQNStepLog, DQNEpisodeLog
from core.models.dqn.dqn_agent import DQNAgent


class EpisodeMemoryCallback(AgentCallback):
    """The callback class for the DQN agent's memory per episode.

    Attributes:
        agent: The agent to store the experiences for.
        reset_interval: The number of episodes after which to reduce resource usage
            by removing old memories which will not be used.
        observations: The observations gathered from the agent, grouped per episode.
        q_values: The q-values gathered from the agent, grouped per episode.
        verbose: The verbosity of the logging information.
    """

    def __init__(self, agent: DQNAgent, reset_interval, min_successful_reward, verbose):
        # Super call
        super(EpisodeMemoryCallback, self).__init__(agent)
        # The experiences to store
        self.reset_interval = reset_interval
        self.min_successful_reward = min_successful_reward
        self.verbose = verbose
        self.observations = {}
        self.q_values = {}
        self.actions = {}
        self._ep_observations = []
        self._ep_q_values = []
        self._ep_actions = []
        self._name = "EpisodeMemory"
        self._step_log = DQNStepLog()
        self._episode_log = DQNEpisodeLog()

    def on_train_begin(self, logs=None):
        """Callback when training begins"""
        self.observations = {}
        self.q_values = {}
        self.actions = {}
        self._ep_observations = []
        self._ep_q_values = []
        self._ep_actions = []

    def on_step_end(self, step, logs=None):
        """Callback when a step ends"""
        # Store the state and q-values of the current episodes
        self._ep_observations.append(logs[self._step_log.state])
        self._ep_q_values.append(logs[self._step_log.q_values])
        self._ep_actions.append(logs[self._step_log.action])

    def on_episode_end(self, step, episode, logs=None):
        """Callback when an episode ends"""
        # Check if the episode was successful, else, remove
        keep_episode = True
        episode_reward = logs[self._episode_log.episode_reward]
        # Don't keep bas episodes if specified
        if self.min_successful_reward is not None and episode_reward < self.min_successful_reward:
            keep_episode = False

        if keep_episode:
            if self.min_successful_reward is not None and self.verbose == 2:
                gym.logger.info(f"[{self._name}] Keeping episode {episode} due to reward {episode_reward} "
                                f">= {self.min_successful_reward}")
            # Store the values
            self.observations[episode] = self._ep_observations.copy()
            self.q_values[episode] = self._ep_q_values.copy()
            self.actions[episode] = self._ep_actions.copy()
            # Reset memories
            if episode > self.reset_interval and episode % self.reset_interval == 0:
                # The last not allowed episode
                last_not_allowed = episode - self.reset_interval
                if self.verbose == 2:
                    gym.logger.info(f"[{self._name}] Removing all episodes from {last_not_allowed} and before")
                while self.q_values.get(last_not_allowed) is not None:
                    # Remove the episode
                    self.q_values.pop(last_not_allowed)
                    self.observations.pop(last_not_allowed)
                    self.actions.pop(last_not_allowed)
                    # Try earlier episodes until all removed
                    last_not_allowed -= 1
        elif self.verbose > 1:
            gym.logger.info(f"[{self._name}] Disregarding episode {episode} due to reward = {episode_reward} "
                            f"< {self.min_successful_reward}")

        # Reset the episode values
        self._ep_observations = []
        self._ep_q_values = []
        self._ep_actions = []

    def get_last_episode_experiences(self, episode, nb_episodes):
        """Get the last number of episode experiences, or as many as possible.

        Arguments:
            episode: The current episode in the experiment.
            nb_episodes: The number of episodes of experiences to return.

        Returns:
            observations, q_values: The observation and q-values of the last episodes.
        """
        nb_possible_episodes = min(episode, nb_episodes)
        observations = []
        q_values = []
        actions = []
        # Get the last nb_possible_episodes
        for episode in range(episode - nb_possible_episodes + 1, episode + 1):
            observations.extend(self.observations[episode])
            q_values.extend(self.q_values[episode])
            actions.extend(self.actions[episode])
        return np.array(observations, dtype=np.float32), np.array(q_values, dtype=np.float32), \
               np.array(actions, dtype=np.float32)
