import numpy as np
import os
from typing import List

import torch

from core.callbacks.loggers import LogEntry
from core.callbacks.sdt.sdt_callback import SDTCallback
from core.experiment.config import ExperimentConfiguration, TreeConfiguration
from core.experiment.environment_description import EnvironmentDescription
from core.paths.compare_path import ComparePaths
from run_config import CONTINUE_ON_EXISTING_DIRS


class ActionsLog(LogEntry):
    """A log entry containing loss data per epoch for a Soft Decision Tree."""
    def __init__(self, sdt_callbacks: List[SDTCallback]):
        # Super call
        super(ActionsLog, self).__init__()
        self.sdt_callbacks = sdt_callbacks
        # The fields of the entry type
        self.step = "Step"
        self.episode = "Episode"
        self.agent_action = "Agent Action"
        self.agent_distribution = "Agent Distribution"
        self.agent_order = "Agent Action Order"
        # All the entry fields
        self.entry_fields = [self.step, self.episode, self.agent_action, self.agent_distribution, self.agent_order]
        for callback in sdt_callbacks:
            callback_fields = self._create_fields(callback)
            self.entry_fields.extend(callback_fields)

    @staticmethod
    def _create_fields(sdt_callback: SDTCallback):
        name = sdt_callback.full_name
        action = f"{name} Action"
        distribution = f"{name} Distribution"
        order = f"{name} Action order"
        leaf = f"{name} Leaf"
        matches = f"{name} Matches"
        return [action, distribution, order, leaf, matches]

    def create_sdt_entry(self, sdt_callback, action, distribution, order, leaf, matches):
        fields = self._create_fields(sdt_callback)
        values = [action, distribution, order, leaf, matches]
        entry = zip(fields, values)
        return entry

    def create_entry(self, step, episode, agent_action, agent_distribution, agent_order, sdt_callback_entries):
        """Method to create an entry for the log."""
        entry = {
            self.step: step,
            self.episode: episode,
            self.agent_action: agent_action,
            self.agent_distribution: agent_distribution,
            self.agent_order: agent_order
        }
        for sdt_entry in sdt_callback_entries:
            for field, value in sdt_entry:
                entry[field] = value
        return entry


class PostActionsLog(LogEntry):
    """A log entry containing loss data per epoch for a Soft Decision Tree."""
    def __init__(self, env_description: EnvironmentDescription, tree_configuration: TreeConfiguration):
        # Super call
        super(PostActionsLog, self).__init__()
        self.env_description = env_description
        self.tree_configuration = tree_configuration
        # The fields of the entry type
        self.action = "Action"
        self.all_actions = "All actions"
        self.total_actions = "Agent"
        self.same_actions = "SDT Match"
        self.accuracy = "SDT Action Accuracy"
        self.order_avg = "Action Order Average"
        self.order_max = "Action Order Max"
        self.order_min = "Action Order Min"
        self.leafs = [f"Leaf Node {leaf_node}" for leaf_node in range(2 ** self.tree_configuration.depth)]
        # All the entry fields
        self.entry_fields = [self.action, self.total_actions, self.same_actions, self.accuracy,
                             self.order_avg, self.order_max, self.order_min]
        self.entry_fields.extend(self.leafs)

    def create_entry(self, action, total_actions, same_actions, order_avg, order_max, order_min, leaf_counts):
        """Method to create an entry for the log."""
        entry = {
            self.action: action,
            self.total_actions: total_actions,
            self.same_actions: same_actions,
            self.accuracy: same_actions / total_actions,
            self.order_avg: order_avg,
            self.order_max: order_max,
            self.order_min: order_min
        }
        for leaf_node, count in leaf_counts:
            field = f"Leaf Node {leaf_node}"
            entry[field] = count
        return entry


class ActionLogger(object):
    """A logger for storing information for actions of a DQN agent and its SDTs.

    Attributes:
        experiment_configuration: The experiment configuration used.
    """
    def __init__(self, experiment_configuration: ExperimentConfiguration,
                 env_description: EnvironmentDescription,
                 sdt_callbacks: List[SDTCallback], deterministic_trees):
        # Store the arguments
        self.exp_config = experiment_configuration
        self.env_description = env_description
        self.sdt_callbacks = sdt_callbacks
        self.deterministic_trees = deterministic_trees
        self.actions_log = ActionsLog(sdt_callbacks)
        # The log data
        self.data = []
        self.actions_count = {}
        self._all_actions = "All actions"
        self.num_steps = {}
        self._init_data()
        # Path to log files
        path = ComparePaths.directory(self.exp_config)
        os.makedirs(path, exist_ok=CONTINUE_ON_EXISTING_DIRS)

    def _init_data(self):
        for action in range(self.env_description.num_actions()):
            # action -> sdt -> (same_action, order_matches, leafs)
            self.actions_count[action] = {c.full_name: [0, [], []] for c in self.sdt_callbacks}
            self.num_steps[action] = 0
        self.actions_count[self._all_actions] = {c.full_name: [0, [], []] for c in self.sdt_callbacks}
        self.num_steps[self._all_actions] = 0

    def compare_actions(self, step, episode, state, agent_action, agent_q_values):
        # Get the prediction of each SDT for the state
        processed_observation = torch.from_numpy(np.array([state]).astype(np.float32))
        agent_order = agent_q_values.argsort()
        sdt_entries = []
        for sdt_callback in self.sdt_callbacks:
            action, leaf, distribution = \
                sdt_callback.soft_decision_tree.predict(processed_observation, deterministic=self.deterministic_trees)
            distribution = distribution.numpy()
            # How accurate does the SDT choose the same action (& distribution) as the DQN agent?
            same_action = 1 if action == agent_action else 0
            order = distribution.argsort()
            # noinspection PyUnresolvedReferences
            matches = (order == agent_order).sum()
            sdt_entry = self.actions_log.create_sdt_entry(sdt_callback, action, distribution, order, leaf, matches)
            sdt_entries.append(sdt_entry)
            # For this specific agent action
            self.actions_count[agent_action][sdt_callback.full_name][0] += same_action
            self.actions_count[agent_action][sdt_callback.full_name][1].append(matches)
            self.actions_count[agent_action][sdt_callback.full_name][2].append(leaf)
            # Overall
            self.actions_count[self._all_actions][sdt_callback.full_name][0] += same_action
            self.actions_count[self._all_actions][sdt_callback.full_name][1].append(matches)
            self.actions_count[self._all_actions][sdt_callback.full_name][2].append(leaf)
        # Create the entry
        entry = self.actions_log.create_entry(step, episode, agent_action, agent_q_values, agent_order, sdt_entries)
        self.data.append(entry)
        self.num_steps[self._all_actions] += 1
        self.num_steps[agent_action] += 1

    def save_logs(self):
        """Save the logs."""
        path = ComparePaths.log_actions_path(self.exp_config)
        self.actions_log.write_data(self.data, path)

    def finish_logs(self):
        # For each SDT callback
        for sdt_callback in self.sdt_callbacks:
            log = PostActionsLog(self.env_description, sdt_callback.tree_conf)
            entries = []
            # Make an overview from the action counts
            for action in self.actions_count.keys():
                # Get the name for the (overall) action
                action_name = action if action == self._all_actions else self.env_description.actions[action]
                # Number of correct predictions for this action
                correct_actions, order_matches, leafs = self.actions_count[action][sdt_callback.full_name]
                order_matches = np.array(order_matches)
                leafs = np.array(leafs)
                nodes, count = np.unique(leafs, return_counts=True)
                leaf_counts = zip(nodes, count)
                # Create the entry
                entry = log.create_entry(action_name, self.num_steps[action], correct_actions, order_matches.mean(),
                                         order_matches.max(), order_matches.min(), leaf_counts)
                entries.append(entry)
            # Store the logs
            directory = ComparePaths.log_tree_config_directory(self.exp_config, sdt_callback.tree_conf)
            os.makedirs(directory, exist_ok=True)
            path = ComparePaths.log_specific_path(self.exp_config, sdt_callback.tree_conf, sdt_callback.name)
            log.write_data(entries, path)
