import os

import gym

from core.callbacks.dqn.agent_callback import AgentCallback
from core.callbacks.loggers import LogEntry
from core.experiment.config import ExperimentConfiguration
from core.models.dqn.dqn_agent import DQNAgent
from core.models.dqn.policy import EpsilonGreedyPolicy, BoltzmannPolicy
from core.paths.dqn_path import DQNPaths
from run_config import CONTINUE_ON_EXISTING_DIRS


LOG_STEP_SAVE_INTERVAL = 100000
LOG_EPISODE_SAVE_INTERVAL = 1000


class DQNStepLog(LogEntry):
    """A log entry containing data per step for a DQN agent."""
    def __init__(self):
        # Super call
        super(DQNStepLog, self).__init__()
        # The fields of the entry type
        self.step = "Step"
        self.episode = "Episode"
        self.state = "State"
        self.action = "Action"
        self.q_values = "Q-values"
        self.reward = "Reward"
        self.loss = "Loss"
        self.policy = "Policy"
        self.policy_value_name = "Policy Value Name"
        self.policy_value = "Policy Value"
        # All the entry fields
        self.entry_fields = [self.step, self.episode, self.state, self.action, self.q_values, self.reward, self.loss,
                             self.policy, self.policy_value_name, self.policy_value]

    def create_entry(self, step, episode, state, action, q_values, reward, loss,
                     # policy_name, policy_value_name, policy_value
                     ):
        """Method to create an entry for the log."""
        return {
            self.step: step,
            self.episode: episode,
            self.state: state,
            self.action: action,
            self.q_values: q_values,
            self.reward: reward,
            self.loss: loss,
            # self.policy: policy_name,
            # self.policy_value_name: policy_value_name,
            # self.policy_value: policy_value,
        }

    def add_policy_data(self, entry, policy, value):
        # Add the policy data to the entry
        entry[self.policy] = policy.name
        entry[self.policy_value_name] = policy.value_name
        entry[self.policy_value] = value
        return entry


class DQNEpisodeLog(LogEntry):
    """A log entry containing data per episode for a DQN agent."""
    def __init__(self):
        # Super call
        super(DQNEpisodeLog, self).__init__()
        # The fields of the entry type
        self.learning_step = "Learning Step"
        self.episode = "Episode"
        self.episode_reward = "Reward"
        self.episode_steps = "Steps"
        # All the entry fields
        self.entry_fields = [self.episode, self.learning_step, self.episode_reward, self.episode_steps, ]

    def create_entry(self, step, episode, episode_reward, episode_steps):
        """Method to create an entry for the log."""
        return {
            self.learning_step: step,
            self.episode: episode,
            self.episode_reward: episode_reward,
            self.episode_steps: episode_steps,
        }


class MemoryLog(LogEntry):
    """A log entry containing data for experience distributions for a replay memory."""

    def __init__(self, env_description):
        # Super call
        super(MemoryLog, self).__init__()
        self._env_description = env_description
        # The fields of the entry type
        self.step = "Step"
        self.episode = "Episode"
        self.counts = [f"{env_description.actions[idx]}" for idx in env_description.actions.keys()]
        # All the entry fields
        self.entry_fields = [self.step, self.episode]
        self.entry_fields.extend(self.counts)

    def create_entry(self, step, episode, counts_per_actions):
        """Method to create an entry for the log."""
        entry = {self.step: step, self.episode: episode}
        for a, count in enumerate(counts_per_actions):
            field = f"{self._env_description.actions[a]}"
            entry[field] = count
        return entry


class SDTMemoryLog(LogEntry):
    """A log entry containing data for experience distributions for an SDT-enhanced replay memory."""

    def __init__(self, num_leafs, env_description):
        # Super call
        super(SDTMemoryLog, self).__init__()
        self._env_description = env_description
        # The fields of the entry type
        self.step = "Step"
        self.episode = "Episode"
        self.leafs = [f"Leaf {leaf}" for leaf in range(num_leafs)]
        self.counts = []
        for leaf in self.leafs:
            for action in self._env_description.actions.keys():
                self.counts.append(f"{leaf}: {self._env_description.actions[action]}")
        # All the entry fields
        self.entry_fields = [self.step, self.episode]
        self.entry_fields.extend(self.counts)

    def create_entry(self, step, episode, counts_per_actions_and_leafs):
        """Method to create an entry for the log."""
        entry = {self.step: step, self.episode: episode}
        for leaf, action_counts in enumerate(counts_per_actions_and_leafs):
            for a, count in enumerate(action_counts):
                field = f"Leaf {leaf}: {self._env_description.actions[a]}"
                entry[field] = count
        return entry


class DQNLogger(AgentCallback):
    """A logger for storing information for a DQN.

    Attributes:
        experiment_configuration: The ExperimentConfiguration in which the callback is used.
        agent: The DQN agent to store the logs for.
        step_save_interval: The save and reset interval for the step log data.
        episode_save_interval: The save and reset interval for the episode log data.
        training_agent: Whether or not the agent is being trained or evaluated.
        best_save: Whether or not the best saved model is used.
    """
    def __init__(self, experiment_configuration: ExperimentConfiguration, agent: DQNAgent,
                 step_save_interval=LOG_STEP_SAVE_INTERVAL, episode_save_interval=LOG_EPISODE_SAVE_INTERVAL,
                 training_agent=True, best_save=False):
        # Super call
        super(DQNLogger, self).__init__(agent)
        # Store the arguments
        self.experiment_configuration = experiment_configuration
        self.step_save_interval = step_save_interval
        self.episode_save_interval = episode_save_interval
        self.training_dqn = training_agent
        self.best_save = best_save
        # The log data
        self.step_log_entry = DQNStepLog()
        self.episode_log_entry = DQNEpisodeLog()
        # memory_log_entry initialised once memory is initialised
        self.memory_log_entry = None
        self.step_data = []
        self._steps = 0
        self.episode_data = []
        self.memory_data = []
        # Path to log directory
        dqn_log_directory = DQNPaths.log_directory(experiment_configuration, training_agent, self.best_save)
        try:
            os.makedirs(dqn_log_directory, exist_ok=CONTINUE_ON_EXISTING_DIRS)
        except FileExistsError:
            gym.logger.error(f"A DQN log directory for the given experiment already exists at "
                             f"'{dqn_log_directory}'.\nExperiment exiting due to "
                             f"'CONTINUE_ON_EXISTING_DIRS' = {CONTINUE_ON_EXISTING_DIRS}"
                             f" to protect experiment results being overwritten.")
            exit()

    def on_step_end(self, step, logs=None):
        """Method to log a step.

        Args:
            step: The current episode the agent is in.
            logs: The logging data produced by the DQN agent.
        """
        # Create the entry
        entry = logs.copy()
        # Get the name and possible value of the policy
        policy = self.agent.policy
        # If policy has a value
        if policy.value_name is not None:
            # Epsilon greedy policy
            if isinstance(policy, EpsilonGreedyPolicy):
                val = policy.epsilon
            # Boltzmann policy
            elif isinstance(policy, BoltzmannPolicy):
                val = policy.tau
            # Not supported (yet)
            else:
                raise ValueError(f"Value extraction not defined for this policy: {policy}")
        # No value for the policy
        else:
            val = None
        entry = self.step_log_entry.add_policy_data(entry, policy, val)
        # Store the entry
        self.step_data.append(entry)
        self._steps += 1

        # Store the memory log
        if self.training_dqn:
            counts = self.agent.memory.last_counts
            entry = self.memory_log_entry.create_entry(step, logs[self.step_log_entry.episode], counts)
            self.memory_data.append(entry)

        # Write and reset the entries
        if step > 0 and step % self.step_save_interval == 0:
            path = DQNPaths.step_logs_path(self.experiment_configuration, self.training_dqn, step, self.best_save)
            self.step_log_entry.write_data(self.step_data, path)
            if self.training_dqn:
                path = DQNPaths.memory_logs_path(self.experiment_configuration, self.training_dqn, step, self.best_save)
                self.memory_log_entry.write_data(self.memory_data, path)
                del self.memory_data
                self.memory_data = []
            # Delete the data from memory
            del self.step_data
            self.step_data = []
            self._steps = 0

    def on_episode_end(self, step, episode, logs=None):
        """Method to log an episode.

        Args:
            step: The step where the episode ended.
            episode: The current episode the agent is in.
            logs: The logging data produced by the DQN agent.
        """
        # Store the entry
        self.episode_data.append(logs)
        # Save and reset each interval
        if episode > 0 and episode % self.episode_save_interval == 0:
            # Save the episode data
            path = DQNPaths.episode_logs_path(self.experiment_configuration, self.training_dqn, episode, self.best_save)
            self.episode_log_entry.write_data(self.episode_data, path)
            # Delete the data from memory
            del self.episode_data
            self.episode_data = []

    def delete_last_episode(self, num_steps):
        self.step_data = self.step_data[:self._steps-num_steps]
        self.memory_data = self.memory_data[:self._steps-num_steps]

    def save_end(self):
        path = DQNPaths.step_logs_path(self.experiment_configuration, self.training_dqn, best_save=self.best_save)
        self.step_log_entry.write_data(self.step_data, path)
        if self.training_dqn:
            path = DQNPaths.memory_logs_path(self.experiment_configuration, self.training_dqn, best_save=self.best_save)
            self.memory_log_entry.write_data(self.memory_data, path)
        path = DQNPaths.episode_logs_path(self.experiment_configuration, self.training_dqn, best_save=self.best_save)
        self.episode_log_entry.write_data(self.episode_data, path)

    def on_train_end(self, logs=None):
        """Make sure everything is saved by writing all remaining data one last time to the files."""
        self.save_end()

    def on_test_end(self, logs=None):
        """Make sure everything is saved by writing all remaining data one last time to the files."""
        self.save_end()
