import os

from core.callbacks.loggers import LogEntry
from core.experiment.config import ExperimentConfiguration
from core.experiment.environment_description import get_env_description, EnvironmentDescription
from core.models.soft_decision_tree import SoftDecisionTree
from core.paths.sdt_path import SDTPaths
from run_config import CONTINUE_ON_EXISTING_DIRS


class SDTLossesLog(LogEntry):
    """A log entry containing loss data per epoch for a Soft Decision Tree."""
    def __init__(self):
        # Super call
        super(SDTLossesLog, self).__init__()
        # The fields of the entry type
        self.step = "Step"
        self.episode = "Episode"
        self.epoch = "Epoch"
        self.agent_loss = "Agent Loss"
        self.total_loss = "Total Loss"
        self.leafs_loss = "Leafs Loss"
        self.inners_loss = "Inners Loss"
        # All the entry fields
        self.entry_fields = [self.epoch, self.step, self.episode, self.agent_loss,
                             self.total_loss, self.leafs_loss, self.inners_loss, ]

    def create_entry(self, step, episode, epoch, agent_loss, total_loss, leafs_loss, inners_loss):
        """Method to create an entry for the log."""
        return {
            self.step: step,
            self.episode: episode,
            self.epoch: epoch,
            self.agent_loss: agent_loss,
            self.total_loss: total_loss,
            self.leafs_loss: leafs_loss,
            self.inners_loss: inners_loss,
        }


class SDTInnerWeightsLog(LogEntry):
    """A log entry containing weights data per epoch for a inner node of a Soft Decision Tree."""
    def __init__(self, environment_description: EnvironmentDescription):
        # Super call
        super(SDTInnerWeightsLog, self).__init__()
        # The fields of the entry type
        self._description = environment_description
        self.step = "Step"
        self.episode = "Episode"
        self.epoch = "Epoch"
        self.inner_node = [f"({idx}) {self._description.observations[idx]}"
                           for idx in range(self._description.num_observations())]
        self.beta = "Beta"
        self.bias = "Bias"
        # All the entry fields
        self.entry_fields = [self.epoch, self.step, self.episode]
        self.entry_fields.extend(self.inner_node)
        self.entry_fields.extend([self.beta, self.bias])

    def create_entry(self, step, episode, epoch, inner_node, beta, bias):
        """Method to create an entry for the log."""
        entry = {
            self.step: step,
            self.episode: episode,
            self.epoch: epoch,
            self.beta: beta,
            self.bias: bias,
        }
        for idx in range(self._description.num_observations()):
            entry[self.inner_node[idx]] = inner_node[idx]
        return entry


class SDTLeafWeightsLog(LogEntry):
    """A log entry containing weights data per epoch for a leaf node of a Soft Decision Tree."""
    def __init__(self, environment_description: EnvironmentDescription):
        # Super call
        super(SDTLeafWeightsLog, self).__init__()
        # The fields of the entry type
        self._description = environment_description
        self.step = "Step"
        self.episode = "Episode"
        self.epoch = "Epoch"
        self.leaf_node = [self._description.actions[idx]
                          for idx in range(self._description.num_actions())]
        # All the entry fields
        self.entry_fields = [self.epoch, self.step, self.episode]
        self.entry_fields.extend(self.leaf_node)

    def create_entry(self, step, episode, epoch, leaf_node):
        """Method to create an entry for the log."""
        entry = {
            self.step: step,
            self.episode: episode,
            self.epoch: epoch,
        }
        for idx in range(self._description.num_actions()):
            entry[self.leaf_node[idx]] = leaf_node[idx]
        return entry


class SDTActionsLog(LogEntry):
    """A log entry containing loss data per epoch for a Soft Decision Tree."""
    def __init__(self):
        # Super call
        super(SDTActionsLog, self).__init__()
        # The fields of the entry type
        self.step = "Step"
        self.episode = "Episode"
        self.epoch = "Epoch"
        self.action = "Action"
        self.leaf = "Leaf"
        self.leaf_distribution = "Leaf Distribution"
        self.agent_action = "Agent Action"
        self.agent_q_values = "Agent q-values"
        #
        self.same_action = "Same Action"
        self.full_order_match = "Full Action Order Match"
        self.order_match = "Action Order Match"
        self.max_match = "Maximum Match"
        self.error = "Distribution Error"
        # All the entry fields
        self.entry_fields = [self.epoch, self.step, self.episode,
                             self.action, self.leaf, self.leaf_distribution,
                             self.agent_action, self.agent_q_values,
                             #
                             self.same_action, self.full_order_match,
                             self.order_match, self.max_match, self.error]

    # noinspection PyTypeChecker,PyUnresolvedReferences
    def create_entry(self, step, episode, epoch, action, leaf, leaf_distribution, agent_action, agent_q_values):
        """Method to create an entry for the log."""
        same_action = 1 if action == agent_action else 0
        sdt_order = leaf_distribution.argsort()
        agent_order = agent_q_values.argsort()
        equals = sdt_order == agent_order
        max_matches = len(equals)
        full_order_match = 1 if equals.all() else 0
        matches = equals.sum()
        error = agent_q_values - leaf_distribution
        return {
            self.step: step,
            self.episode: episode,
            self.epoch: epoch,
            self.action: action,
            self.leaf: leaf,
            self.leaf_distribution: leaf_distribution,
            self.agent_action: agent_action,
            self.agent_q_values: agent_q_values,
            #
            self.same_action: same_action,
            self.full_order_match: full_order_match,
            self.order_match: matches,
            self.max_match: max_matches,
            self.error: error
        }


class SDTPostActionsLog(LogEntry):
    """A log entry containing loss data per epoch for a Soft Decision Tree."""
    def __init__(self, environment_description: EnvironmentDescription):
        # Super call
        super(SDTPostActionsLog, self).__init__()
        # The fields of the entry type
        self._description = environment_description
        # The fields of the entry type
        self.step = "Step"
        self.episode = "Episode"
        self.epoch = "Epoch"
        self.action = "DQN Action"
        self.action_accuracy = "Action Accuracy"
        self.full_order_match = "Full Order Match Accuracy"
        self.order_match_min = "Order Match Min"
        self.order_match_max = "Order Match Max"
        self.order_match_avg = "Order Match Average"
        # All the entry fields
        self.entry_fields = [self.epoch, self.step, self.episode, self.action,
                             self.action_accuracy, self.full_order_match, self.order_match_min,
                             self.order_match_max, self.order_match_avg]

    def create_entry(self, step, episode, epoch, action, action_accuracy, full_order_match,
                     order_match_min, order_match_max, order_match_avg):
        """Method to create an entry for the log."""
        return {
            self.step: step,
            self.episode: episode,
            self.epoch: epoch,
            self.action: action,
            self.action_accuracy: action_accuracy,
            self.full_order_match: full_order_match,
            self.order_match_min: order_match_min,
            self.order_match_max: order_match_max,
            self.order_match_avg: order_match_avg
        }


class SDTLogger(object):
    """A logger for storing information for a SDT.

    Attributes:
        experiment_configuration: The experiment configuration used.
        soft_decision_tree: The soft decision tree for the callback.
        sdt_callback_name: The name of the SDT callback using the logger.
        save_epoch_interval: The save and reset interval for the log data.
    """
    def __init__(self, experiment_configuration: ExperimentConfiguration,
                 soft_decision_tree: SoftDecisionTree,
                 sdt_callback_name, save_epoch_interval):
        # Store the arguments
        self.exp_config = experiment_configuration
        self.soft_decision_tree = soft_decision_tree
        self.sdt_callback_name = sdt_callback_name
        self.tree_config = self.soft_decision_tree.tree_configuration
        # The log data
        self.losses = []
        self.actions = []
        self.inner_nodes_steps = {}
        self.leaf_nodes_steps = {}
        # How often to write data to file
        self.save_epoch_interval = save_epoch_interval
        description = get_env_description(self.exp_config.env_name)
        self.log_loss_entry = SDTLossesLog()
        self.log_inners_entry = SDTInnerWeightsLog(description)
        self.log_leafs_entry = SDTLeafWeightsLog(description)
        self.log_actions_entry = SDTActionsLog()
        # Path to log files
        nodes_directory = SDTPaths.log_weights_directory(self.exp_config, self.tree_config, self.sdt_callback_name)
        os.makedirs(nodes_directory, exist_ok=CONTINUE_ON_EXISTING_DIRS)
        # Initialize the weights
        for idx in range(self.soft_decision_tree.num_inners):
            self.inner_nodes_steps[idx] = []
        for idx in range(self.soft_decision_tree.num_leafs):
            self.leaf_nodes_steps[idx] = []

    def log_losses(self, step, episode, epoch, agent_loss, total_loss, leafs_loss, inners_loss):
        """Method to log the step when the soft decision tree has been trained."""
        # Create and store the entry
        entry = self.log_loss_entry.create_entry(step, episode, epoch, agent_loss, total_loss, leafs_loss, inners_loss)
        self.losses.append(entry)

    def log_actions(self, entries):
        """Method to log the actions chosen by the soft decision tree after each training epoch."""
        # Store the entries
        self.actions.extend(entries)

    def log_weights(self, step, episode, epoch):
        """Method to log the weights of the trained soft decision tree."""
        # Get the parameters of the soft decision tree
        inners, betas, biases, leafs = self.soft_decision_tree.get_parameters()
        # Store the values for each inner node
        for idx, (inner_node, beta, bias) in enumerate(zip(inners, betas, biases)):
            # Create and store the entry
            entry = self.log_inners_entry.create_entry(step, episode, epoch, inner_node, beta, bias)
            self.inner_nodes_steps[idx].append(entry)
        # Store the leaf nodes' parameters
        for idx, leaf_node in enumerate(leafs):
            # Create and store the entry
            entry = self.log_leafs_entry.create_entry(step, episode, epoch, leaf_node)
            self.leaf_nodes_steps[idx].append(entry)

    def save_logs(self, epoch):
        """Save the logs each interval."""
        # Save the logs to a file
        if epoch > 0 and epoch % self.save_epoch_interval == 0:
            self.save_all(epoch)

    def save_all(self, epoch=None):
        """Make sure everything is saved by writing all data one last time to the files."""
        # Losses
        path = SDTPaths.log_losses_path(self.exp_config, self.tree_config, self.sdt_callback_name, epoch)
        self.log_loss_entry.write_data(self.losses, path)
        del self.losses
        self.losses = []
        # Actions
        path = SDTPaths.log_actions_path(self.exp_config, self.tree_config, self.sdt_callback_name, epoch)
        self.log_actions_entry.write_data(self.actions, path)
        del self.actions
        self.actions = []
        # Weights
        for idx in range(self.soft_decision_tree.num_inners):
            path = SDTPaths.log_inner_path(self.exp_config, self.tree_config, self.sdt_callback_name, idx, epoch)
            self.log_inners_entry.write_data(self.inner_nodes_steps[idx], path)
            self.inner_nodes_steps.pop(idx)
            self.inner_nodes_steps[idx] = []
        for idx in range(self.soft_decision_tree.num_leafs):
            path = SDTPaths.log_leaf_path(self.exp_config, self.tree_config, self.sdt_callback_name, idx, epoch)
            self.log_leafs_entry.write_data(self.leaf_nodes_steps[idx], path)
            self.leaf_nodes_steps.pop(idx)
            self.leaf_nodes_steps[idx] = []
