from core.callbacks.dqn.memory_callback import EpisodeMemoryCallback
from core.callbacks.sdt.sdt_callback import SDTCallbackTemplate, SDTCallback
from core.experiment.config import ExperimentConfiguration, TreeConfiguration
from core.models.dqn.policy import Policy


class EpisodeCallback(SDTCallback):
    """The callback class for a SDT trained at regular episode intervals.

    Attributes:
        episode_interval: The number of episodes between each SDT training process.
        episodes: The number of episodes to use for 1 training process.
    """

    def __init__(self, episode_interval, episodes, training_agent, agent_warm_up_steps, **kwargs):
        # Super call
        super(EpisodeCallback, self).__init__(**kwargs)
        # Store attributes
        self.episode_interval = episode_interval
        self.episodes = episodes
        self.step = 0
        self.current_episode = 0
        self.training_agent = training_agent
        self.agent_warm_up_steps = agent_warm_up_steps

    def on_train_begin(self, logs=None):
        """Callback when training begins"""
        super(EpisodeCallback, self).on_train_begin(logs)
        if not self.successful_episodes_only:
            self.current_episode = 0

    def on_episode_end(self, step, episode, logs=None):
        # Train the SDT every episode_interval episodes
        agent_ready = (not self.training_agent) or step >= self.agent_warm_up_steps
        if episode > 0 and episode % self.episode_interval == 0 and agent_ready:
            self.train_sdt(step, episode)

    def _get_train_data(self, step, episode, split_in_batches):
        """Gather the training data for the SDT."""
        # Get the experiences and split them up experience per batch and transform to PyTorch tensors
        observations, q_values, actions = self.memory.get_last_episode_experiences(episode, self.episodes)
        full_batches, remainder = self._get_batches_available(observations=observations)
        if split_in_batches:
            split_observations, split_q_values, split_actions = \
                self._split_in_batches(observations, q_values, actions, full_batches, remainder, step, episode)
        else:
            split_observations, split_q_values, split_actions = observations, q_values, actions
        # Return the observations
        return split_observations, split_q_values, split_actions, full_batches, remainder


class EpisodeCallbackTemplate(SDTCallbackTemplate):
    """The class representing a template for a an Episode SDT callback.

    Attributes:
        episode_interval: The number of episodes between each SDT training process.
        episodes: The number of episodes to use for 1 training process.
    """
    def __init__(self, episode_interval, episodes, training_agent, agent_warm_up_steps, **kwargs):
        # Super call
        super(EpisodeCallbackTemplate, self).__init__(**kwargs)
        # Store the arguments
        self.episode_interval = episode_interval
        self.episodes = episodes
        self.training_agent = training_agent
        self.agent_warm_up_steps = agent_warm_up_steps

    def create_callback(self, experiment_configuration: ExperimentConfiguration,
                        tree_configuration: TreeConfiguration,
                        memory: EpisodeMemoryCallback,
                        agent_policy: Policy):
        """Method to create a callback, based on this template, for the given SDT and arguments."""
        name, soft_decision_tree, logger = self._setup_arguments(experiment_configuration, tree_configuration)
        callback = EpisodeCallback(self.episode_interval, self.episodes, self.training_agent, self.agent_warm_up_steps,
                                   name=name, agent_policy=agent_policy,
                                   soft_decision_tree=soft_decision_tree, memory=memory, logger=logger,
                                   epochs=self.epochs, save_epoch_interval=self.save_epoch_interval,
                                   log_interval=self.log_interval,
                                   successful_episodes_only=self.successful_episodes_only,
                                   boltzmann_tau=self.boltzmann_tau)
        return callback

    def create_name(self):
        """Method to generate the name of the callback."""
        return f"EpisodeCallback_{self.episode_interval}_{self.episodes}"
