import os
from typing import List

import numpy as np
import torch
import torch.multiprocessing as mp

from core.callbacks.callback import Callback, CallbackGroup
from core.callbacks.dqn.memory_callback import EpisodeMemoryCallback
from core.callbacks.loggers.dqn_logger import DQNStepLog
from core.callbacks.loggers.sdt_logger import SDTLogger, SDTActionsLog
from core.experiment.config import ExperimentConfiguration, TreeConfiguration
from core.models.dqn.policy import Policy
from core.models.soft_decision_tree import SoftDecisionTree
from core.paths.sdt_path import SDTPaths
from run_config import CONTINUE_ON_EXISTING_DIRS


class SDTCallback(Callback):
    """The class representing a callback for a Soft Decision Tree.

    Attributes:
        name: The name to give to the callback. Used for logging and saving
            SDT information.
        soft_decision_tree: The SDT the callback is meant for.
        memory: The memory callback for the agent, containing the necessary
            experiences to train a SDT.
        logger: The SDT logger, storing the losses, weights and actions of the SDT.
        agent_policy: The policy used by the agent to select actions.
        epochs: The number of epochs when training on batches.
        save_epoch_interval: The saving interval expressed in number of epochs.
        log_interval: The logging interval expressed in number of epochs.
        successful_episodes_only: (Optional) Whether or not the SDT should only be
            trained for successful episodes.
        boltzmann_tau: (Optional) The temperature tau if using a Boltzmann distribution
            for the target data.
    """

    def __init__(self, name, soft_decision_tree: SoftDecisionTree,
                 memory: EpisodeMemoryCallback, logger: SDTLogger,
                 agent_policy: Policy, epochs, save_epoch_interval, log_interval,
                 successful_episodes_only=False, boltzmann_tau=None):
        # Store the SDT, DQN memory and SDT logger
        self.soft_decision_tree = soft_decision_tree
        self.memory = memory
        self.logger = logger
        self.agent_policy = agent_policy
        # Store other arguments
        self.name = name
        self.epochs = epochs
        self.save_epoch_interval = save_epoch_interval
        self.log_interval = log_interval
        self.successful_episodes_only = successful_episodes_only
        self.boltzmann_tau = boltzmann_tau
        # Further setup
        self.exp_conf = soft_decision_tree.experiment_configuration
        self.tree_conf = soft_decision_tree.tree_configuration
        self.batch_size = self.tree_conf.batch_size
        self.input_shape = self.exp_conf.input_shape[0]
        self.save_directory = SDTPaths.directory(self.exp_conf, self.tree_conf, self.name)
        os.makedirs(self.save_directory, exist_ok=CONTINUE_ON_EXISTING_DIRS)
        self.full_name = f"{self.tree_conf.name} {self.name}"
        # Keep track of progress
        self.current_epoch = 0
        self.agent_loss = 0
        self.log_actions_entry = SDTActionsLog()
        self.dqn_entry = DQNStepLog()

    def on_train_begin(self, logs=None):
        """Callback when training begins"""
        self.current_epoch = 0
        self.agent_loss = 0

    def on_train_end(self, logs=None):
        """Callback when training ends"""
        self.logger.save_all()

    def on_test_end(self, logs=None):
        """Callback when training ends"""
        self.logger.save_all()

    def on_step_end(self, step, logs=None):
        """Callback when a step ends"""
        self.agent_loss = logs[self.dqn_entry.loss]

    def train_sdt(self, step, episode):
        """Train the SDT"""
        # Get the data
        split_in_batches = True
        observations, q_values, actions, full_batches, remainder = self._get_train_data(step, episode, split_in_batches)
        # Train the SDT
        for epoch in range(self.epochs):
            # Save log data
            self.logger.save_logs(self.current_epoch)
            self.current_epoch += 1
            # Gather new data
            data = zip(observations, q_values)
            total_loss, leafs_loss, inners_loss = \
                self.soft_decision_tree.train_on_data_(data, self.current_epoch, self.tree_conf.name, self.name,
                                                       full_batches, remainder, epoch, self.epochs)
            # Log the losses
            self.logger.log_losses(step, episode, self.current_epoch,
                                   self.agent_loss, total_loss, leafs_loss, inners_loss)
            # Log the weights
            self.logger.log_weights(step, episode, self.current_epoch)
            if self.current_epoch > 0:
                # Log the actions
                if self.current_epoch % self.log_interval == 0:
                    action_entries = \
                        self._get_action_predictions(observations, q_values, actions, full_batches, remainder,
                                                     split_in_batches, step, episode, self.current_epoch)
                    self.logger.log_actions(action_entries)
            # An other epoch has passed
            # self.current_epoch += 1
        # Save the SDT
        if self.current_epoch > 0:
            path = SDTPaths.save_path(self.exp_conf, self.tree_conf, self.name, epoch_save=self.current_epoch)
            self.soft_decision_tree.save_sdt(path)

    def _get_action_predictions(self, observations, q_values, actions, full_batches, remainder, split_in_batches,
                                step, episode, epoch):
        entries = []
        if not split_in_batches:
            observations, q_values, actions = \
                self._split_in_batches(np.array(observations), np.array(q_values), np.array(actions),
                                       full_batches, remainder, step, episode)
        for observation_batch, q_values_batch, action_batch in zip(observations, q_values, actions):
            for state, agent_q, agent_a in zip(observation_batch, q_values_batch, action_batch):
                state = torch.stack([state])
                action, chosen_leaf, leaf_distribution = \
                    self.soft_decision_tree.predict(state, deterministic=self.boltzmann_tau is None)
                entry = self.log_actions_entry.create_entry(step, episode, epoch, action, chosen_leaf[0].item(),
                                                            leaf_distribution[0].numpy(),
                                                            int(agent_a.item()), agent_q.numpy())
                entries.append(entry)
        return entries

    def _get_train_data(self, step, episode, split_in_batches):
        """Gather the training data for the SDT."""
        raise NotImplementedError

    def _get_batches_available(self, observations=None):
        """Get the number of batches available."""
        total_experiences = len(self.memory.observations) if observations is None else len(observations)
        full_batches, remainder = divmod(total_experiences, self.batch_size)
        return full_batches, remainder

    def _split_in_batches(self, observations, q_values, actions, full_batches, remainder, step, episode):
        """Split the data into batches."""
        split_observations = []
        split_q_values = []
        split_actions = []
        for batch in range(full_batches + 1):
            if not (batch >= full_batches and remainder == 0):
                batch_size = self.batch_size if batch < full_batches else remainder
                low = batch * self.batch_size
                high = low + batch_size
                batch_observations = torch.from_numpy(observations[low:high])
                batch_q_values = torch.from_numpy(self._calculate_targets(q_values[low:high], step, episode))
                split_observations.append(batch_observations)
                split_q_values.append(batch_q_values)
                if actions is not None:
                    batch_actions = torch.from_numpy(actions[low:high])
                    split_actions.append(batch_actions)
        if actions is not None:
            return split_observations, split_q_values, split_actions
        else:
            return split_observations, split_q_values

    def _calculate_targets(self, q_values, step, episode):
        # Get the probability distribution from the agent's policy
        return self.agent_policy.get_distribution(q_values, step, episode)


class SDTCallbackTemplate(object):
    """The class representing a template for a SDT callback.

    This class is used when setting up the experiments to generate
    1 callback for each type of tree configuration used.

    Attributes:
        epochs: The number of epochs when training on batches.
        save_epoch_interval: The saving interval expressed in number of epochs.
        log_interval: The logging interval expressed in number of epochs.
        successful_episodes_only: (Optional) Whether or not the SDT should only be
            trained for successful episodes.
        boltzmann_tau: (Optional) The temperature tau if using a Boltzmann distribution
            for the target data.
    """

    def __init__(self, epochs, save_epoch_interval, log_interval,
                 successful_episodes_only=False, boltzmann_tau=None):
        # Store the arguments
        self.epochs = epochs
        self.save_epoch_interval = save_epoch_interval
        self.log_interval = log_interval
        self.successful_episodes_only = successful_episodes_only
        self.boltzmann_tau = boltzmann_tau

    def create_callback(self, experiment_configuration: ExperimentConfiguration,
                        tree_configuration: TreeConfiguration,
                        memory: EpisodeMemoryCallback,
                        agent_policy: Policy):
        """Method to create a callback, based on this template, for the given SDT and arguments.

        Args:
            experiment_configuration: The experiment configuration to create the callback for.
            tree_configuration: The tree configuration of the SDT to create.
            memory: The memory callback of the DQN agent.
            agent_policy: The policy used by the agent.

        Returns:
            A SDTCallback.
        """
        raise NotImplementedError

    def create_name(self):
        """Method to generate the name of the callback."""
        raise NotImplementedError

    def _setup_arguments(self, experiment_configuration: ExperimentConfiguration,
                         tree_configuration: TreeConfiguration):
        """Method to setup the default attributes of the callback."""
        name = self.create_name()
        soft_decision_tree = SoftDecisionTree(experiment_configuration, tree_configuration)
        logger = SDTLogger(experiment_configuration, soft_decision_tree, name, self.save_epoch_interval)
        return name, soft_decision_tree, logger


def on_train_begin(callback):
    return callback.on_train_begin


def on_train_end(callback):
    return callback.on_train_end


def on_test_begin(callback):
    return callback.on_test_begin


def on_test_end(callback):
    return callback.on_test_end


def on_step_begin(callback):
    return callback.on_step_begin


def on_step_end(callback):
    return callback.on_step_end


def on_episode_begin(callback):
    return callback.on_episode_begin


def on_episode_end(callback):
    return callback.on_episode_end


class SDTCallbackGroup(CallbackGroup):
    """The class representing a group of callbacks.

    Attributes:
        callbacks: A list of callbacks in the group.
        threads: The number of threads available to run the callbacks.
    """

    def __init__(self, callbacks: List[SDTCallback], threads):
        # Sort the callbacks: start with the biggest SDTs to get those going first
        callbacks.sort(key=lambda c: c.soft_decision_tree.depth, reverse=True)
        # Super call
        super(SDTCallbackGroup, self).__init__(callbacks)
        # Store the arguments
        self.threads = threads
        # Only use the number of threads available, or as many as needed for the callbacks
        self.used_threads = min(len(self.callbacks), threads)

        # Hard-coded switch to decide when to run single threaded instead of multi-threaded
        self.train_begin_multi = False
        self.train_end_multi = False
        self.test_begin_multi = False
        self.test_end_multi = False
        self.step_begin_multi = False
        self.step_end_multi = False
        self.episode_begin_multi = False
        self.episode_end_multi = False

    def _run_all_callbacks(self, method, args, run_multi):
        # Run multi-threaded
        if run_multi:
            # With processes
            all_processes = []
            for callback in self.callbacks:
                process = mp.Process(target=method(callback), args=args)
                process.start()
                all_processes.append(process)
            for process in all_processes:
                process.join()
        # Run single threaded
        else:
            for callback in self.callbacks:
                method(callback)(*args)

    def on_train_begin(self, logs=None):
        args = (logs,)
        self._run_all_callbacks(on_train_begin, args, self.train_begin_multi)

    def on_train_end(self, logs=None):
        args = (logs,)
        self._run_all_callbacks(on_train_end, args, self.train_end_multi)

    def on_test_begin(self, logs=None):
        args = (logs,)
        self._run_all_callbacks(on_test_begin, args, self.test_begin_multi)

    def on_test_end(self, logs=None):
        args = (logs,)
        self._run_all_callbacks(on_test_end, args, self.test_end_multi)

    def on_step_begin(self, step, logs=None):
        args = (step, logs,)
        self._run_all_callbacks(on_step_begin, args, self.step_begin_multi)

    def on_step_end(self, step, logs=None):
        args = (step, logs,)
        self._run_all_callbacks(on_step_end, args, self.step_end_multi)

    def on_episode_begin(self, step, episode, logs=None):
        args = (step, episode, logs,)
        self._run_all_callbacks(on_episode_begin, args, self.episode_begin_multi)

    def on_episode_end(self, step, episode, logs=None):
        args = (step, episode, logs,)
        self._run_all_callbacks(on_episode_end, args, self.episode_end_multi)
