import gym
from pathlib import Path

import torch


class ExperimentConfiguration(object):
    """The configuration for an experiment.

    Attributes:
        env_name: The name of the gym environment to run the experiment in.
        seed: The seed to use for initializing the environment and experiment.
        steps: The number of steps to perform in the environment.
        no_cuda: Whether or not to use CUDA tensors for GPU computations.
            Used by the soft decision tree and agent implementation.
        save_directory: The path to the directory where to store results.
        tensorboard: Whether or not to log to tensorboard.
        make_environment: The function to use for making the gym environment.
        #
        input_shape: The expected shape of the observations received from an environment.
        actions: The expected number of actions that the model can output to an environment.
    """
    def __init__(self, env_name, seed, steps, no_cuda, save_directory, tensorboard=False,
                 make_environment=lambda env: gym.make(env)):
        self.env_name = env_name
        self.seed = seed
        self.steps = steps
        self.no_cuda = no_cuda
        self.device = torch.device('cpu') if no_cuda else torch.device('cuda')
        self.save_directory = Path(save_directory) / f"{seed}"
        self.base_save_directory = save_directory
        self.tensorboard = tensorboard
        self.make_environment = make_environment
        #
        self.input_shape = None
        self.actions = None


class ModelConfiguration(object):
    """The configuration for a model.

    An instance of this class contains all necessary configuration parameters
    to build a keras neural network, used as the model of an agent.

    Attributes:
        create_function: A function (input_shape, output_shape) -> model,
            which creates a torch model given an input and output shape.
        optimizer: The optimizer to use for training the model.
        learning_rate: The learning rate for the optimizer.
        loss: The loss function to use for training the model.
    """
    def __init__(self, create_function, optimizer, learning_rate, loss):
        self.create_function = create_function
        self.optimizer = optimizer
        self.learning_rate = learning_rate
        self.loss = loss


class AgentConfiguration(object):
    """The configuration for a DQN agent.

    Attributes:
        warm_up_steps: The number of steps during training to use a random
            uniform policy for selecting actions.
        memory_size: The size of the experience replay memory.
        batch_size: The batch_size to use for training the agent.
        policy: The policy to use for action selection.
        test_policy: The policy to use if testing the agent.
        gamma: The discount factor for the future rewards.
        enable_double_dqn: Use Double DQN?
        target_update: Number of steps between target model updates.
        target_tau: Integer >= 1 for hard updates, and 0 < float < 1 for soft updates.
        save_interval: The number of steps between saves of the DQN agent.
        prioritised_replay: (Optional) Whether or not to use prioritised experience replay.
        combined_replay: (Optional) Whether or not to use combined experience replay.
        sdt_replay: (Optional) Whether or not the agent should use a SDT-managed replay memory.
        random_sampling: (Optional) Perform random sampling, instead of getting spaced out experiences.
        prioritised_sampling: (Optional) Whether or not to use prioritised sampling with SDT replay.
    """
    def __init__(self, warm_up_steps, memory_size, batch_size,
                 policy, test_policy, gamma,
                 enable_double_dqn, target_update, target_tau,
                 save_interval,
                 prioritised_replay=False, combined_replay=False,
                 sdt_replay=False, random_sampling=True, prioritised_sampling=False):
        self.warm_up_steps = warm_up_steps
        self.memory_size = memory_size
        self.batch_size = batch_size
        self.policy = policy
        self.test_policy = test_policy
        self.gamma = gamma
        self.enable_double_dqn = enable_double_dqn
        self.target_update = target_update
        self.target_tau = target_tau
        self.save_interval = save_interval
        self.combined_replay = combined_replay
        self.prioritised_replay = prioritised_replay
        self.sdt_replay = sdt_replay
        self.random_sampling = random_sampling
        self.prioritised_sampling = prioritised_sampling


class TreeConfiguration(object):
    """The configuration for a soft decision tree.

    Attributes:
        depth: The depth of the tree.
        lmbda: The lambda indicating the penalty strength.
        batch_size: The batch_size to use for training the model.
        name: The name for this configuration of a SDT.
        #
        save_directory: The path to the directory where to store results.
    """
    def __init__(self, depth, lmbda, batch_size, name):
        self.depth = depth
        self.lmbda = lmbda
        self.batch_size = batch_size
        self.name = name
        #
        self.save_directory = None
