class EnvironmentDescription(object):
    """A description of a gym environment, for interpreting observations and actions.

    Attributes:
        name: The name of the environment for which to provide a description.
        observations: The observation index, along with the name of the observation value at that index.
        actions: The actions, along with their descriptions.
        observation_dimensions: The dimensions of the observations.
    """

    def __init__(self):
        # Name
        self.name = None
        # Observations
        self.observations = {}
        self.observation_dimensions = {}
        self.observations_width = 0
        self.observations_height = 0
        # Actions
        self.actions = {}

    def num_observations(self):
        return len(self.observations)

    def num_actions(self):
        return len(self.actions)


class CartPoleEnvDescription(EnvironmentDescription):
    """The description of the CartPole environment."""
    def __init__(self):
        # Super call
        super(CartPoleEnvDescription, self).__init__()
        # Name
        self.name = "CartPole"
        # Observations
        self.observations = {
            0: "Cart Position",
            1: "Cart Velocity",
            2: "Pole Angle",
            3: "Pole Velocity At Tip",
        }
        self.observation_dimensions = {
            0: 4,
        }
        self.observations_width = 4
        self.observations_height = 1
        # Actions
        self.actions = {
            0: "Left",
            1: "Right",
        }


class LunarLanderEnvDescription(EnvironmentDescription):
    """The description of the LunarLander environment."""
    def __init__(self):
        # Super call
        super(LunarLanderEnvDescription, self).__init__()
        # Name
        self.name = "LunarLander"
        # Observations
        self.observations = {
            0: "x",
            1: "y",
            2: "Velocity x",
            3: "Velocity y",
            4: "Angle",
            5: "Angular Velocity",
            6: "Ground Contact Left Leg",
            7: "Ground Contact Right Leg",
        }
        self.observation_dimensions = {
            0: 8,
        }
        self.observations_width = 8
        self.observations_height = 1
        # Actions
        self.actions = {
            0: "No-op",
            1: "Left Orientation",
            2: "Main Engine",
            3: "Right Orientation",
        }


class MarioEnvDescription(EnvironmentDescription):
    """The description of the Mario environment."""
    def __init__(self):
        # Super call
        super(MarioEnvDescription, self).__init__()
        # Name
        self.name = "Mario"
        # Observations
        self.observations = {}
        # The dimensions of the observations
        self.observation_dimensions = {}
        # The 10 x 10 grid around Mario
        for cell in range(100):
            self.observations[cell] = f"Cell {cell}"
            if cell % 10 == 0:
                self.observation_dimensions[int(cell / 10)] = 10
        # The last two values:
        self.observations[100] = "On Ground"
        self.observations[101] = "Can Jump"
        self.observation_dimensions[10] = 2
        self.observations_width = 10
        self.observations_height = 11
        # Actions
        self.actions = {
            0: "No-op",
            1: "Left",
            2: "Right",
            3: "Down",
            4: "Jump",
            5: "Sprint",
        }


def get_env_description(env_name):
    if "CartPole" in env_name:
        return CartPoleEnvDescription()
    elif "LunarLander" in env_name:
        return LunarLanderEnvDescription()
    elif "Mario" in env_name:
        return MarioEnvDescription()
    else:
        raise ValueError(f"No description for {env_name} environment")
