import os
from typing import List

import numpy as np
from gym import Env

from core.callbacks.callback import CallbackGroup
from core.callbacks.loggers.action_logger import ActionLogger
from core.callbacks.loggers.dqn_logger import DQNStepLog, DQNEpisodeLog, SDTMemoryLog, MemoryLog
from core.callbacks.sdt.sdt_callback import SDTCallbackTemplate
from core.experiment.config import *
from core.experiment.environment_description import get_env_description
from core.experiment.experiment_logger import ExperimentLogger
from core.experiment.experiment_setup import ExperimentSetup
from core.models.dqn.dqn_agent import DQNAgent
from core.models.dqn.experience_replay import Experience, SDTReplayMemory
from core.paths.sdt_path import SDTPaths
from visualization.sdt.leaf_plotter import LeafPlotter


class Experiment(object):
    """The class for representing experiments.

    This class provides an abstraction for automatically creating a DQN agent
    with the given experiment-, agent- and model-configuration.

    Attributes:
        experiment_configuration: An ExperimentConfiguration, describing
            the parameters of the experiment and environment.
        agent_configuration: An AgentConfiguration for creating the agent.
        model_configuration: A ModelConfiguration for creating the underlying
            neural network of the agent.
        tree_configurations: A list of TreeConfigurations for creating the tree.
        sdt_callback_templates: The callback templates for differently trained SDTs.
    """

    def __init__(self, experiment_configuration: ExperimentConfiguration,
                 agent_configuration: AgentConfiguration,
                 model_configuration: ModelConfiguration,
                 tree_configurations: List[TreeConfiguration],
                 sdt_callback_templates: List[SDTCallbackTemplate],
                 sdt_threads=4):
        # Store the configurations
        self.exp_conf = experiment_configuration
        self.agent_conf = agent_configuration
        self.model_conf = model_configuration
        self.tree_configs = tree_configurations
        self.sdt_templates = sdt_callback_templates
        self.threads = sdt_threads
        self.setup = ExperimentSetup(self.exp_conf, self.agent_conf, self.model_conf,
                                     self.tree_configs, self.sdt_templates)

    #########
    # Train #
    #########
    def train(self, train_agent_only=False, agent_save_interval=None, min_successful_reward=None,
              deterministic=True,
              visualize=False, verbose=2):
        """Creates a training experiment for the given configurations.

        Use this method to train a DQN agent and SDTs.

        Args:
            train_agent_only: Whether or not to train the DQN agent only, without any
                Soft Decision Trees.
            agent_save_interval: (Optional) The number of steps between saves of the agent.
            min_successful_reward: (Optional) If training SDTs too, whether or not to
                only consider successful episodes.
            deterministic: Whether or not the SDT policy should be deterministic.
            visualize: Indicate if the environment should be visualized.
            verbose: Set the amount of information logged by the agent.

        Returns:
            None.
        """
        # Create the results directory for this experiment, based on the seed
        gym.logger.info(f"Setting up TRAIN experiment...")
        os.makedirs(self.exp_conf.save_directory, exist_ok=True)
        training_agent = True
        training_sdt = not train_agent_only
        use_best = False
        if agent_save_interval is None:
            agent_save_interval = self.agent_conf.save_interval
        # Create and initialise the environment
        with self.setup.make_environment() as env:
            # Create the DQN agent
            agent = self.setup.create_agent(training_agent)
            # Setup all callbacks (DQN, memory, loggers and SDT callbacks)
            all_callbacks, _, agent_logger, sdt_group = \
                self.setup.setup_all_callbacks(agent, training_agent, agent_save_interval, train_agent_only,
                                               min_successful_reward, use_best, verbose, self.threads)
            # Setup the replay memory for if training SDTs
            if training_sdt and agent.agent_config.sdt_replay:
                config = agent.agent_config
                agent_logger.memory_log_entry = SDTMemoryLog(sdt_group.callbacks[0].soft_decision_tree.num_leafs,
                                                             get_env_description(self.exp_conf.env_name))
                if config.prioritised_sampling:
                    # Anneal beta to 1 over the entire training process
                    beta = 0.4
                    beta_growth = (1 - beta) / (self.exp_conf.steps - self.agent_conf.warm_up_steps)
                    agent.memory = SDTReplayMemory(config.memory_size, config.batch_size, sdt_group.callbacks[0],
                                                   deterministic, config.random_sampling,
                                                   config.prioritised_sampling, beta_growth=beta_growth)
                else:
                    agent.memory = SDTReplayMemory(config.memory_size, config.batch_size, sdt_group.callbacks[0],
                                                   deterministic, config.random_sampling,
                                                   config.prioritised_sampling)
            else:
                agent_logger.memory_log_entry = MemoryLog(get_env_description(self.exp_conf.env_name))
            # Train the agent
            self._run_agent(env, agent, training_agent, training_sdt, self.exp_conf.steps, all_callbacks,
                            visualize=visualize, verbose=verbose)

    def train_on_final_agent(self, episodes, step_save=None, use_best=False, min_successful_reward=None,
                             visualize=False, verbose=2):
        """Creates a training experiment for the given configurations.

        Method to train soft decision trees on the policy of an already trained agent.

        Args:
            episodes: The number of episodes to test the agent for.
            step_save: The final step save of the agent to use.
            use_best: Whether or not to use the best saved model instead of a step saved one.
            min_successful_reward: The minimum reward per episode expected to be
                considered a successful episode to train on.
            visualize: Indicate if the environment should be visualized.
            verbose: Set the amount of information logged by the agent.

        Returns:
            None.
        """
        gym.logger.info(f"Setting up TRAIN ON FINAL AGENT experiment...")
        training_agent = False
        training_sdt = True
        train_agent_only = False
        agent_save_interval = np.inf
        # Create and initialise the environment
        with self.setup.make_environment() as env:
            # Load the DQN agent
            agent = self.setup.load_agent(step_save=step_save, use_best=use_best, training=training_agent)
            print(f"Agent test policy: {agent.test_policy}")
            # Setup all callbacks (DQN, memory, loggers and SDT callbacks)
            all_callbacks, memory_callback, agent_logger, _ = \
                self.setup.setup_all_callbacks(agent, training_agent, agent_save_interval, train_agent_only,
                                               min_successful_reward, use_best, verbose, self.threads)
            # Test the agent and train the SDTs
            self._run_agent(env, agent, training_agent, training_sdt, episodes, all_callbacks,
                            memory_callback, agent_logger, min_successful_reward,
                            visualize=visualize, verbose=verbose)

    ########
    # Test #
    ########
    def test_agent(self, episodes, step_save=None, use_best=False, log_results=False, visualize=True, verbose=2):
        """Creates a testing experiment for a DQN agent.

        Args:
            episodes: For how many episodes to run the experiment.
            step_save: The step number of the save.
            use_best: Whether or not to use the best saved model instead of a step saved one.
            log_results: Whether or not to the log results.
            visualize: Indicate if the environment should be visualized.
            verbose: Set the amount of information logged by the agent.

        Returns:
            None.
        """
        gym.logger.info(f"Setting up TEST AGENT experiment...")
        training_agent = False
        training_sdt = False
        # Create and initialise the environment
        with self.setup.make_environment() as env:
            # Load DQN agent architecture and weights
            agent = self.setup.load_agent(step_save=step_save, use_best=use_best, training=training_agent)
            print(f"Agent test policy: {agent.test_policy}")
            # Test the agent
            if log_results:
                # Setup all callbacks (DQN, memory, loggers and SDT callbacks)
                all_callbacks, memory_callback, agent_logger, _ = \
                    self.setup.setup_all_callbacks(agent, training_agent, np.inf, True, None, use_best,
                                                   verbose, self.threads)
                self._run_agent(env, agent, training_agent, training_sdt, episodes, all_callbacks,
                                visualize=visualize, verbose=verbose)
            else:
                self._run_agent(env, agent, training_agent, training_sdt, episodes,
                                visualize=visualize, verbose=verbose)

    #################################
    # Run Agent-centered Experiment #
    #################################
    def _run_agent(self, env: Env, agent: DQNAgent, training_agent, training_sdt, steps_or_episodes, callbacks=None,
                   memory_callback=None, agent_logger=None, min_successful_reward=None, visualize=True, verbose=2):
        # Gather the arguments
        num_steps = steps_or_episodes if training_agent else None
        num_episodes = None if training_agent else steps_or_episodes
        # agent.training = training_agent
        training_sdt_on_successful_episodes = training_sdt and min_successful_reward is not None

        callback_group = CallbackGroup(callbacks=callbacks if callbacks is not None else [])
        policy = agent.policy if training_agent else agent.test_policy
        warm_up_steps = self.agent_conf.warm_up_steps if training_agent else None
        # The logger for the experiment
        logger = ExperimentLogger(training_agent, steps_or_episodes, policy, verbose, time_per_steps=1000, time_avg=20,
                                  progress=5, time_per_eps=10, warm_up_steps=warm_up_steps)

        dqn_step_log_entry = DQNStepLog()
        dqn_episode_log_entry = DQNEpisodeLog()

        # Keep running as long as requested
        step = 1
        episode = 1

        def keep_running():
            if training_agent:
                return step <= num_steps
            else:
                return episode <= num_episodes

        # Callback when training/testing begins
        if training_agent:
            callback_group.on_train_begin()
        else:
            callback_group.on_test_begin()

        while keep_running():
            # Start a new episode
            state = env.reset()
            episode_reward = 0.0
            episode_steps = 0
            done = False

            # Callback when a new episode begins
            callback_group.on_episode_begin(step, episode)

            # Repeat until the episode is over
            while not done and keep_running():
                # Visualize the environment if requested
                if visualize:
                    env.render()

                # Inform when stopping the random policy warm-up steps
                if training_agent:
                    logger.warm_up_info(step)

                # Callback at the start of a time step
                callback_group.on_step_begin(step)

                # Select an action
                action, q_values = self.setup.agent_predict(agent, state, step, episode)
                # Observe the new state and reward
                next_state, reward, done, _ = env.step(action)

                # Store the experience in memory
                experience = Experience(state, action, reward, done, next_state, q_values)
                if training_agent:
                    agent.store_experience(experience)
                # Train the agent if necessary
                agent_loss = agent.train(step) if training_agent else None

                # Gather the data for the step
                step_logs = dqn_step_log_entry.create_entry(step, episode, state, action, q_values, reward, agent_loss)

                # Callback at the end of a time step
                callback_group.on_step_end(step, step_logs)
                # An other step has passed
                step += 1
                state = next_state
                episode_steps += 1
                episode_reward += reward

                # Time estimation
                logger.time_info(step, episode)
                # Progress info
                logger.progress_info(step, episode, done, episode_reward, episode_steps)

            # Gather the data for the episode
            episode_logs = dqn_episode_log_entry.create_entry(step, episode, episode_reward, episode_steps)

            # If not training the agent and training SDTs on a trained agent with successful episodes only
            if training_sdt_on_successful_episodes and episode_reward < min_successful_reward:
                # Delete the episode from logs and memory
                agent_logger.delete_last_episode(episode_steps)
                memory_callback.on_episode_end(step, episode, episode_logs)
                # Revert the experiment back to before this episode
                step -= episode_steps
            # If Keeping the episode, continue
            else:
                # Callback when an episode ends
                callback_group.on_episode_end(step, episode, episode_logs)
                # An other episode has passed
                episode += 1

        # Callback when training/testing ends
        if training_agent:
            callback_group.on_train_end()
        else:
            callback_group.on_test_end()

        # Total time taken
        logger.total_time_info()

    #####################################################################################

    def test_sdt(self, episodes, step_save, tree_config, sdt_callback_name, deterministic=True, agent_policy=None,
                 plot_chosen_leafs=False, visualize=True):
        """Creates a testing experiment for a soft decision tree.

        Args:
            episodes: For how many episodes to run the experiment.
            step_save: The step number of the save.
            tree_config: The tree configuration.
            sdt_callback_name: The name of the callback used to train the soft decision tree.
            deterministic: Whether or not the SDT policy should be deterministic.
            agent_policy: An optional agent policy to use for selecting actions with for the SDT.
            plot_chosen_leafs: Whether or not to plot the chosen leaf and actions per episode while testing the policy.
            visualize: Indicate if the environment should be visualized.

        Returns:
            None.
        """
        gym.logger.info(f"Setting up TEST SDT experiment...")
        # Create and initialise the environment
        with self.setup.make_environment() as env:
            # Load the Soft Decision Tree
            model = self.setup.load_sdt(tree_config, sdt_callback_name, epoch_save=step_save)
            # Plot the leafs
            leaf_plotter = LeafPlotter(tree_config) if plot_chosen_leafs else None

            episode = 0
            while episode < episodes:
                done = False
                episode_reward = 0
                observation = env.reset()

                while not done:
                    if plot_chosen_leafs:
                        rgb_array = env.render(mode='rgb_array')
                    elif visualize:
                        env.render()
                    processed_observation = torch.from_numpy(np.array([observation]).astype(np.float32))
                    action, chosen_leaf, leaf_distribution = model.predict(processed_observation,
                                                                           deterministic=deterministic)
                    # Follow agent's policy if provided
                    if agent_policy is not None:
                        action = agent_policy.select_action(leaf_distribution.numpy())
                    # Plot the leafs
                    if plot_chosen_leafs:
                        leaf_plotter.add_entry(episode, rgb_array, action, chosen_leaf, leaf_distribution)
                    next_observation, reward, done, step = env.step(action)
                    episode_reward += reward
                    observation = next_observation
                print(f"Episode {episode}: reward = {episode_reward}")
                episode += 1

            if plot_chosen_leafs:
                leaf_plotter_save_path = SDTPaths.vis_specific_directory(self.exp_conf, tree_config, sdt_callback_name)
                leaf_plotter.plot_chosen_leafs(leaf_plotter_save_path, episodes=range(episodes), show_plot=False)

    def test_trees_with_agent(self, steps, agent_step_save, deterministic=True, visualize=False, verbose=2):
        """Creates a testing experiment for a DQN agent.

        Args:
            steps: For how many steps to run the experiment.
            agent_step_save: The step number of the save for the DQN agent.
            deterministic: Whether or not the SDT policy should be deterministic.
            visualize: Indicate if the environment should be visualized.
            verbose: Set the amount of information logged in the terminal.

        Returns:
            None.
        """
        gym.logger.info(f"Setting up TEST SDTs experiment...")
        training_agent = False
        train_agent_only = False
        agent_save_interval = np.inf
        min_successful_reward = None
        use_best = False

        # Create and initialise the environment
        with self.setup.make_environment() as env:
            # Load the DQN agent
            agent = self.setup.load_agent(step_save=agent_step_save)
            agent.training = training_agent
            print(f"Agent test policy: {agent.test_policy}")
            # Setup all callbacks (DQN, memory, loggers and SDT callbacks)
            all_callbacks, memory_callback, agent_logger, sdt_group = \
                self.setup.setup_all_callbacks(agent, training_agent, agent_save_interval, train_agent_only,
                                               min_successful_reward, use_best, verbose, self.threads)
            # Load all the soft decision trees on their last save
            sdt_callbacks = []
            for callback in sdt_group.callbacks:
                callback.soft_decision_tree = self.setup.load_sdt(callback.tree_conf, callback.name)
                sdt_callbacks.append(callback)

            env_descr = get_env_description(self.exp_conf.env_name)
            actions_logger = ActionLogger(self.exp_conf, env_descr, sdt_callbacks, deterministic_trees=deterministic)

            step = 0
            episode = 0
            while step < steps:
                done = False
                episode_reward = 0.0
                observation = env.reset()

                while not done and step < steps:
                    if visualize:
                        env.render()

                    # Let the agent select an action
                    agent_action, q_values = self.setup.agent_predict(agent, observation, step, episode)
                    # Compare agent prediction with that of the Soft Decision Trees
                    actions_logger.compare_actions(step, episode, observation, agent_action, q_values)

                    # Perform the agent's action
                    next_observation, reward, done, _ = env.step(agent_action)
                    episode_reward += reward
                    observation = next_observation
                    step += 1

                print(f"Episode {episode}: reward = {episode_reward}, step: {step}")
                episode += 1

            # Save the results
            actions_logger.save_logs()
            actions_logger.finish_logs()
