import time
from collections import deque

import gym

from core.models.dqn.policy import RandomPolicy


class ExperimentLogger(object):
    """An abstraction for the logging functions in experiments.

    Attributes:
        training: Is the experiment training something or evaluating?
        steps_or_episodes: The number of steps/episodes for the training/evaluating process.
        policy: The policy used by the agent.
        verbose: How verbose the logger should be.
        time_per_steps: The interval size of the time calculations.
        time_avg: The window size of the running average.
        progress: If verbose == 0, the percentages of the learning process to log.
        time_per_eps: If verbose == 1, the number of episodes between logging the average time.
        warm_up_steps: (Optional) The warm-up steps if an agent is trained.
    """
    def __init__(self, training, steps_or_episodes, policy, verbose, time_per_steps=1000, time_avg=10, progress=5,
                 time_per_eps=10, warm_up_steps=None):
        assert 0 <= verbose <= 2 and int(verbose) == verbose, f"Verbose must be one of [0, 1, 2]. Given: {verbose}."
        # Keep track of how verbose the logger should be
        self.verbose = verbose
        # Store the arguments
        self.training = training
        self.num_steps = steps_or_episodes if training else None
        self.num_episodes = None if training else steps_or_episodes
        self.warm_up_steps = warm_up_steps
        self.policy = policy
        self._random_policy = str(RandomPolicy())

        self.steps_len = len(str(self.num_steps))
        self.episodes_len = len(str(self.num_episodes))
        self.time_start = time.time()
        self.time_before = self.time_start

        self.time_per_steps = time_per_steps
        self.time_avg = time_avg
        self.progress = progress
        self.time_per_eps = time_per_eps
        self._times = deque(maxlen=self.time_avg)
        # Some constants for calculating/printing time
        self._hour_seconds = 3600
        self._minute_seconds = 60
        self._hour = "hour(s)"
        self._minute = "minute(s)"
        self._seconds = "second(s)"

    def get_time(self, seconds):
        """Return a time  and metric in seconds, minutes or hours depending on the number of seconds provided."""
        if seconds >= self._hour_seconds:
            metric = self._hour
            duration = seconds / self._hour_seconds
        elif seconds >= self._minute_seconds:
            metric = self._minute
            duration = seconds / self._minute_seconds
        else:
            metric = self._seconds
            duration = seconds
        # Return the time and metric
        return f"{round(duration, 5)} {metric}"

    def time_info(self, step, episode):
        """Inform the time per X steps and the estimated time remaining."""
        if self.training and step % self.time_per_steps == 0 and step > 0:
            # Calculate how much time it took
            time_after = time.time()
            diff = time_after - self.time_before
            steps_left = self.num_steps - step
            # Only count after warm-up steps (as the agent is not training yet)
            if step > self.warm_up_steps:
                self._times.append(diff)
            self.time_before = time_after
            # Estimate time left
            if step <= self.warm_up_steps:
                estimation = ""
            else:
                avg_time = sum(self._times) / len(self._times)
                time_left = steps_left / self.time_per_steps * avg_time
                estimation = f"Estimated time remaining: {self.get_time(time_left)}"
            # Log the time
            if self.verbose == 2 or (self.verbose == 1 and episode % self.time_per_eps == 0):
                gym.logger.info(f"{self.time_per_steps} steps took {self.get_time(diff)} seconds. {estimation}")

    def warm_up_info(self, step):
        """Inform when stopping the random policy warm-up steps."""
        # The last step was the last warm-up step
        if (step - 1) == self.warm_up_steps:
            gym.logger.info(f"[Step {step}] Stopping random policy warm-up steps. Starting training...")

    def progress_info(self, step, episode, done, episode_reward, episode_steps):
        """Inform of the progress"""
        if self.training:
            # Calculate the percentage done
            percent_done = step / self.num_steps * 100
            # Verbose 0: logging every progress %
            if self.verbose == 0 and percent_done % self.progress == 0:
                gym.logger.info(f"{step: {self.steps_len}d} / {self.num_steps} ({round(percent_done): 3d}%%)")
            # Verbose 1 and 2: log every episode
            elif done:
                progress = f"{step: {self.steps_len}d} / {self.num_steps} ({round(percent_done): 3d}%%)"
                reward_str = f"{episode_reward: 6f}"
                steps_str = f"{episode_steps: 4d}"
                policy_str = str(self.policy) if self.training and step > self.warm_up_steps else self._random_policy
                gym.logger.info(f"[{progress}] Episode {episode} | "
                                f"reward: {reward_str}, steps: {steps_str}, {policy_str}")
        elif done:
            # Log every episode when evaluating
            gym.logger.info(f"Episode {episode: {self.episodes_len}d} / {self.num_episodes} | "
                            f"reward: {episode_reward: 6f}, steps: {episode_steps: 4d}")

    def total_time_info(self):
        """Inform the total time taken to run the experiment."""
        total_time = time.time() - self.time_start
        gym.logger.info(f"{'Training' if self.training else 'Testing'} experiment took {self.get_time(total_time)}.")
