import os
import random
from typing import List

import numpy as np

from core.callbacks.dqn.agent_callback import SaveAgentCallback
from core.callbacks.dqn.memory_callback import EpisodeMemoryCallback
from core.callbacks.loggers.dqn_logger import DQNLogger
from core.callbacks.sdt.episode_callback import EpisodeCallbackTemplate
from core.callbacks.sdt.sdt_callback import SDTCallbackTemplate, SDTCallbackGroup
from core.experiment.config import *
from core.models.dqn.dqn_agent import DQNAgent
from core.models.dqn.policy import Policy
from core.models.soft_decision_tree import SoftDecisionTree
from core.paths.dqn_path import DQNPaths
from core.paths.sdt_path import SDTPaths
from run_config import CONTINUE_ON_EXISTING_DIRS


class ExperimentSetup(object):
    """The class containing different methods to setup an experiment.

    Attributes:
        experiment_configuration: An ExperimentConfiguration, describing
            the parameters of the experiment and environment.
        agent_configuration: An AgentConfiguration for creating the agent.
        model_configuration: A ModelConfiguration for creating the underlying
            neural network of the agent.
        tree_configurations: A list of TreeConfigurations for creating the tree.
        sdt_callback_templates: The callback templates for differently trained SDTs.
    """

    def __init__(self, experiment_configuration: ExperimentConfiguration,
                 agent_configuration: AgentConfiguration,
                 model_configuration: ModelConfiguration,
                 tree_configurations: List[TreeConfiguration],
                 sdt_callback_templates: List[SDTCallbackTemplate]):
        # Store the configurations
        self.exp_conf = experiment_configuration
        self.agent_conf = agent_configuration
        self.model_conf = model_configuration
        self.tree_configs = tree_configurations
        self.sdt_templates = sdt_callback_templates

    ###############
    # Environment #
    ###############
    def _init_environment(self, env):
        """Initialize the environment."""
        # Seed the environment and numpy
        random.seed(self.exp_conf.seed)
        np.random.seed(self.exp_conf.seed)
        env.seed(self.exp_conf.seed)
        torch.manual_seed(self.exp_conf.seed)
        # Get the observations and action shape of the environment
        observation_shape = env.observation_space.shape
        number_of_actions = env.action_space.n
        self.exp_conf.input_shape = observation_shape
        self.exp_conf.actions = number_of_actions
        gym.logger.info(f"Created environment {self.exp_conf.env_name}:\n\t"
                        f"observation_shape = {observation_shape}\n\t"
                        f"number_of_actions = {number_of_actions}")

    def make_environment(self):
        """Create the environment and initialise it."""
        env = self.exp_conf.make_environment(self.exp_conf.env_name)
        self._init_environment(env)
        return env

    @staticmethod
    def agent_predict(agent: DQNAgent, observation, step, episode):
        """Get an action from a DQN prediction, along with the q_values."""
        return agent.select_action(observation, step, episode, return_q_values=True)

    @staticmethod
    def sdt_predict(tree: SoftDecisionTree, observation, deterministic):
        """Get an action from a SDT prediction, along with the chosen leaf and leaf distribution."""
        processed_observation = torch.from_numpy(np.array([observation]).astype(np.float32))
        sdt_action, chosen_leaf, leaf_distribution = tree.predict(processed_observation, deterministic)
        leaf_distribution = leaf_distribution.numpy()[0]
        return sdt_action, chosen_leaf, leaf_distribution

    #############
    # DQN Agent #
    #############
    def create_agent(self, training=True):
        """Create a new DQN agent."""
        gym.logger.info(f"Creating new DQN agent...")
        agent = DQNAgent(self.exp_conf, self.agent_conf, self.model_conf, training=training)
        # The save directory for the agent
        agent_directory = DQNPaths.directory(self.exp_conf)
        try:
            os.makedirs(agent_directory, exist_ok=CONTINUE_ON_EXISTING_DIRS)
        except FileExistsError:
            gym.logger.error(f"A DQN directory for the given experiment already exists at "
                             f"'{self.exp_conf.save_directory}'.\nExperiment exiting due to "
                             f"'CONTINUE_ON_EXISTING_DIRS' = {CONTINUE_ON_EXISTING_DIRS}"
                             f" to protect experiment results being overwritten.")
            exit()
        # Return the agent
        return agent

    def load_agent(self, step_save=None, use_best=False, training=False):
        """Load DQN agent architecture and weights."""
        if use_best:
            path = DQNPaths.save_best_path(self.exp_conf)
        else:
            path = DQNPaths.save_path(self.exp_conf, step_save=step_save)
        gym.logger.info(f"Loading DQN at '{path}'...")
        agent = DQNAgent(self.exp_conf, self.agent_conf, self.model_conf, training=training)
        agent.load(path)
        agent.model.eval()
        return agent

    @staticmethod
    def create_agent_save_callback(agent: DQNAgent, agent_save_interval):
        """Create a save callback for the DQN agent."""
        save_callback = SaveAgentCallback(agent, agent_save_interval)
        return save_callback

    ##############
    # DQN Logger #
    ##############
    def create_dqn_logger(self, agent: DQNAgent, training_agent, best_save):
        """Create a logger callback for the DQN agent."""
        agent_logger = DQNLogger(self.exp_conf, agent, training_agent=training_agent, best_save=best_save)
        return agent_logger

    ###################
    # Memory Callback #
    ###################
    def create_memory_callback(self, agent: DQNAgent, min_successful_reward, verbose):
        """Create a memory callback for gathering an agent's experiences."""
        # Only keep as many episodes as the episode SDT callbacks need (but don't delete too often)
        reset_interval = 50
        for template in self.sdt_templates:
            if isinstance(template, EpisodeCallbackTemplate):
                reset_interval = max(reset_interval, template.episodes)
        memory_callback = EpisodeMemoryCallback(agent, reset_interval, min_successful_reward, verbose)
        return memory_callback

    #############
    # Callbacks #
    #############
    def create_sdt_callback_group(self, memory: EpisodeMemoryCallback, agent_policy: Policy):
        """Create a soft decision tree for each SDT callback."""
        gym.logger.info(f"Setting up SDT callbacks...")
        all_sdt_callbacks = []
        # For each tree configuration
        for tree_conf in self.tree_configs:
            # For each SDT callback, create a callback from the template
            for template in self.sdt_templates:
                callback = template.create_callback(self.exp_conf, tree_conf, memory, agent_policy)
                all_sdt_callbacks.append(callback)
        gym.logger.info(f"Created {len(all_sdt_callbacks)} SDT callbacks for:\n\t"
                        f"{len(self.tree_configs)} TreeConfigurations\n\t"
                        f"{len(self.sdt_templates)} different callback templates")
        return all_sdt_callbacks

    ###################
    # Callbacks Setup #
    ###################
    def setup_all_callbacks(self, agent: DQNAgent, training_agent, agent_save_interval, train_agent_only,
                            min_successful_reward, best_save, verbose, threads):
        """Sets up all the callbacks for an experiment."""
        all_callbacks = []
        # Create a save callback for the agent
        if training_agent:
            agent_save_callback = self.create_agent_save_callback(agent, agent_save_interval)
            all_callbacks.append(agent_save_callback)
        # Set up the logger for the DQN agent
        agent_logger = self.create_dqn_logger(agent, training_agent=training_agent, best_save=best_save)
        all_callbacks.append(agent_logger)
        # Soft Decision Trees are trained
        if not train_agent_only:
            # The memory for gathering DQN experiences for the SDTs
            memory_callback = self.create_memory_callback(agent, min_successful_reward, verbose)
            all_callbacks.append(memory_callback)
            # Set up the soft decision trees and their callbacks
            sdt_callbacks = self.create_sdt_callback_group(memory_callback,
                                                           agent.policy if training_agent else agent.test_policy)
            callback_group = SDTCallbackGroup(sdt_callbacks, threads)
            all_callbacks.append(callback_group)
        else:
            memory_callback = None
            callback_group = None
            gym.logger.info(f"Created no SDT callbacks, training only agent.")
        # Return all callbacks
        return all_callbacks, memory_callback, agent_logger, callback_group

    ######################
    # Soft Decision Tree #
    ######################
    def load_sdt(self, tree_config: TreeConfiguration, sdt_callback_name, epoch_save=None):
        """Load a Soft Decision Tree."""
        if epoch_save is None:
            path = SDTPaths.last_save_path(self.exp_conf, tree_config, sdt_callback_name)
        else:
            path = SDTPaths.save_path(self.exp_conf, tree_config, sdt_callback_name, epoch_save)
        model = SoftDecisionTree(experiment_configuration=self.exp_conf, tree_configuration=tree_config)
        gym.logger.info(f"Loading Soft decision tree at '{path}'...")
        model.load_state_dict(torch.load(path))
        model.eval()
        return model
