import numpy as np
from copy import deepcopy
from typing import List

import torch

from core.experiment.config import ExperimentConfiguration, AgentConfiguration, ModelConfiguration
from core.models.dqn.dqn_model import QNeuralNetwork
from core.models.dqn.experience_replay import Experience, ReplayMemory, CombinedReplayMemory, PrioritisedReplayMemory
from core.models.dqn.policy import RandomPolicy


class DQNAgent(object):
    """The DQN agent.

    This class provides an abstraction for automatically creating a DQN agent
    with the given experiment-, agent- and model-configuration.

    Attributes:
        experiment_config: An ExperimentConfiguration, describing
            the parameters of the experiment and environment.
        agent_config: An AgentConfiguration for creating the agent.
        model_config: A ModelConfiguration for creating the underlying
            neural network of the agent.

        memory: The ReplayMemory used by the agent.
        training: (Optional) A Boolean, indicating if the agent is training or not.
    """
    def __init__(self, experiment_config: ExperimentConfiguration,
                 agent_config: AgentConfiguration,
                 model_config: ModelConfiguration, training=True):
        # Store the configurations
        self.experiment_config = experiment_config
        self.agent_config = agent_config
        self.model_config = model_config
        # Store some parameters for easier access
        self.gamma = self.agent_config.gamma
        self.double_dqn = self.agent_config.enable_double_dqn
        self.policy = self.agent_config.policy
        self.test_policy = self.agent_config.test_policy
        # Create the neural network and the target network
        self.model = QNeuralNetwork(self.experiment_config, self.model_config)
        self.target_model = deepcopy(self.model) if self.double_dqn else None
        self.target_update = self.agent_config.target_update
        self.target_tau = self.agent_config.target_tau
        if self.double_dqn:
            self.target_model.train(mode=False)
        # Create the requested memory type
        if agent_config.combined_replay and agent_config.sdt_replay:
            raise ValueError(f"Both Combined Replay and SDT replay have been requested. Can only accept one.")
        else:
            # Non-SDT replay memories
            if not agent_config.sdt_replay:
                # CER
                if agent_config.combined_replay:
                    self.memory = CombinedReplayMemory(self.agent_config.memory_size, self.agent_config.batch_size)
                # Prioritised
                elif agent_config.prioritised_replay:
                    # Anneal beta to 1 over the entire training process
                    beta = 0.4
                    beta_growth = (1 - beta) / (experiment_config.steps - agent_config.warm_up_steps)
                    self.memory = PrioritisedReplayMemory(self.agent_config.memory_size, self.agent_config.batch_size,
                                                          beta=beta, beta_growth=beta_growth)
                # Default
                else:
                    self.memory = ReplayMemory(self.agent_config.memory_size, self.agent_config.batch_size)
            # SDT replay memory is initialized later
            else:
                self.memory = None
        # Are we training the agent?
        self.training = training
        self.model.train(mode=training)

    def predict_q_values(self, state):
        """Return the predicted q-values for a state.

        Args:
            state: a state to predict the q-values for.
        """
        with torch.no_grad():
            q_values = self.model.get_q_values(state, is_single_sample=True).data.numpy()
            return q_values

    def select_action(self, state, step, episode, return_q_values=False):
        """Select an action with the current policy.

        Args:
            state: The state observed.
            step: The current time step.
            episode: The current episode.
            return_q_values: (Optional) Whether or not to also return the q_values.

        Returns:
            The action chosen by the agent's policy. (Optional) The q-values returned by the agent.
        """
        # Get the q-values predicted for the state
        q_values = self.predict_q_values(state)
        # If training
        if self.training:
            # Use a random policy for the warm up steps
            if step < self.agent_config.warm_up_steps:
                action = RandomPolicy.select(q_values)
            # Use the policy
            else:
                action = self.policy.select_action(q_values, step, episode)
        # If testing
        else:
            action = self.test_policy.select_action(q_values, step, episode)
        # Return the action, along with the q_values if requested
        if return_q_values:
            return action, q_values
        else:
            return action

    def store_experience(self, experience: Experience):
        """Store an experience in the agent's replay memory."""
        self.memory.store_experience(experience, self)

    def train(self, step):
        """Train the DQN agent and store the latest experience in memory.

        Args:
            step: The current time step.

        Returns:
            loss: The training loss of the agent.
        """
        # Train on a batch of experiences if the number of warm up steps are done
        if step >= self.agent_config.warm_up_steps:
            # Get a batch of experiences
            if isinstance(self.memory, PrioritisedReplayMemory) or \
                    (self.agent_config.sdt_replay and self.agent_config.prioritised_sampling):
                experiences, indices, is_weights, leafs = self.memory.get_experience_batch()
                # Train the underlying Q model on the batch of experiences
                return self._train_on_batch(experiences, step, indices, is_weights, leafs)
            else:
                experiences = self.memory.get_experience_batch()
                # Train the underlying Q model on the batch of experiences
                return self._train_on_batch(experiences, step)

    def _train_on_batch(self, experiences: List[Experience], step, indices=None, is_weights=None, leafs=None):
        """Train the underlying Q network and update the target network."""
        # Preprocess the data
        samples = len(experiences)
        states = np.empty(shape=(samples, self.experiment_config.input_shape[0]))
        next_states = np.empty_like(states)
        actions = np.empty(shape=samples, dtype=np.int)
        rewards = np.empty(shape=samples)
        terminals = np.empty_like(rewards, np.bool)

        for idx, experience in enumerate(experiences):
            state, action, reward, terminate, next_state, _ = experience
            states[idx] = state
            next_states[idx] = next_state
            actions[idx] = action
            rewards[idx] = reward
            terminals[idx] = terminate
        # Get the predicted q-values for each state
        q_values = self.model.get_q_values(states)
        # The target q_value: reward + gamma * future_reward
        target_q_values = q_values.clone().data.numpy()
        # If using double Q-learning, predict with the target network
        if self.double_dqn:
            next_q_values = self.target_model.get_q_values(next_states).data.numpy()
        else:
            next_q_values = self.model.get_q_values(next_states).data.numpy()
        # The future reward is max Q(next_state) or 0 if next_state is terminal
        future_rewards = np.max(next_q_values, axis=1) * ~terminals
        target_q_values[np.arange(len(target_q_values)), actions] = rewards + self.gamma * future_rewards
        target_q_values = self.model.to_variable(target_q_values)

        # Doing prioritised ER
        prioritised_replay = indices is not None and is_weights is not None
        if prioritised_replay:
            # The TD-errors
            errors = torch.abs(q_values - target_q_values).sum(dim=1).data.numpy()
            # TD-error clipped between [-1, 1] for stability reasons (Schaul et al. 2016)
            errors = np.clip(errors, -1, 1)
            # Update the priorities
            if prioritised_replay and leafs is not None:
                self.memory.update_priorities(leafs, indices, errors)
            else:
                self.memory.update_priorities(indices, errors)

        # Reset gradients
        self.model.optimizer.zero_grad()
        # Compute the loss
        loss = self.model.loss(q_values, target_q_values)
        # Prioritised replay uses importance sampling weights
        if prioritised_replay:
            is_weights = torch.FloatTensor(is_weights)
            loss = (is_weights * loss).mean()
        # Optimize the model
        loss.backward()
        self.model.optimizer.step()

        # Double DQN
        if self.double_dqn:
            # Hard update every target_update steps
            if self.target_update >= 1 and step % self.target_update == 0:
                self.target_model.load_state_dict(self.model.state_dict())
            # Soft update every target_update steps
            elif step % self.target_update == 0:
                for target_param, param in zip(self.target_model.parameters(), self.model.parameters()):
                    # noinspection PyUnresolvedReferences
                    target_param.data.copy_(target_param.data * (1.0 - self.target_tau) + param.data * self.target_tau)

        # Return the loss
        return loss.item()

    def save(self, path):
        """Save the model at the given path."""
        with open(path, 'wb') as output_file:
            torch.save(self.model.state_dict(), output_file)

    def load(self, path):
        """Load the model from the given path."""
        self.model.load_state_dict(torch.load(path))
        if self.double_dqn:
            self.target_model.eval()
        self.model.eval()
