import torch
import torch.nn as nn
from torch.autograd import Variable

from core.experiment.config import ExperimentConfiguration, ModelConfiguration


class QNeuralNetwork(nn.Module):
    """A Q-learning neural network.

    This class provides an abstraction for automatically creating a Q-learning
    neural network with the given experiment-, agent- and model-configuration.

    Attributes:
        experiment_config: An ExperimentConfiguration, describing
            the parameters of the experiment and environment.
        model_config: A ModelConfiguration for creating the underlying
            neural network of the agent.
    """
    def __init__(self, experiment_config: ExperimentConfiguration,
                 model_config: ModelConfiguration):
        # Super call
        super(QNeuralNetwork, self).__init__()
        # Store the configurations
        self.experiment_config = experiment_config
        self.model_config = model_config
        self.observation_shape = self.experiment_config.input_shape[0]

        # Create the internal layers of the neural network
        self.sequential_model = self.model_config.create_function(
            input_shape=experiment_config.input_shape,
            output_shape=experiment_config.actions
        )
        self.to(device=self.experiment_config.device)
        # Set the optimizer and loss function
        self.optimizer = self.model_config.optimizer(self.parameters(), lr=self.model_config.learning_rate)
        self.loss = self.model_config.loss()

    def forward(self, x):
        """The forward pass through the model."""
        return self.sequential_model(x)

    def get_q_values(self, x, is_single_sample=False):
        """Get the q-values of a given input."""
        state = self.to_variable(x.reshape(-1, self.observation_shape))
        self.train(False)
        q_values = self(state)
        return q_values[0] if is_single_sample else q_values

    @staticmethod
    def to_variable(x):
        """Convert input to a torch tensor."""
        # noinspection PyArgumentList
        return Variable(torch.Tensor(x))
