import random
from collections import namedtuple

import gym
import numpy as np
import torch
from scipy.spatial import ConvexHull
from scipy.spatial.distance import cdist

from lib.per_master.prioritized_memory import Memory


##############
# Experience #
##############
# The named tuple representing an experience
Experience = namedtuple("Experience", "state, action, reward, done, next_state, q_values")


################################
# The Experience Replay Memory #
################################
class ReplayMemory(object):
    """The class representing an experience replay memory for an agent.

    Attributes:
        memory_size: The maximum number of experiences in the replay memory.
        batch_size: The batch_size used for sampling the replay memory.
    """
    def __init__(self, memory_size, batch_size):
        # Store the arguments
        self.memory_size = memory_size
        self.batch_size = batch_size
        # The replay memory is implemented as a circular list for faster sampling (faster than deque sampling)
        self._memory = [None] * memory_size
        self._index = 0
        self._size = 0

        self.last_counts = []

    def __len__(self):
        """Get the length of the internal memory."""
        return self._size

    def _count_experiences(self, experiences):
        if len(experiences) == 0:
            self.last_counts = []
        else:
            # Store the counts per action
            counts = [0 for _ in range(len(experiences[0].q_values))]
            for x in experiences:
                counts[x.action] += 1
            self.last_counts = counts

    def store_experience(self, experience, agent=None):
        """Store an experience in the memory."""
        # Replace the oldest memory in the circularly iterated list
        self._memory[self._index] = experience
        self._size = min(self._size + 1, self.memory_size)
        self._index = (self._index + 1) % self.memory_size

    def get_experience_batch(self):
        """Get a random batch of experiences from the memory."""
        # You can only get as many experiences as present in the memory, with a maximum of batch_size
        number_of_experiences = min(self._size, self.batch_size)
        # Get the random experiences from the memory
        if self._size < self.memory_size:
            random_experiences = random.sample(self._memory[:self._size], number_of_experiences)
        else:
            random_experiences = random.sample(self._memory, number_of_experiences)
        # Count the number of sampled experiences per action
        self._count_experiences(random_experiences)
        return random_experiences


class CombinedReplayMemory(ReplayMemory):
    """The class representing an experience replay memory for an agent.
        Based on Combined Experience Replay (CER).
    """
    def __init__(self, memory_size, batch_size):
        # Super Call
        super(CombinedReplayMemory, self).__init__(memory_size, batch_size - 1)
        # Store the last experience for efficient access
        self._last_experience = None

    def store_experience(self, experience, agent=None):
        """Store an experience in the memory."""
        # Super Call
        super(CombinedReplayMemory, self).store_experience(experience)
        # Store the last experience
        self._last_experience = experience

    def get_experience_batch(self):
        """Get a random batch of experiences from the memory."""
        # Super Call
        random_experiences = super(CombinedReplayMemory, self).get_experience_batch()
        # Add the last experience to the sampled experiences
        random_experiences.append(self._last_experience)
        # Count the number of sampled experiences per action
        self._count_experiences(random_experiences)
        return random_experiences


class PrioritisedReplayMemory(ReplayMemory):
    """The class representing a prioritised experience replay memory for an agent.

    This class serves as a wrapper for the underlying library: https://github.com/rlcode/per

    Attributes:
        memory_size: The maximum number of experiences in the replay memory.
        batch_size: The batch_size used for sampling the replay memory.
        e: A small positive constant that prevents the edge-case of transitions not being revisited
            once their error is zero.

    """
    # noinspection PyMissingConstructor
    def __init__(self, memory_size, batch_size, e=0.01, alpha=0.6, beta=0.4, beta_growth=0.001):
        # Store the arguments
        self.memory_size = memory_size
        self.batch_size = batch_size
        self.e = e
        self.alpha = alpha
        self.beta = beta
        self.beta_growth = beta_growth
        # The replay memory is implemented as a circular list for faster sampling (faster than deque sampling)
        self._memory = Memory(capacity=memory_size, e=e, a=alpha, beta=beta, beta_increment_per_sampling=beta_growth)
        self.last_counts = []

    def __len__(self):
        """Get the length of the internal memory."""
        return self._memory.tree.n_entries

    def store_experience(self, experience, agent=None):
        """Store an experience in the memory."""
        # Let the agent predict the target for the experience's state
        q_values = agent.model.get_q_values(experience.state, is_single_sample=True).data.numpy()
        old_action_prediction = q_values[experience.action]
        target_q_values = agent.target_model.get_q_values(experience.next_state, is_single_sample=True).data.numpy()
        # Calculate the future reward
        if experience.done:
            future_reward = experience.reward
        else:
            future_reward = experience.reward + agent.gamma * np.max(target_q_values)
        # Calculate the TD-error
        td_error = np.abs(old_action_prediction - future_reward)
        # TD-error clipped between [-1, 1] for stability reasons (Schaul et al. 2016)
        td_error = np.clip(td_error, -1, 1)
        # Add the experience to the underlying memory, prioritised based on its error
        self._memory.add(td_error, experience)

    def get_experience_batch(self):
        """Get a random batch of experiences from the memory."""
        # Sample a batch of experiences, based on their priority
        if self._memory.tree.n_entries == 0:
            self.last_counts = []
            return [], [], [], None
        else:
            mini_batch, indices, is_weights = self._memory.sample(self.batch_size)
            # Count the number of sampled experiences per action
            self._count_experiences(mini_batch)
            return mini_batch, indices, is_weights, None

    def update_priorities(self, indices, errors):
        """Update the priorities of the used experiences."""
        for i in range(self.batch_size):
            idx = indices[i]
            self.update_single_priority(idx, errors[i])

    def update_single_priority(self, index, error):
        self._memory.update(index, error)


class SDTReplayMemory(object):
    """The class representing an experience replay memory for an agent, using SDTs.

        Attributes:
            memory_size: The maximum total number of experiences in the replay memory.
            leaf_batch_size: The total batch_size used for sampling the leaf replay memories.
            sdt_callback: The SDT callback to implement the memory with.
            deterministic: Whether or not the SDTs are supposed to predict deterministically.
            random_sampling: Whether or not to perform random sampling, instead of getting spaced out experiences.
            prioritised_sampling: Whether or not to use prioritised sampling with SDT replay.
    """
    def __init__(self, memory_size, batch_size, sdt_callback, deterministic,
                 random_sampling,
                 prioritised_sampling, e=0.01, alpha=0.6, beta=0.4, beta_growth=0.001):
        # Store the arguments
        self.sdt_callback = sdt_callback
        self.memory_size = memory_size
        self.leaf_batch_size = batch_size
        self.deterministic = deterministic
        self.random_sampling = random_sampling
        self.prioritised_sampling = prioritised_sampling
        # The memories, per callback and per leaf
        self._memories = []
        self._tree = sdt_callback.soft_decision_tree
        self._num_actions = self.sdt_callback.soft_decision_tree.args.output_dim
        self.last_counts = []

        gym.logger.info(f"Using SDT experience replay...")
        # Sample equally from each leaf
        num_leafs = self._tree.num_leafs
        leaf_memory_size = int(memory_size / num_leafs)
        leaf_batch_size = int(batch_size / num_leafs)
        gym.logger.info(f"[{sdt_callback.full_name}] {num_leafs} leafs with batch size = {batch_size} "
                        f"=> {leaf_batch_size} samples per leaf")
        for leaf in range(num_leafs):
            if prioritised_sampling:
                leaf_memory = SDTPrioritisedLeafMemory(leaf_memory_size, leaf_batch_size, e, alpha, beta, beta_growth)
            else:
                leaf_memory = SDTLeafMemory(leaf_memory_size, leaf_batch_size, random_sampling)
            self._memories.append(leaf_memory)

    def __len__(self):
        """Get the length of the internal memory."""
        raise NotImplementedError

    def _reset_count_experiences(self):
        self.last_counts = [[0 for _ in range(self._num_actions)] for _ in range(self._tree.num_leafs)]

    def _count_experiences(self, leaf, experiences):
        if len(experiences) != 0:
            # Store the counts per action
            for x in experiences:
                self.last_counts[leaf][x.action] += 1

    def store_experience(self, experience, agent=None):
        """Store an experience in the memory."""
        # Let the tree process the experience
        processed_observation = torch.from_numpy(np.array([experience.state]).astype(np.float32))
        _, chosen_leaf, _ = self._tree.predict(processed_observation, deterministic=self.deterministic)
        # Store the experience in the chosen leaf
        self._memories[chosen_leaf.item()].store_experience(experience, agent)

    def get_experience_batch(self):
        """Get a random batch of experiences from the memory."""
        # From each tree & leaf, get the experiences
        self._reset_count_experiences()
        if self.random_sampling or not self.prioritised_sampling:
            experiences = []
            for leaf, leaf_memory in enumerate(self._memories):
                leaf_experiences = leaf_memory.get_experience_batch()
                self._count_experiences(leaf, leaf_experiences)
                experiences.extend(leaf_experiences)
            # Shuffle and return the experiences
            random.shuffle(experiences)
            return experiences
        # Prioritised sampling
        else:
            experiences = []
            for leaf, leaf_memory in enumerate(self._memories):
                leaf_mini_batch, leaf_indices, leaf_is_weights, _ = leaf_memory.get_experience_batch()
                self._count_experiences(leaf, leaf_mini_batch)
                batch = zip(leaf_mini_batch, leaf_indices, leaf_is_weights, [leaf] * len(leaf_indices))
                experiences.extend(batch)
            # Shuffle the data
            batches = []
            indices = []
            is_weights = []
            leafs = []
            random_indices = list(range(len(experiences)))
            random.shuffle(random_indices)
            for idx in random_indices:
                leaf_mini_batch, leaf_indices, leaf_is_weights, leaf = experiences[idx]
                batches.append(leaf_mini_batch)
                indices.append(leaf_indices)
                is_weights.append(leaf_is_weights)
                leafs.append(leaf)
            # Return the experiences
            return batches, indices, is_weights, leafs

    def update_priorities(self, leafs, indices, errors):
        """Update the priorities of the used experiences."""
        for i in range(min(self.leaf_batch_size, len(indices))):
            idx = indices[i]
            leaf = leafs[i]
            # noinspection PyUnresolvedReferences
            self._memories[leaf].update_single_priority(idx, errors[i])


class SDTLeafMemory(ReplayMemory):
    """A replay memory for a tree leaf

    Attributes:
        memory_size: The maximum number of experiences in the replay memory.
        batch_size: The batch_size used for sampling the replay memory.
        random_sampling: Whether or not to perform random sampling, instead of getting spaced out experiences.
    """
    def __init__(self, memory_size, batch_size, random_sampling):
        # Super call
        super(SDTLeafMemory, self).__init__(memory_size, batch_size)
        # Store the arguments
        self.random_sampling = random_sampling

    def _get_spread_experiences(self):
        # Get all the states as points in the n-dimensional state space
        size = len(self)
        points = np.array(list(map(lambda e: e.state, self._memory[:size])))
        # ConvexHull needs at least 5 points (library requirement)
        if len(self) >= 5:
            # TODO: masking here
            # TODO: use feature masking for larger states than cart pole
            # For all experiences in the memory, get the convex hull
            hull = ConvexHull(points)
            hull_points = points[hull.vertices, :]
            # Get the distances between all hull points and get the most distanced pair
            hull_distances = cdist(hull_points, hull_points, metric='euclidean')
            pair_idx = np.unravel_index(hull_distances.argmax(), hull_distances.shape)
            pair = [hull_points[pair_idx[0]], hull_points[pair_idx[1]]]
            # Get the most distant points
            return self._get_most_distant_points(pair, pair_idx, points)
        # Memory contains between 1 and 4 experiences
        else:
            # Only 1 experience in memory, or only 1 required
            if size == 1 or self.batch_size == 1:
                return [self._memory[0]] * self.batch_size
            # Only 2 experiences, 2 or more required
            elif size == 2:
                doubles = int(np.floor(self.batch_size / 2))
                experiences = self._memory[:size] * doubles
                if len(experiences) < self.batch_size:
                    experiences.append(random.choice(self._memory[:size], 1))
                return experiences
            # 3 - 4 experiences, 2 or more required
            else:
                # Calculate the distance between all points
                distances = cdist(points, points, metric='euclidean')
                # Get the points in order of distances
                pair_idx = np.unravel_index(distances.argmax(), distances.shape)
                pair = [points[pair_idx[0]], points[pair_idx[1]]]
                return self._get_most_distant_points(pair, pair_idx, points)

    def _get_most_distant_points(self, original_pair, original_pair_idx, points):
        # Add the most distant pair to the experiences to return
        experiences = [self._memory[original_pair_idx[0]], self._memory[original_pair_idx[1]]]
        experience_points = np.array(original_pair)
        # As long as we don't have a batch, keep searching for the most distant point w.r.t.
        #   the already selected ones
        while len(experiences) < self.batch_size:
            # Get the distance of all points to the already selected ones
            new_distances = cdist(points, experience_points, metric='euclidean')
            # Get the smallest distance of each point to the selected experiences
            #   => find closest neighbouring state
            minimum_to_experiences = np.min(new_distances, axis=1)
            # Get the next point which is the furthest away and add it
            new_point_idx = np.argmax(minimum_to_experiences)
            new_point = np.expand_dims(points[new_point_idx, :], 0)
            experience_points = np.append(experience_points, new_point, axis=0)
            # noinspection PyTypeChecker
            experiences.append(self._memory[new_point_idx])
        # Return the experiences
        return experiences

    def get_experience_batch(self):
        """Gather as many different experiences as possible."""
        size = len(self)
        # If there are no experiences, don't return any
        if size == 0:
            return []
        # Default random sampling
        elif self.random_sampling:
            # If not enough experiences are available, sample with replacement
            if size < self.batch_size:
                random_experiences = self._memory[:size]
                random_experiences.extend(random.choices(self._memory[:size], k=self.batch_size - size))
                return random_experiences
            # Else, sample without replacement
            else:
                return random.sample(self._memory if size == self.memory_size else self._memory[:size], self.batch_size)
        # Prioritised sampling
        else:
            return self._get_spread_experiences()


class SDTPrioritisedLeafMemory(PrioritisedReplayMemory):
    """A prioritised replay memory for a tree leaf"""
