import torch
import torch.nn.functional as F
from torch.distributions import categorical

import lib.soft_decision_tree.sdt.model as sdt
from core.experiment.config import ExperimentConfiguration, TreeConfiguration
from lib.soft_decision_tree.args import parser


class SoftDecisionTree(sdt.SoftDecisionTree):
    """The Soft Decision Tree.

    This class serves as a wrapper to provide parameter inputs to the
    underlying class as initialization arguments, instead of command-line
    arguments.

    Attributes:
        experiment_configuration: An ExperimentConfiguration, describing
            the parameters of the experiment and environment.
        tree_configuration: A TreeConfiguration for creating the tree.
    """
    def __init__(self, experiment_configuration: ExperimentConfiguration,
                 tree_configuration: TreeConfiguration):
        # Store the configurations
        self.experiment_configuration = experiment_configuration
        self.tree_configuration = tree_configuration
        # Extract the arguments for the tree and parse them
        args = [
            "--input-dim", f"{experiment_configuration.input_shape[0]}",
            "--output-dim", f"{experiment_configuration.actions}",
            "--depth", f"{tree_configuration.depth}",
            "--lmbda", f"{tree_configuration.lmbda}",
            "--batch-size", f"{tree_configuration.batch_size}",
            "--seed", f"{experiment_configuration.seed}",
            "--save", f"{tree_configuration.save_directory}",
            "--log-interval", "-1",
            ]
        # Train on GPU or CPU
        no_cuda = self.experiment_configuration.no_cuda
        cuda = not no_cuda and torch.cuda.is_available()
        if cuda:
            args.extend(["--no-cuda"])
        if experiment_configuration.tensorboard:
            args.extend(["--tensorboard"])
        # Parse the arguments
        args = parser.parse_args(args)
        args.device = torch.device('cuda') if cuda else torch.device('cpu')
        # Set the seed
        torch.manual_seed(args.seed)
        # Super call
        super(SoftDecisionTree, self).__init__(args)
        self.prev_epoch_loss = None

    def train_on_data_(self, train_data, current_epoch, sdt_name, callback_name,
                       full_batches, remainder, epoch, epochs):
        """Train the Soft Decision Tree with the given train data.

        Arguments:
            train_data: A list of input values and their corresponding target values.
            current_epoch: The current training epoch.
            sdt_name: The name given to the SDT.
            callback_name: The name of the callback requesting this training.
            full_batches: The number of full batches used in training.
            remainder: The remained of experiences besides full_batches.
            epoch: The training epoch in the current iteration.
            epochs: The number of epochs per iteration.

        Returns:
            total_loss, leafs_loss, inners_loss: the average losses for the training epoch.
        """
        self.train()
        # Keep track of the losses
        sum_total_loss = 0.0
        sum_leafs_loss = 0.0
        sum_inners_loss = 0.0
        # For each batch
        for batch_idx, (data, target) in enumerate(train_data):
            data, target = data.to(self.args.device), target.to(self.args.device)
            # Optimize
            self.optimizer.zero_grad()
            loss_total, loss_leafs, loss_inners = self._calc_loss(data, target)
            loss_total.backward()
            # Gradient norm clipping
            torch.nn.utils.clip_grad_norm_(self.parameters(), 40)
            self.optimizer.step()

            batch_number = batch_idx + 1
            size = len(data)
            if batch_idx % self.args.log_interval == 0 and self.args.log_interval > 0:
                print(f'[{sdt_name} {callback_name}] Train Epoch: {current_epoch} '
                      f'[{batch_number * size}/{size * full_batches + remainder}'
                      f' ({100. * batch_number / (full_batches + 1):.0f}%)]\tLoss: {loss_total.item():.6f}')
            self.step += 1
            # Add the new losses
            sum_total_loss += loss_total.item()
            sum_leafs_loss += loss_leafs.item()
            sum_inners_loss += loss_inners.item()
        # Get the average losses over the used batches
        num_experiences = full_batches + 1
        sum_total_loss /= num_experiences
        sum_leafs_loss /= num_experiences
        sum_inners_loss /= num_experiences

        # Only log at the end if this was requested
        # if self.args.log_interval == -1:
        #     print(f'[{sdt_name} {callback_name}] Train Epoch: {current_epoch} '
        #           f'[{full_batches * self.args.batch_size + remainder}/'
        #           f'{self.args.batch_size * full_batches + remainder}'
        #           f' (100%)]\tLoss: {sum_total_loss:.6f}')
        if self.args.log_interval == -1:
            if epoch == 0:
                self.prev_epoch_loss = (current_epoch, sum_total_loss)
            elif epoch == epochs - 1:
                print(f'[{sdt_name} {callback_name}] Train Epoch: {self.prev_epoch_loss[0]} -> {current_epoch} '
                      f'[{full_batches * self.args.batch_size + remainder}/'
                      f'{self.args.batch_size * full_batches + remainder}]\t'
                      f'Loss: {self.prev_epoch_loss[1]:.6f} -> {sum_total_loss:.6f}')
        # Return the losses
        return sum_total_loss, sum_leafs_loss, sum_inners_loss

    def predict(self, sample_x, deterministic=False):
        """Predict and output value given the input sample."""
        with torch.no_grad():
            leafs_pred = F.softmax(self.leafs, dim=1)
            probs_right = torch.sigmoid(self.betas * (sample_x.matmul(self.weights.t()) + self.biases))
            leaf_path_probs = self._calc_path_probs(probs_right, self.ancestors_leafs)
            if deterministic:
                chosen_leaf = torch.argmax(leaf_path_probs, dim=1)
                chosen_act = torch.argmax(leafs_pred[chosen_leaf]).item()
            else:
                chosen_leaf = categorical.Categorical(probs=leaf_path_probs[0]).sample().item()
                chosen_act = categorical.Categorical(probs=leafs_pred[chosen_leaf]).sample().item()
                # chosen_act = torch.distributions.categorical.Categorical(
                #     probs=leaf_path_probs.matmul(leafs_pred)[0]).sample().item()
            leaf_distribution = leafs_pred[chosen_leaf]
            return chosen_act, chosen_leaf, leaf_distribution

    def get_parameters(self):
        """Get the parameters of this Soft Decision Tree.

        Args:
            None.

        Returns:
            inners, betas, biases, leafs: The parameters for the current configuration of the tree.
        """
        with torch.no_grad():
            # Get the inner weights and their betas and bias
            inners = self.weights.detach().numpy()
            betas = self.betas.detach().tolist()
            biases = self.biases.detach().tolist()
            # List of leaf distributions
            leafs = [F.softmax(leaf.view(-1), dim=0).numpy() for leaf in self.leafs]
            # Return the parameters
            return inners, betas, biases, leafs

    def save_sdt(self, path):
        with open(path, 'wb') as output_file:
            torch.save(self.state_dict(), output_file)
