from core.experiment.environment_description import get_env_description
from core.experiment.experiment import Experiment
from core.models.dqn.policy import DecayingEpsilonGreedyPolicy, GreedyPolicy

from experiments.cart_pole import *
from experiments.templates import get_episode_callbacks
from visualization.dqn.dqn_visualization import DQNVisualizer


###########################
# ExperimentConfiguration #
###########################
from visualization.policy.policy_visualization import PolicyVisualizer
from visualization.sdt.sdt_visualization import SDTVisualizer
from visualization.sdt.tree_visualization import TreeVisualizer

experiment_configuration = ExperimentConfiguration(
    env_name=env_name,
    seed=1,
    steps=steps,
    no_cuda=no_cuda,
    save_directory=f"../../../../Data/Experiments/CartPole/epsilon_greedy/post/Default/"
)
# Whether or not SDTs are trained together with the DQN
train_agent_only = True
# Whether or not the DQN agent is training at the same time as the SDT
training_agent = False


######################
# AgentConfiguration #
######################
# Set the experiment specific policies
agent_configuration.policy = DecayingEpsilonGreedyPolicy(epsilon_max=1.0, epsilon_min=0.1, warm_up_steps=warm_up_steps,
                                                         steps=int((steps - warm_up_steps) * 0.5))
agent_configuration.test_policy = GreedyPolicy()


#####################
# Episode Callbacks #
#####################
# Generate the episode callbacks
episode_callbacks = get_episode_callbacks(episode_templates, training_agent, agent_configuration.warm_up_steps,
                                          epochs, save_epoch_interval, log_interval)

##############
# Experiment #
##############
experiment = Experiment(experiment_configuration, agent_configuration, model_configuration,
                        tree_configurations, episode_callbacks, sdt_threads=8)
env_description = get_env_description(experiment_configuration.env_name)


if __name__ == '__main__':
    gym.logger.set_level(gym.logger.DEBUG)
    torch.set_num_threads(1)  # Any other value than single core creates overhead

    # Train DQN agent
    experiment.train(train_agent_only=train_agent_only, verbose=1)
    # Plot DQN results
    dqn_visualizer = DQNVisualizer()
    dqn_visualizer.plot_all(experiment_configuration, training_dqn=True, num_leafs=None, colour_idx=0)
    # Test the best agent
    experiment.test_agent(episodes, log_results=True, visualize=False, use_best=True)
    dqn_visualizer.plot_all(experiment_configuration, training_dqn=False, use_best=True)

    # Train Soft Decision Trees on an already trained DQN agent
    experiment.train_on_final_agent(episodes=episodes)
    dqn_visualizer.plot_all(experiment_configuration, training_dqn=False, num_leafs=None)
    # Plot SDT results
    sdt_visualizer = SDTVisualizer()
    sdt_visualizer.plot_all(experiment_configuration, tree_configurations, episode_callbacks, save_epoch_interval,
                            training_agent)
    # Plot SDT Trees
    tree_visualizer = TreeVisualizer(env_description)
    tree_visualizer.visualize_all(experiment_configuration, tree_configurations, episode_callbacks)

    # Compare Policies
    experiment.test_trees_with_agent(steps=10000, agent_step_save=None, deterministic=True)
    # Plot policy comparisons
    policy_visualizer = PolicyVisualizer()
    policy_visualizer.plot_all(experiment_configuration, tree_configurations, episode_callbacks, training_agent)

    # Compare all agents
    # dqn_visualizer = DQNVisualizer()
    # dqn_visualizer.compare_same_config(experiment_configuration, seeds=range(30))
    # dqn_visualizer.plot_seeds_all(experiment_configuration, seeds=range(30))

    # sdt_visualizer = SDTVisualizer()
    # sdt_visualizer.compare_all(experiment_configuration, tree_configurations, episode_callbacks, save_epoch_interval,
    #                            range(30))

    # policy_visualizer = PolicyVisualizer()
    # policy_visualizer.plot_avg_comparisons(experiment_configuration, tree_configurations, episode_callbacks,
    #                                        training_agent, range(30))
