import csv

from core.callbacks.loggers.dqn_logger import *
from core.experiment.config import ExperimentConfiguration
from core.experiment.environment_description import EnvironmentDescription
from core.paths import load_partial_files
from core.paths.dqn_path import DQNPaths
from visualization import print_loading, print_num_entries


class DQNLoader(object):
    """The loader for csv for a DQN agent."""
    def __init__(self):
        self.step_log = DQNStepLog()
        self.episode_log = DQNEpisodeLog()

    def load_step_data(self, exp_config: ExperimentConfiguration, training_dqn, use_best,
                       env_description: EnvironmentDescription):
        """Load step data for a DQN agent."""
        # Define the data to be loaded (skip state and q-values)
        skipped_fields = [self.step_log.state, self.step_log.q_values,
                          self.step_log.policy, self.step_log.policy_value_name]
        data = {field: [] for field in self.step_log.entry_fields if field not in skipped_fields}
        # For the actions, we collect frequencies
        data[self.step_log.action] = {idx: 0 for idx in range(env_description.num_actions())}
        # The policy name and value name
        policy = None
        policy_value_name = None

        # Load the data
        def get_path(interval=None):
            return DQNPaths.step_logs_path(exp_config, training_dqn, interval, use_best)
        paths = load_partial_files(get_path, LOG_STEP_SAVE_INTERVAL)
        for path in paths:
            print_loading(path)
            with open(path, "r") as file:
                reader = csv.reader(file)
                fields = None
                nb_entries = 0
                for entry in reader:
                    # Read the header
                    if fields is None:
                        fields = entry
                    # Read the entries
                    else:
                        # Load the data
                        for field, value in zip(fields, entry):
                            # Policy name
                            if field == self.step_log.policy:
                                policy = value
                            # Policy value name
                            elif field == self.step_log.policy_value_name:
                                policy_value_name = value
                            # Actions are used as frequencies
                            elif field == self.step_log.action:
                                data[field][int(value)] += 1
                            # Skip fields if necessary
                            elif field not in skipped_fields:
                                # Agent loss may not be given => pre-trained agent
                                if field == self.step_log.loss:
                                    value = 0 if len(value) == 0 else float(value)
                                # Only the policy value and reward need to be floats
                                elif field == self.step_log.policy_value or field == self.step_log.reward:
                                    try:
                                        value = float(value)
                                    except ValueError:
                                        value = None
                                # All other values are integers
                                else:
                                    value = int(value)
                                data[field].append(value)
                        # An other entry has been loaded
                        nb_entries += 1
                print_num_entries(nb_entries)
        # Return the data
        return data, policy, policy_value_name

    @staticmethod
    def load_memory_data(exp_config: ExperimentConfiguration, training_dqn, use_best,
                         env_description: EnvironmentDescription, num_leafs=None):
        """Load memory data for a DQN agent."""
        log = MemoryLog(env_description) if num_leafs is None else SDTMemoryLog(num_leafs, env_description)
        # Define the data to be loaded
        data = {field: [] for field in log.entry_fields}

        # Load the data
        def get_path(interval=None):
            return DQNPaths.memory_logs_path(exp_config, training_dqn, interval, use_best)
        paths = load_partial_files(get_path, LOG_STEP_SAVE_INTERVAL)
        for path in paths:
            print_loading(path)
            with open(path, "r") as file:
                reader = csv.reader(file)
                fields = None
                nb_entries = 0
                for entry in reader:
                    # Read the header
                    if fields is None:
                        fields = entry
                    # Read the entries
                    else:
                        # Load the data
                        for field, value in zip(fields, entry):
                            # Empty fields are considered 0
                            value = 0 if len(value) == 0 else int(value)
                            data[field].append(value)
                        # An other entry has been loaded
                        nb_entries += 1
                print_num_entries(nb_entries)
        # Return the data
        return data

    def load_episode_data(self, exp_config: ExperimentConfiguration, training_dqn, use_best):
        """Load episode data for a DQN agent."""
        # Define the data to be loaded
        skipped_fields = []
        data = {field: [] for field in self.episode_log.entry_fields if field not in skipped_fields}

        # Load the data
        def get_path(interval=None):
            return DQNPaths.episode_logs_path(exp_config, training_dqn, interval, use_best)
        paths = load_partial_files(get_path, LOG_EPISODE_SAVE_INTERVAL)
        for path in paths:
            print_loading(path)
            with open(path, "r") as file:
                reader = csv.reader(file)
                fields = None
                nb_entries = 0
                for entry in reader:
                    # Read the header
                    if fields is None:
                        fields = entry
                    # Read the entries
                    else:
                        # Load the data
                        for field, value in zip(fields, entry):
                            # Only the rewards need to be floats
                            if field == self.episode_log.episode_reward:
                                value = float(value)
                            # All other values are integers
                            else:
                                value = int(value)
                            data[field].append(value)
                        # An other entry has been loaded
                        nb_entries += 1
                print_num_entries(nb_entries)
        # Return the data
        return data

    def get_last_episode(self, exp_config: ExperimentConfiguration, training_dqn, use_best):
        """Get the last episode the DQN agent has played."""
        episode = 0
        episode_idx = 0
        for idx, field in enumerate(self.episode_log.entry_fields):
            if field == self.episode_log.episode:
                episode_idx = idx
                break
        path = DQNPaths.episode_logs_path(exp_config, training_dqn, best_save=use_best)
        with open(path, "r") as file:
            reader = csv.reader(file)
            for entry in reader:
                episode = entry[episode_idx]
        return int(episode)
