import os
from pathlib import Path
from typing import List

import numpy as np

from core.experiment.config import ExperimentConfiguration
from core.experiment.environment_description import get_env_description
from core.paths.compare_path import ComparePaths
from core.paths.dqn_path import DQNPaths
from visualization.dqn.dqn_loader import DQNLoader, MemoryLog, SDTMemoryLog
from visualization.theme import *
from visualization.visualizer import Visualizer


class DQNVisualizer(Visualizer):
    """The visualizer for a DQN agent."""
    def __init__(self, theme: ColourTheme = OrangeColours3(), lines: LineStyles = LineStyles(), dpi=72):
        # Super call
        super(DQNVisualizer, self).__init__(theme, lines, dpi)
        # The data loader
        self.loader = DQNLoader()
        self.moving_avg_episodes = [10, 100, 1000]
        self.moving_avg_steps = [100, 1000, 10000, 100000]
        self.ext = "pdf"

    @staticmethod
    def _create_default_title(exp_config: ExperimentConfiguration, field_name, x_label, val=None):
        val_string = "" if val is None else f" | {val}"
        return f"[Seed {exp_config.seed}] {field_name} per {x_label}{val_string}"

    @staticmethod
    def _create_default_average_title(field_name, x_label, val=None):
        val_string = "" if val is None else f" | {val}"
        return f"{field_name} per {x_label}{val_string}"

    @staticmethod
    def _create_custom_title(exp_config: ExperimentConfiguration, subtitle):
        return f"[Seed {exp_config.seed}] {subtitle}"

    def _plot_default_field(self, exp_config: ExperimentConfiguration, y_label, field_name, field_data, x_label, x_data,
                            save_directory, moving_average=None, is_step_window=True, colour_idx=None, line_idx=None):
        print(f"Plotting {y_label}...")
        title = self._create_default_title(exp_config, field_name, x_label)
        save_path = f"{save_directory}/{x_label}_{y_label}.{self.ext}"
        self._plot(title, x_label, y_label, x_data, field_data, save_path, colour_idx=colour_idx, line_idx=line_idx)
        # Plot a moving average if requested
        if moving_average is not None:
            points = len(x_data)
            for window in moving_average:
                if self._use_window(window, points, is_step_window):
                    new_x, new_y = self._calculate_moving_average(window, x_data, field_data)
                    save_path = f"{save_directory}/{x_label}_{y_label}_ma_{window}.{self.ext}"
                    self._plot(title, x_label, y_label, new_x, new_y, save_path,
                               colour_idx=colour_idx, line_idx=line_idx)

    def _plot_averaged_field(self, field, field_name, field_data_mean, field_data_std_dev, x_label, x_data, save_path,
                             colour_idx=None, line_idx=None, y_lim=None):
        print(f"Plotting {field}...")
        self._plot_averages(self._create_default_average_title(field_name, x_label), x_label, field, x_data,
                            field_data_mean, field_data_std_dev, save_path, colour_idx=colour_idx, line_idx=line_idx,
                            y_lim=y_lim)

    def _plot_actions(self, exp_config: ExperimentConfiguration, nb_steps, field_data, save_directory,
                      colour_idx=None, line_idx=None):
        print(f"Plotting action frequencies...")
        env_description = get_env_description(exp_config.env_name)
        save_path = f"{save_directory}/Action_Frequencies.{self.ext}"
        # Calculate the frequency
        for key in field_data.keys():
            field_data[key] /= nb_steps
        # Get the labels
        x_label = "Actions"
        y_label = "Frequency"
        action_labels = [env_description.actions[key] for key in field_data.keys()]
        x = list(field_data.keys())
        y = field_data.values()
        self._bar_plot(self._create_custom_title(exp_config, "Action Frequencies"), x_label, y_label, x, y,
                       action_labels, save_path, show_labels=False, show_values=True,
                       colour_idx=colour_idx, line_idx=line_idx)

    def _plot_memory_default(self, exp_config: ExperimentConfiguration, x_label, xs, ys, labels, save_directory,
                             moving_average=None, moving_average_only=False):
        print(f"Plotting Sampled Experiences...")
        title = self._create_default_title(exp_config, "Sampled Experiences", x_label)
        y_label = "Count"
        save_path = f"{save_directory}/Sampled_Experiences.{self.ext}"
        show_labels = True
        legend_loc = "lower right"
        if not moving_average_only:
            self._plots(title, x_label, y_label, xs, ys, labels, save_path, show_labels, legend_loc)
        # Plot a moving average if requested
        if moving_average is not None:
            points = len(xs[0])
            for window in moving_average:
                if self._use_window(window, points):
                    new_ys = []
                    new_xs = []
                    new_x = self._calculate_moving_average(window, xs[0])
                    for y in ys:
                        new_y = self._calculate_moving_average(window, y)
                        new_ys.append(*new_y)
                        new_xs.append(*new_x)
                    save_path = f"{save_directory}/Sampled_Experiences_ma_{window}.{self.ext}"
                    self._plots(title, x_label, y_label, new_xs, new_ys, labels, save_path, show_labels, legend_loc)

    def _plot_memory_leafs(self, exp_config: ExperimentConfiguration, x_label, xs, ys, labels, save_directory,
                           num_leafs, moving_average=None):
        print(f"Plotting Leaf Sampled Experiences...")
        y_label = "Count"
        legend_loc = "lower right"
        show_labels = True
        data_x = {}
        data_y = {}
        data_l = {}
        for x, y, label in zip(xs, ys, labels):
            leaf = int(label.split(": ")[0].replace("Leaf ", ""))
            action_label = label.split(": ")[1]
            if data_x.get(leaf) is None:
                data_x[leaf] = []
            if data_y.get(leaf) is None:
                data_y[leaf] = []
            if data_l.get(leaf) is None:
                data_l[leaf] = []
            data_x[leaf].append(x)
            data_y[leaf].append(y)
            data_l[leaf].append(action_label)
        # Make a plot pro each leaf separately
        for leaf in range(num_leafs):
            xl = data_x[leaf]
            yl = data_y[leaf]
            ll = data_l[leaf]
            title = self._create_default_title(exp_config, "Sampled Experiences", x_label, f"Leaf {leaf}")
            save_path = f"{save_directory}/Sampled_Experiences_Leaf_{leaf}.{self.ext}"
            self._plots(title, x_label, y_label, xl, yl, ll, save_path, show_labels, legend_loc)
            # Plot a moving average if requested
            if moving_average is not None:
                points = len(xs[0])
                for window in moving_average:
                    if self._use_window(window, points):
                        new_ys = []
                        new_xs = []
                        new_x = self._calculate_moving_average(window, xl[0])
                        for y in yl:
                            new_y = self._calculate_moving_average(window, y)
                            new_ys.append(*new_y)
                            new_xs.append(*new_x)
                        save_path = f"{save_directory}/Sampled_Experiences_Leaf_{leaf}_ma_{window}.{self.ext}"
                        self._plots(title, x_label, y_label, new_xs, new_ys, ll, save_path, show_labels, legend_loc)

    def plot_memory_data(self, exp_config: ExperimentConfiguration, training_dqn, use_best, num_leafs):
        """Plot all the requested values for step logs."""
        # Load the data
        env_description = get_env_description(exp_config.env_name)
        log = MemoryLog(env_description) if num_leafs is None else SDTMemoryLog(num_leafs, env_description)
        data = self.loader.load_memory_data(exp_config, training_dqn, use_best, env_description, num_leafs)
        x_label = self.loader.step_log.step
        labels = log.counts
        x_data = data[x_label]
        xs = [x_data for _ in labels]
        ys = [data[c] for c in labels]

        # Plot the data
        directory = DQNPaths.vis_log_directory(exp_config, training_dqn, use_best)
        os.makedirs(directory, exist_ok=True)
        self._plot_memory_default(exp_config, x_label, xs, ys, labels, directory, self.moving_avg_steps,
                                  moving_average_only=num_leafs is not None)
        if num_leafs is not None:
            self._plot_memory_leafs(exp_config, x_label, xs, ys, labels, directory, num_leafs, self.moving_avg_steps)

    def plot_step_data(self, exp_config: ExperimentConfiguration, training_dqn, use_best,
                       colour_idx=None, line_idx=None):
        """Plot all the requested values for step logs."""
        # Load the data
        env_description = get_env_description(exp_config.env_name)
        data, policy_name, policy_value_name = \
            self.loader.load_step_data(exp_config, training_dqn, use_best, env_description)
        x_label = self.loader.step_log.step
        x_data = data[x_label]
        nb_steps = len(x_data)

        # Plot the data
        directory = DQNPaths.vis_log_directory(exp_config, training_dqn, use_best)
        os.makedirs(directory, exist_ok=True)
        for field in data.keys():
            if field == self.loader.step_log.step:
                continue
            field_data = data[field]
            # Actions
            if field is self.loader.step_log.action:
                self._plot_actions(exp_config, nb_steps, field_data, directory, colour_idx=colour_idx, line_idx=line_idx)
            # Policy (only if a policy value exists to be plotted)
            elif training_dqn and field is self.loader.step_log.policy_value:
                if policy_value_name is not None:
                    self._plot_default_field(exp_config, policy_value_name, policy_name, field_data, x_label, x_data,
                                             directory, colour_idx=colour_idx, line_idx=line_idx)
            # All other data
            else:
                self._plot_default_field(exp_config, field, field, field_data, x_label, x_data, directory,
                                         colour_idx=colour_idx, line_idx=line_idx)

    def plot_episode_data(self, exp_config: ExperimentConfiguration, training_dqn, use_best,
                          colour_idx=None, line_idx=None):
        """Plot all the requested values for episode logs."""
        # Load the data
        data = self.loader.load_episode_data(exp_config, training_dqn, use_best)
        x_label = self.loader.episode_log.episode
        x_data = data[x_label]

        # Plot the data
        directory = DQNPaths.vis_log_directory(exp_config, training_dqn, use_best)
        os.makedirs(directory, exist_ok=True)
        for field in data.keys():
            if field == self.loader.episode_log.episode:
                continue
            field_data = data[field]
            moving_average = self.moving_avg_episodes if field == self.loader.episode_log.episode_reward else None
            self._plot_default_field(exp_config, field, field, field_data, x_label, x_data, directory,
                                     moving_average, is_step_window=False, colour_idx=colour_idx, line_idx=line_idx)

    def plot_seeds_episode_data(self, exp_config: ExperimentConfiguration, training_dqn, use_best, seeds,
                                colour_idx=None, line_idx=None, y_lim=None):
        # Load the data for each seed
        data = {}
        x_label = self.loader.episode_log.episode
        x_data = []
        for seed in seeds:
            exp_config.save_directory = Path(exp_config.base_save_directory) / f"{seed}"
            agent_data = self.loader.load_episode_data(exp_config, training_dqn, use_best)
            data[seed] = agent_data
            agent_x_data = agent_data[x_label]
            if len(x_data) < len(agent_x_data):
                x_data = agent_x_data

        # Plot the data
        directory = ComparePaths.vis_avg_dqn_comparison_directory(exp_config, training_dqn, use_best)
        os.makedirs(directory, exist_ok=True)
        for field in data[0].keys():
            if field == self.loader.episode_log.episode:
                continue
            field_data = list(map(lambda d: np.array(d[field]), data.values()))
            field_data_mean, field_data_std_dev = self._calculate_statistics(field_data)

            if training_dqn and field == self.loader.episode_log.episode_reward:
                # print(field_data_mean[-30:-20])
                idx = np.where(~np.isnan(field_data_mean))[0][-100:]
                # print(idx)
                print("(MA 100) train mean:", np.mean(field_data_mean[idx]),
                      "(MA 100) std dev:", np.mean(field_data_std_dev[idx]))
                idx = np.where(~np.isnan(field_data_mean))[0][-10:]
                # print(idx)
                print("(MA 10) train mean:", np.mean(field_data_mean[idx]),
                      "(MA 10) std dev:", np.mean(field_data_std_dev[idx]))
                idx = np.where(~np.isnan(field_data_mean))[0][-1]
                # print(idx)
                print("train mean:", np.mean(field_data),
                      "std dev:", np.std(field_data))

            save_path = f"{directory}/{x_label}_{field}.{self.ext}"
            self._plot_averaged_field(field, field, field_data_mean, field_data_std_dev, x_label, x_data, save_path,
                                      colour_idx=colour_idx, line_idx=line_idx, y_lim=y_lim)
            for window in self.moving_avg_episodes:
                if self._use_window(window, len(x_data)):
                    new_field_data_mean, new_field_data_std_dev, new_x = \
                        self._calculate_moving_average(window, field_data_mean, field_data_std_dev, x_data)
                    save_path = f"{directory}/{x_label}_{field}_moving_average_{window}.{self.ext}"
                    self._plot_averaged_field(field, field, new_field_data_mean, new_field_data_std_dev,
                                              x_label, new_x, save_path, colour_idx=colour_idx, line_idx=line_idx)

    def plot_seeds_all(self, exp_config: ExperimentConfiguration, seeds, colour_idx=None, line_idx=None,
                       y_lim=(0, 500)):
        # Load the data for each seed
        data_best = {}
        data_final = {}
        x_label = self.loader.episode_log.episode
        x_data = []
        for seed in seeds:
            exp_config.save_directory = Path(exp_config.base_save_directory) / f"{seed}"
            agent_data = self.loader.load_episode_data(exp_config, training_dqn=False, use_best=False)
            data_final[seed] = agent_data
            agent_x_data = agent_data[x_label]
            if len(x_data) < len(agent_x_data):
                x_data = agent_x_data
            agent_data = self.loader.load_episode_data(exp_config, training_dqn=False, use_best=True)
            data_best[seed] = agent_data
            agent_x_data = agent_data[x_label]
            if len(x_data) < len(agent_x_data):
                x_data = agent_x_data

        # Plot the data
        directory = ComparePaths.vis_all_dqn_directory(exp_config)
        os.makedirs(directory, exist_ok=True)
        field = self.loader.episode_log.episode_reward
        field_data_best = list(map(lambda d: np.array(d[field]), data_best.values()))
        field_data_final = list(map(lambda d: np.array(d[field]), data_final.values()))
        fd_best_mean, fd_best_std_dev = self._calculate_statistics(field_data_best)
        fd_final_mean, fd_final_std_dev = self._calculate_statistics(field_data_final)
        save_path = f"{directory}/{x_label}_{field}.{self.ext}"
        self._plot_multiple_averages(self._create_default_average_title(field, self.loader.episode_log.episode),
                                     self.loader.episode_log.episode, field, [x_data, x_data],
                                     [fd_final_mean, fd_best_mean], [fd_final_std_dev, fd_best_std_dev],
                                     ["final", "best"], save_path, y_lim=y_lim, plot_between=False,
                                     legend_location="lower right", colour_idx=colour_idx, line_idx=line_idx)
        print("final mean:", np.mean(fd_final_mean),
              "std dev", np.mean(fd_final_std_dev))
        print("best mean:", np.mean(fd_best_mean),
              "std dev", np.mean(fd_best_std_dev))

    def plot_all(self, exp_config: ExperimentConfiguration, training_dqn, use_best=False, num_leafs=None,
                 colour_idx=None, line_idx=None, plot_memory=True):
        # Step logs
        self.plot_step_data(exp_config, training_dqn, use_best, colour_idx, line_idx)
        # Episode logs
        self.plot_episode_data(exp_config, training_dqn, use_best, colour_idx, line_idx)
        # Memory logs
        if training_dqn and plot_memory:
            self.plot_memory_data(exp_config, training_dqn, use_best, num_leafs)

    def compare_same_config(self, exp_config: ExperimentConfiguration, seeds, colour_idx=None, line_idx=None,
                            y_lim=(0, 500)):
        # Compare training process
        training_dqn = True
        use_best = False
        self.plot_seeds_episode_data(exp_config, training_dqn, use_best, seeds, colour_idx, line_idx, y_lim)
        # Compare testing process
        training_dqn = False
        self.plot_seeds_episode_data(exp_config, training_dqn, use_best, seeds, colour_idx, line_idx)
        use_best = True
        self.plot_seeds_episode_data(exp_config, training_dqn, use_best, seeds, colour_idx, line_idx)

    def plot_seeds_configs(self, exp_configs: List[ExperimentConfiguration], names, seeds, y_lim=(0, 500)):
        # Load the data for each seed
        data_train_mean = []
        data_best_mean = []
        data_final_mean = []
        data_train_std_dev = []
        data_best_std_dev = []
        data_final_std_dev = []
        x_label = self.loader.episode_log.episode
        x_data_train = []
        x_data = []
        field = self.loader.episode_log.episode_reward
        for idx, exp_config in enumerate(exp_configs):
            temp_train = []
            temp_final = []
            temp_best = []
            temp_x_train = []
            temp_x = []
            for seed in seeds:
                exp_config.save_directory = Path(exp_config.base_save_directory) / f"{seed}"
                # Train
                agent_data = self.loader.load_episode_data(exp_config, training_dqn=True, use_best=False)
                temp_train.append(agent_data[field])
                agent_x_data = agent_data[x_label]
                if len(temp_x_train) < len(agent_x_data):
                    temp_x_train = agent_x_data
                # Final
                agent_data = self.loader.load_episode_data(exp_config, training_dqn=False, use_best=False)
                temp_final.append(agent_data[field])
                agent_x_data = agent_data[x_label]
                if len(temp_x) < len(agent_x_data):
                    temp_x = agent_x_data
                # Best
                agent_data = self.loader.load_episode_data(exp_config, training_dqn=False, use_best=True)
                temp_best.append(agent_data[field])
                agent_x_data = agent_data[x_label]
                if len(temp_x) < len(agent_x_data):
                    temp_x = agent_x_data
            # Calculate averages over seeds
            train_mean, train_std_dev = self._calculate_statistics(temp_train)
            final_mean, final_std_dev = self._calculate_statistics(temp_final)
            best_mean, best_std_dev = self._calculate_statistics(temp_best)
            data_train_mean.append(train_mean)
            data_train_std_dev.append(train_std_dev)
            data_final_mean.append(final_mean)
            data_final_std_dev.append(final_std_dev)
            data_best_mean.append(best_mean)
            data_best_std_dev.append(best_std_dev)

            x_data_train.append(temp_x_train)
            x_data.append(temp_x)

        # Plot the data
        directory = ComparePaths.vis_full_comparison_dqn(exp_configs, no_seed=True)
        os.makedirs(directory, exist_ok=True)
        # Train
        save_path = f"{directory}/{x_label}_{field}_Train.{self.ext}"
        self._plot_multiple_averages(self._create_default_average_title(field, x_label), x_label, field,
                                     x_data_train, data_train_mean, data_train_std_dev, names, save_path,
                                     y_lim=y_lim, plot_between=False, legend_location="upper left")

        print("train:", list(map(np.mean, data_train_mean)))
        print("final:", list(map(np.mean, data_final_mean)))
        print("best:", list(map(np.mean, data_best_mean)))

        for window in self.moving_avg_episodes:
            if self._use_window(window, len(x_data_train[0])):
                x_data_train = list(map(lambda d: self._calculate_moving_average(window, d)[0], x_data_train))
                data_train_mean = list(map(lambda d: self._calculate_moving_average(window, d)[0], data_train_mean))
                data_train_std_dev = list(map(lambda d: self._calculate_moving_average(window, d)[0], data_train_std_dev))
                save_path = f"{directory}/{x_label}_{field}_moving_average_{window}.{self.ext}"
                self._plot_multiple_averages(self._create_default_average_title(field, x_label), x_label, field,
                                             x_data_train, data_train_mean, data_train_std_dev, names, save_path,
                                             y_lim=y_lim, plot_between=False, legend_location="upper left")
        # Final
        save_path = f"{directory}/{x_label}_{field}_Final.{self.ext}"
        self._plot_multiple_averages(self._create_default_average_title(field, x_label), x_label, field,
                                     x_data, data_final_mean, data_final_std_dev, names, save_path,
                                     y_lim=y_lim, plot_between=False, legend_location="lower right")
        # Best
        save_path = f"{directory}/{x_label}_{field}_Best.{self.ext}"
        self._plot_multiple_averages(self._create_default_average_title(field, x_label), x_label, field,
                                     x_data, data_best_mean, data_best_std_dev, names, save_path,
                                     y_lim=y_lim, plot_between=False, legend_location="lower right")
