from pathlib import Path
from typing import List

import numpy as np

from core.callbacks.loggers.action_logger import PostActionsLog
from core.callbacks.loggers.sdt_logger import *
from core.callbacks.sdt.episode_callback import EpisodeCallbackTemplate
from core.callbacks.sdt.sdt_callback import SDTCallbackTemplate
from core.experiment.config import ExperimentConfiguration, TreeConfiguration
from core.experiment.environment_description import get_env_description
from core.paths.compare_path import ComparePaths
from visualization import compare_per_tree_config
from visualization.dqn.dqn_loader import DQNLoader
from visualization.policy.policy_loader import PolicyLoader
from visualization.theme import *
from visualization.visualizer import Visualizer


class PolicyVisualizer(Visualizer):
    """The visualizer for a soft decision tree."""
    def __init__(self, theme: ColourTheme = BlueColours3(), lines: LineStyles = LineStyles(), dpi=72):
        # Super call
        super(PolicyVisualizer, self).__init__(theme, lines, dpi)
        # The data loader
        self.loader = PolicyLoader()
        self.ext = "pdf"

    @staticmethod
    def _create_general_title(exp_config: ExperimentConfiguration, value=None, average=False):
        seed_string = "" if average else f"[Seed {exp_config.seed}] "
        return f"{seed_string}{value}"

    @staticmethod
    def _create_tree_config_title(exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                                  value=None, average=False):
        seed_string = "" if average else f"[Seed {exp_config.seed}] "
        val_string = "" if value is None else f": {value}"
        return f"{seed_string}{tree_config.name}{val_string}"

    @staticmethod
    def _create_callback_title(exp_config: ExperimentConfiguration, sdt_callback: SDTCallbackTemplate,
                               value=None, average=False):
        seed_string = "" if average else f"[Seed {exp_config.seed}] "
        val_string = "" if value is None else f": {value}"
        return f"{seed_string}{sdt_callback.create_name()}{val_string}"

    def _create_specific_title(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                               sdt_callback: SDTCallbackTemplate, value=None, average=False):
        seed_string = "" if average else f"[Seed {exp_config.seed}] "
        val_string = "" if value is None else f": {value}"
        name = self._get_short_name(sdt_callback)
        return f"{seed_string}{tree_config.name} | {name}{val_string}"

    @staticmethod
    def _get_short_name(sdt_callback: SDTCallbackTemplate):
        new_name = sdt_callback.create_name().replace("EpisodeCallback_", "").replace("_", "@")
        return new_name

    @staticmethod
    def _to_string(val):
        return str(round(val, 4))

    def _vis_per_action(self, x_label, y_labels, data, labels, env_description: EnvironmentDescription, log,
                        create_title, save_directory, legend_location=None, plot_name=None):
        for y_label in y_labels:
            full_ys = list(map(lambda c: c[y_label], data))
            # For each action (+ All actions combined)
            for action in range(env_description.num_actions() + 1):
                ys = list(map(lambda c: c[action], full_ys))
                action_name = env_description.actions[action] \
                    if action < env_description.num_actions() else log.all_actions
                label = f"{y_label} |\n{action_name}"
                name = f"{y_label}_{action_name}" if plot_name is None else f"{plot_name}_{y_label}_{action_name}"
                name = name.replace(" ", "_")
                title = create_title(label.replace("Action Order Average", "Order Match Average"))
                save_path = f"{save_directory}/{name}.{self.ext}"
                self._bar_plot(title, x_label, y_label.replace("Action Order Average", "Order Match Average"), range(len(labels)), ys, labels, save_path, show_labels=False,
                               show_values=True, legend_location=legend_location, to_string=self._to_string)

    def visualize_per_tree_config(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                                  sdt_callbacks: List[SDTCallbackTemplate], plot_name=None):
        # Load all callbacks for the given tree configuration
        env_description = get_env_description(exp_config.env_name)
        log = PostActionsLog(env_description, tree_config)
        data = []
        labels = []
        # Load the data
        for callback in sdt_callbacks:
            callback_data = self.loader.load_actions_data(exp_config, tree_config, callback)
            data.append(callback_data)
            c_name = self._get_short_name(callback)
            labels.append(c_name)
        # Gather the rest of the plot data
        directory = ComparePaths.vis_per_tree_directory(exp_config, tree_config)
        os.makedirs(directory, exist_ok=True)
        x_label = "Episode Callbacks"
        data_labels = [log.accuracy, log.order_avg]
        create_title = lambda x: self._create_tree_config_title(exp_config, tree_config, x)
        self._vis_per_action(x_label, data_labels, data, labels, env_description, log, create_title, directory,
                             plot_name=plot_name)

    def visualize_per_callback(self, exp_config: ExperimentConfiguration,
                               tree_configs: List[TreeConfiguration], sdt_callback: SDTCallbackTemplate):
        # Load all callbacks for the given tree configuration
        env_description = get_env_description(exp_config.env_name)
        log = None
        data = []
        labels = []
        # Load the data
        for tree_config in tree_configs:
            tree_data = self.loader.load_actions_data(exp_config, tree_config, sdt_callback)
            data.append(tree_data)
            labels.append(tree_config.name)
            log = PostActionsLog(env_description, tree_config)
        # Gather the rest of the plot data
        directory = ComparePaths.vis_per_callback_directory(exp_config, sdt_callback.create_name())
        os.makedirs(directory, exist_ok=True)
        x_label = "Soft Decision Trees"
        data_labels = [log.accuracy, log.order_avg]
        create_title = lambda x: self._create_callback_title(exp_config, sdt_callback, x)
        self._vis_per_action(x_label, data_labels, data, labels, env_description, log, create_title, directory)

    def visualize_all_per_callback_and_tree_config(self, exp_config: ExperimentConfiguration,
                                                   tree_configs: List[TreeConfiguration],
                                                   sdt_callbacks: List[SDTCallbackTemplate], plot_name):
        # Load all callbacks for per tree configuration
        env_description = get_env_description(exp_config.env_name)
        log = PostActionsLog(env_description, tree_configs[0])
        data = {depth: [] for depth in map(lambda tc: tc.depth, tree_configs)}
        callback_labels = []
        # Load the data
        for callback in sdt_callbacks:
            callback_labels.append(self._get_short_name(callback))
            for tree_config in tree_configs:
                d = self.loader.load_actions_data(exp_config, tree_config, callback)
                data[tree_config.depth].append(d)
        tree_labels = list(map(lambda tc: tc.name, tree_configs))
        # Gather the rest of the plot data
        directory = ComparePaths.vis_per_all(exp_config)
        os.makedirs(directory, exist_ok=True)
        x_label = "Soft Decision Trees"
        data_labels = [log.accuracy, log.order_avg]

        for y_label in data_labels:
            full_ys = []
            for tc in data.values():
                y = list(map(lambda c: c[y_label], tc))
                y = list(zip(*y))
                full_ys.append(y)
            # For each action (+ All actions combined)
            for action in range(env_description.num_actions() + 1):
                ys = list(map(lambda c: c[action], full_ys))
                action_name = env_description.actions[action] \
                    if action < env_description.num_actions() else log.all_actions
                label = f"{y_label} | {action_name}"
                name = f"{y_label}_{action_name}" if plot_name is None else f"{plot_name}_{y_label}_{action_name}"
                name = name.replace(" ", "_")
                title = self._create_general_title(exp_config, label)
                save_path = f"{directory}/{name}.{self.ext}"
                self._grouped_bar_plot(title, x_label, y_label, ys, callback_labels, tree_labels, save_path,
                                       show_values=True, legend_location="lower right", to_string=self._to_string)

    def visualize_leafs_per_tree_config(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                                        sdt_callbacks: List[SDTCallbackTemplate], plot_name):
        # Load all callbacks for the given tree configuration
        env_description = get_env_description(exp_config.env_name)
        log = PostActionsLog(env_description, tree_config)
        data = []
        labels = []
        # Load the data
        for callback in sdt_callbacks:
            callback_data = self.loader.load_actions_data(exp_config, tree_config, callback)
            data.append(callback_data)
            c_name = self._get_short_name(callback)
            labels.append(c_name)
        # Gather the rest of the plot data
        directory = ComparePaths.vis_per_tree_directory(exp_config, tree_config)
        os.makedirs(directory, exist_ok=True)
        x_label = "Leaf Nodes"
        y_label = "Times Used"
        leaf_labels = log.leafs
        action_labels = list(env_description.actions.keys()) + [log.all_actions]

        # For each action (+ All actions combined), gather the number of times each leaf is used
        leafs_per_action_data = {action: {} for action in range(len(action_labels))}
        for leaf in leaf_labels:
            leaf_data = list(map(lambda c: c[leaf], data))
            for action in range(env_description.num_actions() + 1):
                leaf_action_data = list(map(lambda c: c[action], leaf_data))
                leafs_per_action_data[action][leaf] = leaf_action_data
        # Draw the plots
        for action, data_per_leaf in leafs_per_action_data.items():
            ys = list(zip(*data_per_leaf.values()))
            action_name = env_description.actions[action] if action < env_description.num_actions() else log.all_actions
            label = f"{y_label} | {action_name}"
            name = f"{plot_name}_{y_label}_{action_name}".replace(" ", "_")
            title = self._create_tree_config_title(exp_config, tree_config, label)
            save_path = f"{directory}/{name}.{self.ext}"
            self._grouped_bar_plot(title, x_label, y_label, ys, leaf_labels, labels, save_path, show_values=True,
                                   legend_location="lower right")

    def compare_per_tree_config(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                                sdt_callbacks: List[SDTCallbackTemplate], training_agent):
        # Sort the comparisons to make the plots more readable
        per_interval, per_episodes, per_full_int, per_times_trained = \
            compare_per_tree_config(sdt_callbacks, DQNLoader(), exp_config, training_agent)
        # Make plots comparing per interval
        for interval, callbacks in per_interval.items():
            if len(callbacks) > 1:
                plot_name = f"I_{interval}"
                self.visualize_per_tree_config(exp_config, tree_config, callbacks, plot_name)
                if "Cart" in exp_config.env_name:
                    self.visualize_leafs_per_tree_config(exp_config, tree_config, callbacks, plot_name)
        # Make plots comparing per episodes used
        for episodes, callbacks in per_episodes.items():
            if len(callbacks) > 1:
                plot_name = f"E_{episodes}"
                self.visualize_per_tree_config(exp_config, tree_config, callbacks, plot_name)
                if "Cart" in exp_config.env_name:
                    self.visualize_leafs_per_tree_config(exp_config, tree_config, callbacks, plot_name)
        # Make plots comparing per times trained
        for times, callbacks in per_times_trained.items():
            if len(callbacks) > 1:
                plot_name = f"U_{times}"
                self.visualize_per_tree_config(exp_config, tree_config, callbacks, plot_name)
                if "Cart" in exp_config.env_name:
                    self.visualize_leafs_per_tree_config(exp_config, tree_config, callbacks, plot_name)
        # Make plots per full intervals used
        if len(per_full_int) > 0:
            plot_name = f"F"
            self.visualize_per_tree_config(exp_config, tree_config, per_full_int, plot_name)
            if "Cart" in exp_config.env_name:
                self.visualize_leafs_per_tree_config(exp_config, tree_config, per_full_int, plot_name)

    def compare_experiments(self, env_description: EnvironmentDescription,
                            tree_config: TreeConfiguration, sdt_callback: SDTCallbackTemplate, seed, args):
        # Load the data for all trees
        log = PostActionsLog(env_description, tree_config)
        labels = [a[0] for a in args]
        configs = list(map(lambda a: a[1], args))
        for c in configs:
            c.save_directory = Path(c.base_save_directory) / f"{seed}"
        data = [self.loader.load_actions_data(c, tree_config, sdt_callback) for c in configs]
        directory = ComparePaths.vis_full_comparison_tree(configs, tree_config, sdt_callback.create_name())
        os.makedirs(directory, exist_ok=True)
        x_label = "Soft Decision Trees"
        data_labels = [log.accuracy, log.order_avg]
        create_title = lambda x: self._create_specific_title(configs[0], tree_config, sdt_callback, x)
        self._vis_per_action(x_label, data_labels, data, labels, env_description, log, create_title, directory)

    def plot_all(self, exp_config: ExperimentConfiguration, tree_configs: List[TreeConfiguration],
                 sdt_callbacks: List[SDTCallbackTemplate], training_agent):
        # For each tree configuration
        for tree_config in tree_configs:
            self.compare_per_tree_config(exp_config, tree_config, sdt_callbacks, training_agent)
        # For each callback
        for callback in sdt_callbacks:
            self.visualize_per_callback(exp_config, tree_configs, callback)

        # Sort the comparisons to make the plots more readable
        per_interval, per_episodes, per_full_int, per_times_trained = \
            compare_per_tree_config(sdt_callbacks, DQNLoader(), exp_config, training_agent)
        # Make plots comparing per interval
        for interval, callbacks in per_interval.items():
            if len(callbacks) > 1:
                plot_name = f"I_{interval}"
                self.visualize_all_per_callback_and_tree_config(exp_config, tree_configs, callbacks, plot_name)
        # Make plots comparing per episodes used
        for episodes, callbacks in per_episodes.items():
            if len(callbacks) > 1:
                plot_name = f"E_{episodes}"
                self.visualize_all_per_callback_and_tree_config(exp_config, tree_configs, callbacks, plot_name)
        # Make plots comparing per times trained
        for times, callbacks in per_times_trained.items():
            if len(callbacks) > 1:
                plot_name = f"U_{times}"
                self.visualize_all_per_callback_and_tree_config(exp_config, tree_configs, callbacks, plot_name)
        # Make plots per full intervals used
        if len(per_full_int) > 1:
            plot_name = f"F"
            self.visualize_all_per_callback_and_tree_config(exp_config, tree_configs, per_full_int, plot_name)

    def plot_comparison(self, post_arg, *args):
        # Extract the post arguments
        post_name, post_exp_conf, post_tree_configs, post_callbacks = post_arg
        # Get a description of the environment for the given experiment configurations
        env_description = get_env_description(post_exp_conf.env_name)
        # For each of the given argument pairs (name, exp_conf, tree_configs, sdt_callbacks), plot comparisons
        arguments = {}
        names = []
        print("Comparing:", post_name)
        for arg in args:
            name, exp_conf, tree_configs, sdt_callbacks = arg
            print("Comparing:", name)
            arguments[name] = arg
            names.append(name)

        # For each concurrent SDT, compare with an equivalent pre-trained SDT
        # (We don't care about non-perfect SDT matches, as we want to compare the exact same configuration here)
        for tree_config in post_tree_configs:
            for callback in post_callbacks:
                # Try finding the equivalent for all other arguments
                found = {name: False for name in names}
                found[post_name] = post_arg
                comp_args = [post_arg]
                for name, arg in arguments.items():
                    name, exp_conf, tree_configs, sdt_callbacks = arg
                    # Try finding the equivalent pre trained version
                    if isinstance(callback, EpisodeCallbackTemplate):
                        tree_config_found = False
                        callback_found = False
                        for t in tree_configs:
                            # Only the depth differs for now
                            if t.depth == tree_config.depth:
                                tree_config_found = True
                                break
                        # Only keep searching if a tree config has been found already
                        if tree_config_found:
                            for c in sdt_callbacks:
                                # Make sure the episode callback is equal
                                if isinstance(c, EpisodeCallbackTemplate) and \
                                        (c.episode_interval == callback.episode_interval) and \
                                        (c.episodes == callback.episodes):
                                    callback_found = True
                                    break
                        # Indicate whether of not a match has been found
                        found[name] = tree_config_found and callback_found
                        if found[name]:
                            comp_args.append(arg)

                # Only compare the if there is something to compare with
                can_compare = any(found.values())
                if can_compare and len(comp_args) > 1:
                    print(f"Comparing {[name for name, a in found.items() if a]} "
                          f"For SDT {tree_config.name}, {callback.create_name()}")
                    self.compare_experiments(env_description, tree_config, callback, post_exp_conf.seed, comp_args)

    def visualize_all_averages(self, exp_config: ExperimentConfiguration, tree_configs: List[TreeConfiguration],
                               sdt_callbacks: List[SDTCallbackTemplate], plot_name, seeds):
        # Load all callbacks for per tree configuration
        env_description = get_env_description(exp_config.env_name)
        log = PostActionsLog(env_description, tree_configs[0])
        data = {}
        callback_labels = [self._get_short_name(callback) for callback in sdt_callbacks]
        tree_labels = [t.name for t in tree_configs]
        # Load the data
        for tree_config in tree_configs:
            data[tree_config.depth] = {}
            for callback in sdt_callbacks:
                c_name = callback.create_name()
                data[tree_config.depth][c_name] = []
                for seed in seeds:
                    exp_config.save_directory = Path(exp_config.base_save_directory) / f"{seed}"
                    d = self.loader.load_actions_data(exp_config, tree_config, callback)
                    data[tree_config.depth][c_name].append(d)

        # Gather the rest of the plot data
        directory = ComparePaths.vis_avg_all_sdt_directory(exp_config)
        os.makedirs(directory, exist_ok=True)
        x_label = "Soft Decision Trees"
        data_labels = [log.accuracy, log.order_avg]

        for y_label in data_labels:
            full_ys_mean = []
            full_ys_std_dev = []
            for depth_k, tc in data.items():
                fm = []
                fs = []
                for call_k, c in tc.items():
                    y_seeds_per_callback = list(map(lambda d: d[y_label], c))
                    y_seeds_per_callback = np.vstack(y_seeds_per_callback)
                    mean, std_dev = self._calculate_statistics(y_seeds_per_callback)
                    fm.append(mean)
                    fs.append(std_dev)
                full_ys_mean.append(list(zip(*fm)))
                full_ys_std_dev.append(list(zip(*fs)))
            # For each action (+ All actions combined)
            for action in range(env_description.num_actions() + 1):
                ys_mean = list(map(lambda c: c[action], full_ys_mean))
                ys_std_dev = list(map(lambda c: c[action], full_ys_std_dev))
                action_name = env_description.actions[action] \
                    if action < env_description.num_actions() else log.all_actions
                # y_label = y_label.replace("Action Order Average", "Match Order Average")
                label = f"{y_label} | {action_name}"
                name = f"{y_label}_{action_name}" if plot_name is None else f"{plot_name}_{y_label}_{action_name}"
                name = name.replace(" ", "_")
                title = self._create_general_title(exp_config, label.replace("Action Order Average", "Order Match Average"), average=True)
                save_path = f"{directory}/{name}.{self.ext}"
                self._averaged_grouped_bar_plot(title, x_label, y_label.replace("Action Order Average", "Order Match Average"), ys_mean, ys_std_dev, callback_labels,
                                                tree_labels, save_path, show_values=True, legend_location="lower right",
                                                to_string=self._to_string)

    def plot_avg_comparisons(self, exp_config: ExperimentConfiguration, tree_configs: List[TreeConfiguration],
                             sdt_callbacks: List[SDTCallbackTemplate], training_agent, seeds, plot_if_one=False):
        # Sort the comparisons to make the plots more readable
        per_interval, per_episodes, per_full_int, per_times_trained = \
            compare_per_tree_config(sdt_callbacks, DQNLoader(), exp_config, training_agent)
        # Make plots comparing per interval
        for interval, callbacks in per_interval.items():
            if plot_if_one or len(callbacks) > 1:
                plot_name = f"I_{interval}"
                self.visualize_all_averages(exp_config, tree_configs, callbacks, plot_name, seeds)
        # Make plots comparing per episodes used
        for episodes, callbacks in per_episodes.items():
            if len(callbacks) > 1:
                plot_name = f"E_{episodes}"
                self.visualize_all_averages(exp_config, tree_configs, callbacks, plot_name, seeds)
        # Make plots comparing per times trained
        for times, callbacks in per_times_trained.items():
            if len(callbacks) > 1:
                plot_name = f"U_{times}"
                self.visualize_all_averages(exp_config, tree_configs, callbacks, plot_name, seeds)
        # Make plots per full intervals used
        if len(per_full_int) > 1:
            plot_name = f"F"
            self.visualize_all_averages(exp_config, tree_configs, per_full_int, plot_name, seeds)
