import csv

from core.callbacks.loggers.sdt_logger import *
from core.callbacks.sdt.sdt_callback import SDTCallbackTemplate
from core.experiment.config import ExperimentConfiguration, TreeConfiguration
from core.experiment.environment_description import get_env_description
from core.paths import load_partial_files
from core.paths.sdt_path import SDTPaths
from visualization import print_loading, print_num_entries, string_to_list


class SDTLoader(object):
    """The loader for csv for a Soft Decision Tree."""
    def __init__(self):
        self.losses_log = SDTLossesLog()
        self.actions_log = SDTActionsLog()

    def load_loss_data(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                       sdt_callback: SDTCallbackTemplate, epoch_save_interval):
        # Define the data to be loaded
        data = {field: [] for field in self.losses_log.entry_fields}

        # Load the data
        def get_path(interval=None):
            return SDTPaths.log_losses_path(exp_config, tree_config, sdt_callback.create_name(), interval)
        paths = load_partial_files(get_path, epoch_save_interval)
        for path in paths:
            print_loading(path)
            with open(path, "r") as file:
                reader = csv.reader(file)
                nb_entries = 0
                fields = None
                for entry in reader:
                    # Read the header
                    if fields is None:
                        fields = entry
                    # Read the entries
                    else:
                        # Load the data
                        for field, value in zip(fields, entry):
                            # Agent loss may not exist (for pre-trained agents)
                            if field == self.losses_log.agent_loss and len(value) == 0:
                                value = 0
                            # Store the entry value
                            data[field].append(float(value))
                        # An other entry has been loaded
                        nb_entries += 1
                print_num_entries(nb_entries)
        # Return the data
        return data

    @staticmethod
    def _load_node_weights(node_idx, is_leaf, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                           sdt_callback: SDTCallbackTemplate, epoch_save_interval):
        # Define the data to be loaded
        env_description = get_env_description(exp_config.env_name)
        log = SDTLeafWeightsLog(env_description) if is_leaf else SDTInnerWeightsLog(env_description)
        data = {field: [] for field in log.entry_fields}

        # Load the data
        def get_path(interval):
            if is_leaf:
                return SDTPaths.log_leaf_path(exp_config, tree_config, sdt_callback.create_name(), node_idx, interval)
            else:
                return SDTPaths.log_inner_path(exp_config, tree_config, sdt_callback.create_name(), node_idx, interval)
        paths = load_partial_files(get_path, epoch_save_interval)
        for path in paths:
            # print_loading(path)
            with open(path, "r") as file:
                reader = csv.reader(file)
                # nb_entries = 0
                fields = None
                for entry in reader:
                    # Read the header
                    if fields is None:
                        fields = entry
                    # Read the entries
                    else:
                        # Load the data
                        for field, value in zip(fields, entry):
                            data[field].append(float(value))
                        # An other entry has been loaded
                        # nb_entries += 1
                # print_num_entries(nb_entries)
        # Return the data
        return data

    def load_weights_data(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                          sdt_callback: SDTCallbackTemplate, epoch_save_interval):
        # The data to be returned
        inners = {}
        leafs = {}
        # Load inner nodes
        for inner in range(2 ** tree_config.depth - 1):
            data = self._load_node_weights(inner, False, exp_config, tree_config, sdt_callback, epoch_save_interval)
            inners[inner] = data
        # Load leaf nodes
        for leaf in range(2 ** tree_config.depth):
            data = self._load_node_weights(leaf, True, exp_config, tree_config, sdt_callback, epoch_save_interval)
            leafs[leaf] = data
        # Return the data
        return inners, leafs

    def load_actions_data(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                          sdt_callback: SDTCallbackTemplate, epoch_save_interval):
        # Define the data to be loaded
        data = {field: [] for field in self.actions_log.entry_fields}

        # Load the data
        def get_path(interval=None):
            return SDTPaths.log_actions_path(exp_config, tree_config, sdt_callback.create_name(), interval)
        paths = load_partial_files(get_path, epoch_save_interval)
        for path in paths:
            print_loading(path)
            with open(path, "r") as file:
                reader = csv.reader(file)
                nb_entries = 0
                fields = None
                for entry in reader:
                    # Read the header
                    if fields is None:
                        fields = entry
                    # Read the entries
                    else:
                        # Load the data
                        for field, value in zip(fields, entry):
                            # Leaf distributions and q-values are lists
                            if field == self.actions_log.leaf_distribution or field == self.actions_log.agent_q_values \
                                    or field == self.actions_log.error:
                                data[field].append(string_to_list(value))
                            else:
                                # Store the entry value
                                data[field].append(float(value))
                        # An other entry has been loaded
                        nb_entries += 1
                print_num_entries(nb_entries)
        # Return the data
        return data

    @staticmethod
    def load_post_actions_data(exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                               sdt_callback: SDTCallbackTemplate, action):
        # Define the data to be loaded
        env_description = get_env_description(exp_config.env_name)
        log = SDTPostActionsLog(env_description)
        data = {field: [] for field in log.entry_fields}

        # Load the data
        path = SDTPaths.log_post_actions_path(exp_config, tree_config, sdt_callback.create_name(), action)
        print_loading(path)
        with open(path, "r") as file:
            reader = csv.reader(file)
            nb_entries = 0
            fields = None
            for entry in reader:
                # Read the header
                if fields is None:
                    fields = entry
                # Read the entries
                else:
                    # Load the data
                    for field, value in zip(fields, entry):
                        # Store the entry value
                        data[field].append(float(value))
                    # An other entry has been loaded
                    nb_entries += 1
            print_num_entries(nb_entries)
        # Return the data
        return data
