from pathlib import Path
from typing import List

import numpy as np

from core.callbacks.loggers.sdt_logger import *
from core.callbacks.sdt.sdt_callback import SDTCallbackTemplate
from core.experiment.config import ExperimentConfiguration, TreeConfiguration
from core.experiment.environment_description import get_env_description
from core.paths.compare_path import ComparePaths
from core.paths.sdt_path import SDTPaths
from visualization import compare_per_tree_config
from visualization.dqn.dqn_loader import DQNLoader
from visualization.sdt.sdt_loader import SDTLoader
from visualization.theme import *
from visualization.visualizer import Visualizer


class SDTVisualizer(Visualizer):
    """The visualizer for a soft decision tree."""
    def __init__(self, theme: ColourTheme = BlueColours3(), lines: LineStyles = LineStyles(), dpi=72):
        # Super call
        super(SDTVisualizer, self).__init__(theme, lines, dpi)
        # The loss labels, logs and data loader
        self.loader = SDTLoader()
        self.losses_log = self.loader.losses_log
        self.actions_log = self.loader.actions_log
        self.sdt_loss_labels = [self.losses_log.total_loss, self.losses_log.inners_loss, self.losses_log.leafs_loss]
        self._all_actions = "All actions"
        self.ext = "pdf"

    @staticmethod
    def _create_tree_config_title(exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                                  value=None, average=False):
        seed_string = "" if average else f"[Seed {exp_config.seed}] "
        val_string = "" if value is None else f": {value}"
        return f"{seed_string}{tree_config.name}{val_string}"

    @staticmethod
    def _create_callback_title(exp_config: ExperimentConfiguration, sdt_callback: SDTCallbackTemplate,
                               value=None, average=False):
        seed_string = "" if average else f"[Seed {exp_config.seed}] "
        val_string = "" if value is None else f": {value}"
        return f"{seed_string}{sdt_callback.create_name()}{val_string}"

    def _create_specific_title(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                               sdt_callback: SDTCallbackTemplate, value=None, average=False):
        seed_string = "" if average else f"[Seed {exp_config.seed}] "
        val_string = "" if value is None else f": {value}"
        name = self._get_short_name(sdt_callback)
        return f"{seed_string}{tree_config.name} | {name}{val_string}"

    @staticmethod
    def _get_short_name(sdt_callback: SDTCallbackTemplate):
        new_name = sdt_callback.create_name().replace("EpisodeCallback_", "").replace("_", "@")
        return new_name

    def _plot_per_step_episode_and_epoch(self, title, y_label, epochs, steps, episodes, ys, labels,
                                         save_dir, save_name, show_labels, legend_location="lower left",
                                         plot_epochs=True, plot_steps=True, plot_episodes=True, xs_given=False,
                                         y_lim=None):
        x_data = [(self.losses_log.epoch, epochs, plot_epochs), (self.losses_log.step, steps, plot_steps),
                  (self.losses_log.episode, episodes, plot_episodes)]
        # Only plot what is requested
        for x_label, x, plot in x_data:
            if plot:
                save_path = f"{save_dir}/{x_label}_{save_name.replace(' ', '_')}.{self.ext}"
                xs = x if xs_given else [x for _ in labels]
                self._plots(title, x_label, y_label, xs, ys, labels, save_path, show_labels,
                            legend_location=legend_location, y_lim=y_lim)

    def _visualize_tree_losses(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                               sdt_callback: SDTCallbackTemplate, epoch_save_int):
        # Load the data
        data = self.loader.load_loss_data(exp_config, tree_config, sdt_callback, epoch_save_int)
        # Plot the losses per step and epoch
        directory = SDTPaths.vis_specific_directory(exp_config, tree_config, sdt_callback.create_name())
        os.makedirs(directory, exist_ok=True)
        y_label = "Loss"
        title = self._create_specific_title(exp_config, tree_config, sdt_callback, y_label)
        steps = data[self.losses_log.step]
        epochs = data[self.losses_log.epoch]
        episodes = data[self.losses_log.episode]
        ys = list(map(lambda l: data[l], self.sdt_loss_labels))
        self._plot_per_step_episode_and_epoch(title, y_label, epochs, steps, episodes, ys, self.sdt_loss_labels,
                                              directory, y_label, show_labels=True)
        return data

    def visualize_losses_per_tree_config(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                                         sdt_callbacks: List[SDTCallbackTemplate], epoch_save_int, plot_name=None):
        # Load all callbacks for the given tree config
        data = []
        labels = []
        # Load the data
        for callback in sdt_callbacks:
            callback_data = self.loader.load_loss_data(exp_config, tree_config, callback, epoch_save_int)
            data.append(callback_data)
            labels.append(callback.create_name())
        # Gather the rest of the plot data
        directory = SDTPaths.vis_per_tree_directory(exp_config, tree_config)
        os.makedirs(directory, exist_ok=True)
        steps = list(map(lambda c: c[self.losses_log.step], data))
        epochs = list(map(lambda c: c[self.losses_log.epoch], data))
        episodes = list(map(lambda c: c[self.losses_log.episode], data))
        # For each loss
        for y_label in self.sdt_loss_labels:
            name = y_label if plot_name is None else f"{plot_name}_{y_label}"
            title = self._create_tree_config_title(exp_config, tree_config, y_label)
            ys = list(map(lambda c: c[y_label], data))
            # Plot per step, episode and epoch
            self._plot_per_step_episode_and_epoch(title, y_label, epochs, steps, episodes, ys, labels,
                                                  directory, name, show_labels=True, xs_given=True)

    def visualize_losses_per_callback(self, exp_config: ExperimentConfiguration, tree_configs: List[TreeConfiguration],
                                      sdt_callback: SDTCallbackTemplate, epoch_save_int):
        # Only visualise if there are more than 1 tree configs
        directory = f"{SDTPaths.vis_directory(exp_config)}/"
        os.makedirs(directory, exist_ok=True)
        if len(tree_configs) > 1:
            directory = f"{directory}/{sdt_callback.create_name()}/"
            os.makedirs(directory, exist_ok=True)
            # Load all callbacks for the given callback
            data = []
            labels = []
            # Load the data
            for tree_config in tree_configs:
                tree_config_data = self.loader.load_loss_data(exp_config, tree_config, sdt_callback, epoch_save_int)
                data.append(tree_config_data)
                labels.append(tree_config.name)
            # Gather the rest of the plot data
            steps = list(map(lambda c: c[self.losses_log.step], data))
            epochs = list(map(lambda c: c[self.losses_log.epoch], data))
            episodes = list(map(lambda c: c[self.losses_log.episode], data))
            # For the total, inner and leaf loss
            for loss_type in self.sdt_loss_labels:
                title = self._create_callback_title(exp_config, sdt_callback, loss_type)
                ys = list(map(lambda c: c[loss_type], data))
                # Plot per step, episode and epoch
                self._plot_per_step_episode_and_epoch(title, loss_type, epochs, steps, episodes, ys, labels,
                                                      directory, loss_type, show_labels=True, xs_given=True)

    def _plot_node(self, node_idx, node_data, is_leaf, partial_title, save_path_prefix,
                   inners_log: SDTInnerWeightsLog, leafs_log: SDTLeafWeightsLog):
        # Gather the data
        prefix = 'Leaf' if is_leaf else 'Inner'
        name = f"{prefix} Node {node_idx}"
        title = f"{partial_title} | {name}"
        y_label = "Probability" if is_leaf else "Weights"
        save_name = f"{prefix}_node_{node_idx}"
        steps = node_data[leafs_log.step if is_leaf else inners_log.step]
        episodes = node_data[leafs_log.episode if is_leaf else inners_log.episode]
        ys = []
        labels = []
        # Leaf node => Action probabilities
        if is_leaf:
            for field in leafs_log.leaf_node:
                ys.append(node_data[field])
                labels.append(field)
        # Inner node => Inner weights
        else:
            for field in inners_log.inner_node:
                ys.append(node_data[field])
                labels.append(field)
            # Beta and bias
            ys.extend([node_data[inners_log.beta], node_data[inners_log.bias]])
            labels.extend([inners_log.beta, inners_log.bias])
        # Plot per epoch, episode and step
        self._plot_per_step_episode_and_epoch(title, y_label, None, steps, episodes, ys, labels, save_path_prefix,
                                              save_name, show_labels=True, plot_epochs=False)

    def _visualize_weights(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                           sdt_callback: SDTCallbackTemplate, epoch_save_int):
        """Visualize weights per tree."""
        # Load the data
        env_description = get_env_description(exp_config.env_name)
        inners, leafs = self.loader.load_weights_data(exp_config, tree_config, sdt_callback, epoch_save_int)
        inners_log = SDTInnerWeightsLog(env_description)
        leafs_log = SDTLeafWeightsLog(env_description)
        title = self._create_specific_title(exp_config, tree_config, sdt_callback)
        directory = SDTPaths.vis_nodes_directory(exp_config, tree_config, sdt_callback.create_name())
        os.makedirs(directory, exist_ok=True)
        # All inner nodes
        for inner in range(2 ** tree_config.depth - 1):
            self._plot_node(inner, inners[inner], False, title, directory, inners_log, leafs_log)
        # All leaf nodes
        for leaf in range(2 ** tree_config.depth):
            self._plot_node(leaf, leafs[leaf], True, title, directory, inners_log, leafs_log)

    def visualize_specific(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                           sdt_callback: SDTCallbackTemplate, epoch_save_int, vis_weights, vis_weights_max_depth):
        # Visualize the weights
        if vis_weights and tree_config.depth <= vis_weights_max_depth:
            self._visualize_weights(exp_config, tree_config, sdt_callback, epoch_save_int)
        # Visualize the losses
        data = self._visualize_tree_losses(exp_config, tree_config, sdt_callback, epoch_save_int)
        # Visualize the epochs per steps
        steps = data[self.losses_log.step]
        epochs = data[self.losses_log.epoch]
        episodes = data[self.losses_log.episode]
        y_label = self.losses_log.epoch
        title = self._create_specific_title(exp_config, tree_config, sdt_callback, y_label)
        directory = SDTPaths.vis_specific_directory(exp_config, tree_config, sdt_callback.create_name())
        self._plot_per_step_episode_and_epoch(title, y_label, None, steps, episodes, [epochs], [None], directory,
                                              y_label, show_labels=False, plot_epochs=False, plot_episodes=False)

    def _calculate_action_statistics(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                                     sdt_callback: SDTCallbackTemplate, epoch_save_int):
        # Load the actions data for the tree
        env_description = get_env_description(exp_config.env_name)
        log = self.actions_log
        post_log = SDTPostActionsLog(env_description)
        data = self.loader.load_actions_data(exp_config, tree_config, sdt_callback, epoch_save_int)
        iterator = zip(
            data[log.epoch], data[log.step], data[log.episode], data[log.action], data[log.agent_action],
            data[log.leaf], data[log.leaf_distribution], data[log.agent_q_values], data[log.same_action],
            data[log.full_order_match], data[log.order_match], data[log.max_match], data[log.error]
        )
        general_counts = {}
        _all_idx = env_description.num_actions()
        for epoch, step, episode, sdt_action, agent_action, leaf, leaf_distribution, agent_q_values, same_action, \
            full_order_match, order_match, max_match, distribution_error in iterator:
            epoch = int(epoch)
            step = int(step)
            episode = int(episode)
            # Initialise epoch data if necessary
            if general_counts.get(epoch) is None:
                general_counts[epoch] = {
                    # action -> (step, episode, count, sdt_count, full_match, matches)
                    action: [step, episode, 0, 0, 0, []] for action in range(_all_idx + 1)
                }
            general_counts[epoch][agent_action][2] += 1
            general_counts[epoch][agent_action][3] += same_action
            general_counts[epoch][agent_action][4] += full_order_match
            general_counts[epoch][agent_action][5].append(order_match)

            general_counts[epoch][_all_idx][2] += 1
            general_counts[epoch][_all_idx][3] += same_action
            general_counts[epoch][_all_idx][4] += full_order_match
            general_counts[epoch][_all_idx][5].append(order_match)

        # Calculate the statistics per epoch per action
        new_data_entries = {action: [] for action in range(env_description.num_actions())}
        new_data_entries[self._all_actions] = []
        for epoch in general_counts.keys():
            all_sdt_count = 0
            all_count = 0
            all_full_match = 0
            all_matches = []

            for action in range(_all_idx):
                step, episode, count, sdt_count, full_match, matches = general_counts[epoch][action]
                matches = np.array(matches)
                if count == 0:  # Avoid divisions by 0 for unused actions
                    count = 1
                action_accuracy = sdt_count / count
                full_match_accuracy = full_match / count
                match_average = matches.mean()
                match_min = matches.min()
                match_max = matches.max()
                entry = post_log.create_entry(step, episode, epoch, action, action_accuracy, full_match_accuracy,
                                              match_min, match_max, match_average)
                new_data_entries[action].append(entry)

                all_count += count
                all_sdt_count += sdt_count
                all_full_match += full_match
                all_matches.extend(matches)

            all_accuracy = all_sdt_count / all_count
            all_match_accuracy = all_full_match / all_count
            all_matches = np.array(all_matches)
            all_average = all_matches.mean()
            all_min = all_matches.min()
            all_max = all_matches.max()
            # noinspection PyUnboundLocalVariable
            entry = post_log.create_entry(step, episode, epoch, -1, all_accuracy, all_match_accuracy,
                                          all_min, all_max, all_average)
            new_data_entries[self._all_actions].append(entry)
        # Write the entries to a file
        for action in range(_all_idx):
            path = SDTPaths.log_post_actions_path(exp_config, tree_config, sdt_callback.create_name(), action)
            post_log.write_data(new_data_entries[action], path)
        path = SDTPaths.log_post_actions_path(exp_config, tree_config, sdt_callback.create_name(), self._all_actions)
        post_log.write_data(new_data_entries[self._all_actions], path)

    def _plot_actions_accuracy_and_order(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                                         sdt_callbacks: List[SDTCallbackTemplate], plot_name=None):
        # Load all callbacks for the given tree config
        env_description = get_env_description(exp_config.env_name)
        post_log = SDTPostActionsLog(env_description)
        data = {action: [] for action in range(env_description.num_actions())}
        data[self._all_actions] = []
        labels = []
        # Load the data
        for callback in sdt_callbacks:
            for action in range(env_description.num_actions()):
                callback_data = \
                    self.loader.load_post_actions_data(exp_config, tree_config, callback, action)
                data[action].append(callback_data)
            data[self._all_actions].append(self.loader.load_post_actions_data(exp_config, tree_config, callback,
                                                                              self._all_actions))
            name = self._get_short_name(callback)
            labels.append(name)
        # Gather the rest of the plot data
        directory = SDTPaths.vis_per_tree_directory(exp_config, tree_config)
        os.makedirs(directory, exist_ok=True)

        steps = list(map(lambda c: c[post_log.step], data[0]))
        epochs = list(map(lambda c: c[post_log.epoch], data[0]))
        episodes = list(map(lambda c: c[post_log.episode], data[0]))

        # Overall: Draw action accuracy, Order accuracy, Full order accuracy
        data_to_plot = [post_log.action_accuracy, post_log.order_match_avg, post_log.full_order_match]

        def _plot(label, a, a_name, x_epochs, x_steps, x_episodes, create_title):
            name = f"{label}_{a_name}" if plot_name is None else f"{plot_name}_{label}_{a_name}"
            title = create_title(exp_config, tree_config, f"{label} |\n{a_name}")
            ys = list(map(lambda c: c[label], data[a]))
            # Plot per step, episode and epoch
            self._plot_per_step_episode_and_epoch(title, label, x_epochs, x_steps, x_episodes, ys, labels, directory,
                                                  name, show_labels=True, xs_given=True, legend_location='lower left')

        # For each plot
        for y_label in data_to_plot:
            # For each action
            for action, action_name in env_description.actions.items():
                _plot(y_label, action, action_name, epochs, steps, episodes, self._create_tree_config_title)
            # For all actions
            action = action_name = self._all_actions
            steps = list(map(lambda c: c[post_log.step], data[action]))
            epochs = list(map(lambda c: c[post_log.epoch], data[action]))
            episodes = list(map(lambda c: c[post_log.episode], data[action]))
            _plot(y_label, action, action_name, epochs, steps, episodes, self._create_tree_config_title)

    def compare_per_tree_config(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                                sdt_callbacks: List[SDTCallbackTemplate], epoch_save_int, training_agent):
        # Sort the comparisons to make the plots more readable
        per_interval, per_episodes, per_full_int, per_episodes_used = \
            compare_per_tree_config(sdt_callbacks, DQNLoader(), exp_config, training_agent)
        # Make plots comparing per interval
        for interval, callbacks in per_interval.items():
            # Only compare if there are at least 2 callbacks
            if len(callbacks) > 1:
                plot_name = f"I_{interval}"
                self.visualize_losses_per_tree_config(exp_config, tree_config, callbacks, epoch_save_int, plot_name)
                self._plot_actions_accuracy_and_order(exp_config, tree_config, callbacks, plot_name)
        # Make plots comparing per episodes used
        for interval, callbacks in per_episodes.items():
            # Only compare if there are at least 2 callbacks
            if len(callbacks) > 1:
                plot_name = f"E_{interval}"
                self.visualize_losses_per_tree_config(exp_config, tree_config, callbacks, epoch_save_int, plot_name)
                self._plot_actions_accuracy_and_order(exp_config, tree_config, callbacks, plot_name)
        # Make plots comparing per episodes used in total
        for episodes, callbacks in per_episodes_used.items():
            if len(callbacks) > 1:
                plot_name = f"U_{episodes}"
                self.visualize_losses_per_tree_config(exp_config, tree_config, callbacks, epoch_save_int, plot_name)
                self._plot_actions_accuracy_and_order(exp_config, tree_config, callbacks, plot_name)
        # Make plots per full intervals used
        if len(per_full_int) > 0:
            plot_name = f"F"
            self.visualize_losses_per_tree_config(exp_config, tree_config, per_full_int, epoch_save_int, plot_name)
            self._plot_actions_accuracy_and_order(exp_config, tree_config, per_full_int, plot_name)

    def plot_all(self, exp_config: ExperimentConfiguration, tree_configs: List[TreeConfiguration],
                 sdt_callbacks: List[SDTCallbackTemplate], epoch_save_int, training_agent,
                 vis_weights=True, vis_weights_max_depth=5):
        # For each specific configuration
        for tree_config in tree_configs:
            for callback in sdt_callbacks:
                self._calculate_action_statistics(exp_config, tree_config, callback, epoch_save_int)

        # For each tree configuration
        for tree_config in tree_configs:
            # For all sdt_callbacks for this tree configuration
            self.compare_per_tree_config(exp_config, tree_config, sdt_callbacks, epoch_save_int, training_agent)
            # Specific configuration
            for callback in sdt_callbacks:
                if tree_config.depth == 1 and callback.episode_interval == 100:
                    self.visualize_specific(exp_config, tree_config, callback, epoch_save_int, vis_weights,
                                            vis_weights_max_depth)
        # For each callback
        for callback in sdt_callbacks:
            self.visualize_losses_per_callback(exp_config, tree_configs, callback, epoch_save_int)

    def plot_seeds_specific(self, exp_config: ExperimentConfiguration, tree_config: TreeConfiguration,
                            sdt_callback: SDTCallbackTemplate, epoch_save_int, seeds):
        # Load the data for each seed
        data = {}
        x_label = self.loader.losses_log.episode
        x_data = []
        for seed in seeds:
            exp_config.save_directory = Path(exp_config.base_save_directory) / f"{seed}"
            sdt_data = self.loader.load_loss_data(exp_config, tree_config, sdt_callback, epoch_save_int)
            data[seed] = sdt_data
            sdt_x_data = sdt_data[x_label]
            if len(x_data) < len(sdt_x_data):
                x_data = sdt_x_data

        # Plot the data
        directory = ComparePaths.vis_avg_specific_sdt_directory(exp_config, tree_config, sdt_callback.create_name())
        os.makedirs(directory, exist_ok=True)
        # Plot the losses
        y_label = "Loss"
        title = self._create_specific_title(exp_config, tree_config, sdt_callback, y_label, average=True)

        ys_total = list(map(lambda d: d[self.losses_log.total_loss], data.values()))
        ys_inner = list(map(lambda d: d[self.losses_log.leafs_loss], data.values()))
        ys_leafs = list(map(lambda d: d[self.losses_log.inners_loss], data.values()))

        total_mean, total_std_dev = self._calculate_statistics(ys_total)
        inner_mean, inner_std_dev = self._calculate_statistics(ys_inner)
        leafs_mean, leafs_std_dev = self._calculate_statistics(ys_leafs)

        save_path = f"{directory}/{x_label}_{y_label}.{self.ext}"
        self._plot_multiple_averages(title, x_label, y_label, [x_data, x_data, x_data],
                                     [total_mean, inner_mean, leafs_mean],
                                     [total_std_dev, inner_std_dev, leafs_std_dev], self.sdt_loss_labels,
                                     save_path)

    def compare_all(self, exp_config: ExperimentConfiguration, tree_configs: List[TreeConfiguration],
                    sdt_callbacks: List[SDTCallbackTemplate], epoch_save_int, seeds, training_agent):
        # For each tree configuration
        for tree_config in tree_configs:
            # For all sdt_callbacks for this tree configuration
            self.compare_per_tree_config(exp_config, tree_config, sdt_callbacks, epoch_save_int, training_agent)
            # Specific configuration
            for callback in sdt_callbacks:
                self.plot_seeds_specific(exp_config, tree_config, callback, epoch_save_int, seeds)
                # self.visualize_specific(exp_config, tree_config, callback, epoch_save_int, vis_weights,
                #                         vis_weights_max_depth)
        # For each callback
        for callback in sdt_callbacks:
            self.visualize_losses_per_callback(exp_config, tree_configs, callback, epoch_save_int)
