import os
from typing import List

import numpy as np
import torch
from PIL import Image, ImageDraw, ImageFont

from core.callbacks.sdt.sdt_callback import SDTCallbackTemplate
from core.experiment.config import ExperimentConfiguration, TreeConfiguration
from core.experiment.environment_description import EnvironmentDescription, get_env_description
from core.paths.sdt_path import SDTPaths
from core.models.soft_decision_tree import SoftDecisionTree


class TreeVisualizer(object):
    """The class for visualizing a soft decision tree policy.

    Attributes:
        environment_description: A description of the en
        pixels_per_unit: An indicator for the resolution of the images.
    """

    def __init__(self, environment_description: EnvironmentDescription, pixels_per_unit=200, **kwargs):
        # Input & Output shape
        self.environment_description = environment_description
        self.input_width = environment_description.observations_width
        self.input_height = environment_description.observations_height
        self.output_shape = len(environment_description.actions)
        # Pixels per unit (resolution)
        self.pixels_per_unit = pixels_per_unit
        self.ext = "pdf"

        # Initialize the constants
        self._init_draw_constants(**kwargs)
        self._init_colours_and_fonts(**kwargs)

    def _init_draw_constants(self, node_width=10, node_height=4, node_spacing=0.1,
                             branch_length=4, branch_width=5, padding=2,
                             text_size=0.4, text_size_small=0.3, **kwargs):
        self.node_height = node_height * self.pixels_per_unit * self.input_height
        self.node_width = node_width * self.pixels_per_unit
        add = max(0, self.input_width - 4)
        self.node_width += self.node_width / 4 * add
        self.node_spacing = node_spacing * self.pixels_per_unit
        self.branch_length = branch_length * self.pixels_per_unit
        self.branch_width = branch_width  # No pixels here!
        self.padding = padding * self.pixels_per_unit
        self.text_size = int(text_size * self.pixels_per_unit)  # (default 80)
        self.text_size_small = int(text_size_small * self.pixels_per_unit)  # (default 54)
        # Derived constants
        self.half_node_width = self.node_width / 2
        self.half_node_height = self.node_height / 2
        self.state_value_size = self.node_width / (self.input_width + 1)
        self.half_state_value_size = self.state_value_size / 2

    def _init_colours_and_fonts(self, font_name="Arial", **kwargs):
        # Colours
        self.background_colour = (255, 255, 255, 255)  # TODO: make background transparent?
        self.text_colour = (0, 0, 0, 255)
        self.node_fill = (245, 244, 243, 255)
        self.branch_colour = (58, 58, 58, 255)
        # Fonts
        font_file = f"{font_name}.ttf"
        self.font = ImageFont.truetype(font_file, size=self.text_size)
        self.font_small = ImageFont.truetype(font_file, size=self.text_size_small)

    def _create_image(self, soft_decision_tree: SoftDecisionTree):
        max_depth = soft_decision_tree.depth
        number_leafs = soft_decision_tree.num_leafs
        number_inners = soft_decision_tree.num_inners
        # Calculate the image width and height
        self.image_width = int(number_leafs * self.node_width +
                               number_inners * self.node_spacing + self.padding * 2)
        self.image_height = int(max_depth * self.node_height +
                                (max_depth - 0.5) * self.branch_length + self.padding * 2)
        self.half_width = self.image_width / 2
        self.half_height = self.image_height / 2
        # print(f"Creating image: (width, height) = {self.image_width, self.image_height}")
        # Create a new image
        # image = Image.new(mode='RGBA', size=(self.image_width, self.image_height), color=self.background_colour)
        image = Image.new(mode='RGB', size=(self.image_width, self.image_height), color=self.background_colour)
        return image

    def _reset_drawing_parameters(self):
        self.depth = 1
        self.inner_pos_y = self.padding
        self.row_pos_x_start = self.half_width
        self.row_pos_x = self.row_pos_x_start
        self.child_number = 1

    def _get_positions(self):
        x1 = self.row_pos_x - self.half_node_width
        y1 = self.inner_pos_y - self.half_node_height
        x2 = x1 + self.node_width
        y2 = y1 + self.node_height
        return (x1, y1), (x2, y2)

    def _get_children_positions(self, parent_pos2):
        x1 = self.row_pos_x
        y1 = parent_pos2[1]
        x2 = self.row_pos_x_start * self.child_number / 2
        y2 = y1 + self.branch_length - self.half_node_height
        return (x1, y1), (x2, y2)

    def _increase_child_count(self):
        self.child_number += 2

    def _increase_depth(self):
        self.depth += 1
        self.inner_pos_y += self.branch_length + self.half_node_height
        self.row_pos_x_start /= 2
        self.row_pos_x = self.row_pos_x_start
        self.child_number = 1

    def _update_sibling_position(self):
        self.row_pos_x += 2 * self.row_pos_x_start

    def _get_value_colour(self, observation_value, extrema):
        min_value, max_value = extrema
        value_interval = max_value - min_value
        colour_value = int(255 * ((observation_value - min_value) / value_interval))
        colour = (colour_value, colour_value, colour_value, 255)
        value_text = 255 if colour_value < 255 / 2 else 0
        colour_text_value = (value_text, value_text, value_text, 255)
        return colour, colour_text_value

    def _draw_observation_value(self, draw, pos1, pos2, value, extrema):
        # Get the colour of the value within the domain
        colour, colour2 = self._get_value_colour(value, extrema=extrema)
        # Draw the observation value
        draw.rectangle(xy=(pos1, pos2), fill=colour, outline=self.branch_colour, width=1)
        # Draw the value as text
        state_name = str(value)  # str(round(value, 5))  # TODO: rounding
        n_w, n_h = self.font_small.getsize(state_name)
        tx = pos1[0] + self.half_state_value_size - n_w / 2
        ty = pos1[1] + self.half_state_value_size - n_h / 2
        draw.text(xy=(tx, ty), text=state_name, font=self.font_small, fill=colour2)

    def _draw_inner(self, draw, values, extrema, node_pos1, node_pos2):
        # Draw the parameter values
        x1 = node_pos1[0] + self.half_node_width - self.input_width * self.half_state_value_size
        y1 = node_pos1[1] + self.half_node_height - self.half_state_value_size
        x2 = x1 + self.state_value_size
        y2 = y1 + self.state_value_size
        for value in values:
            self._draw_observation_value(draw, pos1=(x1, y1), pos2=(x2, y2), value=value, extrema=extrema)
            x1 += self.state_value_size
            x2 = x1 + self.state_value_size
            y2 = y1 + self.state_value_size
        # If inner node: draw branches to children (all inner nodes have children)
        # Left child
        child_pos1, child_pos2 = self._get_children_positions(parent_pos2=node_pos2)
        draw.line(xy=(child_pos1, child_pos2), fill=self.branch_colour, width=self.branch_width)
        self._increase_child_count()
        # Right child
        child_pos1, child_pos2 = self._get_children_positions(parent_pos2=node_pos2)
        draw.line(xy=(child_pos1, child_pos2), fill=self.branch_colour, width=self.branch_width)
        self._increase_child_count()

    def _draw_leaf(self, draw, values, extrema, node_pos1):
        # Draw the action values
        x1 = node_pos1[0] + self.half_node_width - self.output_shape * self.half_state_value_size
        y1 = node_pos1[1] + self.half_node_height - self.half_state_value_size
        x2 = x1 + self.state_value_size
        y2 = y1 + self.state_value_size

        name = "Actions:"
        t_w, t_h = self.font.getsize(name)
        draw.text(xy=(x1 - self.half_state_value_size - t_w, y1 + self.half_state_value_size - t_h / 2),
                  text=name, font=self.font, fill=self.text_colour, align='center')

        for idx, value in enumerate(values):
            self._draw_observation_value(draw, pos1=(x1, y1), pos2=(x2, y2), value=value, extrema=extrema)
            x1 += self.state_value_size
            x2 = x1 + self.state_value_size
            y2 = y1 + self.state_value_size
            name = f"{idx}"
            t_w, _ = self.font.getsize(name)
            draw.text(xy=(x1 - self.half_state_value_size - t_w / 2, y2), text=name,
                      font=self.font, fill=self.text_colour, align='center')

    def _draw_node(self, draw, node_idx, values, extrema, is_leaf=False):
        # Get the node's positions
        pos1, pos2 = self._get_positions()
        name = f"{'Leaf' if is_leaf else 'Inner'} Node {node_idx}"
        # Draw the circle
        draw.ellipse(xy=(pos1, pos2), outline=self.branch_colour, fill=self.node_fill, width=self.branch_width)
        # Name of the node
        name_w, name_h = self.font.getsize(name)
        name_x = pos1[0] + self.half_node_width - name_w / 2
        name_y = pos1[1] + self.half_node_height / 4 - name_h / 2
        draw.text(xy=(name_x, name_y), text=name, font=self.font, fill=self.text_colour, align='center')
        # Draw the values of the node
        if is_leaf:
            self._draw_leaf(draw, values, extrema=extrema, node_pos1=pos1)
        else:
            self._draw_inner(draw, values, extrema=extrema, node_pos1=pos1, node_pos2=pos2)

    def _draw_info(self, draw, experiment_configuration: ExperimentConfiguration,
                   tree_configuration: TreeConfiguration, sdt_callback_name, step):
        # Write the name of the soft decision tree being displayed
        name = f"{tree_configuration.name} {sdt_callback_name}"
        experiment_name = str(experiment_configuration.save_directory)
        experiment_name = experiment_name.rsplit("Experiments/")[-1]
        step_name = f"Training epoch {step}"

        exp_w, exp_h = self.font.getsize(experiment_name)
        exp_x = exp_y = 4
        draw.text(xy=(exp_x, exp_y), text=experiment_name, font=self.font_small, fill=self.text_colour, align='left')
        name_x = exp_x
        name_y = exp_y + exp_h * 1.1
        name_w, name_h = self.font.getsize(name)
        draw.text(xy=(name_x, name_y), text=name, font=self.font_small, fill=self.text_colour, align='left')
        step_x = exp_x
        step_y = name_y + name_h * 1.1
        draw.text(xy=(step_x, step_y), text=step_name, font=self.font_small, fill=self.text_colour, align='left')

    def visualize_tree(self, experiment_configuration: ExperimentConfiguration,
                       tree_configuration: TreeConfiguration,
                       sdt_callback_name, step, temp_save=False):
        """Function to visualize the soft decision tree policy"""
        path = SDTPaths.save_path(experiment_configuration, tree_configuration, sdt_callback_name, step)
        soft_decision_tree = SoftDecisionTree(experiment_configuration, tree_configuration)
        soft_decision_tree.train(False)
        soft_decision_tree.load_state_dict(torch.load(path))
        soft_decision_tree.eval()
        # Get the parameters of the soft decision tree
        inners, betas, biases, leafs = soft_decision_tree.get_parameters()

        # Create a new image and drawing context for it
        image = self._create_image(soft_decision_tree)
        draw = ImageDraw.Draw(image)
        self._draw_info(draw, experiment_configuration, tree_configuration, sdt_callback_name, step)
        self._reset_drawing_parameters()
        # Draw inner nodes
        for idx, (inner_node, beta) in enumerate(zip(inners, betas)):
            # Get the parameter values of the node
            values = inner_node * beta
            extrema = (np.min(values), np.max(values))
            # Draw the node
            self._draw_node(draw, idx, values=values, extrema=extrema, is_leaf=False)
            # The current depth is done
            if (idx + 1) == (2 ** self.depth - 1):
                self._increase_depth()
            # Set the location of the next inner node on this depth
            else:
                self._update_sibling_position()
        # Draw leaf nodes
        for idx, leaf_node in enumerate(leafs):
            values = leaf_node
            extrema = (np.min(values), np.max(values))
            # Draw the node
            self._draw_node(draw, idx, values=values, extrema=extrema, is_leaf=True)
            # Set the position of the next leaf
            self._update_sibling_position()

        if temp_save:
            image.save(f"temp.{self.ext}")

        # Return the image
        return image

    def visualize_tree_process(self, experiment_configuration: ExperimentConfiguration,
                               tree_configuration: TreeConfiguration, sdt_callback_name, vis_frames=False):
        visual_dir = SDTPaths.vis_tree_nodes_directory(experiment_configuration, tree_configuration, sdt_callback_name)
        os.makedirs(visual_dir, exist_ok=True)
        # Get the directory where the SDT saves are
        directory = SDTPaths.directory(experiment_configuration, tree_configuration, sdt_callback_name)
        print(f"Drawing {tree_configuration.name} {sdt_callback_name}...", end="")
        steps = []
        all_files = [file for file in os.listdir(directory) if file.endswith(".pt")]
        num_files = len(all_files)
        # Reduce the number of images produced (RAM memory 30gb+ and disk space reasons)
        create_less = tree_configuration.depth > 3 and num_files > 10
        # Always plot the first and last
        to_create = {10, num_files*10}
        step_size = max(int(num_files / 10), 1)
        if create_less:
            to_create = to_create.union(
                [e for e in range(0, (num_files + 1) * 10, step_size * 10) if 0 < e <= num_files * 10])
            print(f"\n\tDrawing reduced number of trees due to large depth and number of saves: "
                  f"{len(to_create)} out of {num_files}...", end="")

        for file in all_files:
            step = file.replace(".pt", "").replace("epoch-", "")
            if create_less and int(step) not in to_create:
                continue
            # Visualize and save the tree
            image = self.visualize_tree(experiment_configuration, tree_configuration, sdt_callback_name, step)
            steps.append((step, image))
            save_path = f'{visual_dir}/{step}.{self.ext}'
            image.save(save_path)
        # For videos
        if vis_frames:
            steps = sorted(steps, key=lambda s: int(s[0]))
            frames_dir = f"{visual_dir}/frames/"
            os.makedirs(frames_dir, exist_ok=True)
            for frame, (step, image) in enumerate(steps):
                save_path = f'{frames_dir}/{frame}.png'
                image.save(save_path)
        print("Done")

    def visualize_all(self, experiment_configuration: ExperimentConfiguration,
                      tree_configurations: List[TreeConfiguration], sdt_callbacks: List[SDTCallbackTemplate],
                      vis_frames=False):
        env_description = get_env_description(experiment_configuration.env_name)
        experiment_configuration.input_shape = (env_description.num_observations(),)
        experiment_configuration.actions = env_description.num_actions()
        for tree_config in tree_configurations:
            for callback in sdt_callbacks:
                self.visualize_tree_process(experiment_configuration, tree_config, callback.create_name(), vis_frames)
