

###############
# Line Styles #
###############
class LineStyles(object):
    """The class containing all line styles to use."""
    def __init__(self):
        # Store the line styles (for matplotlib)
        #   more styles at https://matplotlib.org/3.1.0/gallery/lines_bars_and_markers/linestyles.html
        self._line_styles = ['solid', 'dashed', 'dotted', 'dashdot', (0, (3, 3, 1, 3, 1, 3)), ]
        self._num_styles = len(self._line_styles)
        self._current = 0

    def __getitem__(self, item):
        """Get a line style for the given item number."""
        return self._line_styles[item % self._num_styles]

    def get_current(self):
        """Get the current line style."""
        return self[self._current]

    def get_next(self):
        """Get the next style after the last used one."""
        style = self[self._current]
        self._current += 1
        return style

    def reset(self):
        """Reset the last used style."""
        self._current = 0


class MarkerStyles(object):
    """The class containing all marker styles to use."""
    def __init__(self):
        # Store the line styles (for matplotlib)
        #   more styles at https://matplotlib.org/3.1.0/gallery/lines_bars_and_markers/linestyles.html
        self._line_styles = ['o', '^', 's', 'P', 'D', 'X', 4]
        self._num_styles = len(self._line_styles)
        self._current = 0

    def __getitem__(self, item):
        """Get a line style for the given item number."""
        return self._line_styles[item % self._num_styles]

    def get_current(self):
        """Get the current line style."""
        return self[self._current]

    def get_next(self):
        """Get the next style after the last used one."""
        style = self[self._current]
        self._current += 1
        return style

    def reset(self):
        """Reset the last used style."""
        self._current = 0


#################
# Colour Themes #
#################
class ColourTheme(object):
    """The class containing a list of colours as a VUB theme."""
    def __init__(self, colours):
        # Normalize from 0-255 to between 0-1
        self._max_value = 255
        # Store the colours
        self._colours = self._normalize(colours)
        self._num_colours = len(self._colours)
        self._current = 0

    def _normalize(self, colours):
        """Normalize RGB values from between 0-255 to 0-1."""
        normalized_colours = []
        for red, green, blue in colours:
            colour = (red / self._max_value, green / self._max_value, blue / self._max_value)
            normalized_colours.append(colour)
        return normalized_colours

    def __getitem__(self, item):
        """Get a colour for the given item number."""
        return self._colours[item % self._num_colours]

    def get_current(self):
        """Get the current colour."""
        return self[self._current]

    def get_next(self):
        """Get the next colour after the last used one."""
        colour = self[self._current]
        self._current += 1
        return colour

    def reset(self):
        """Reset the last used colour."""
        self._current = 0


###############
# Blue Themes #
###############
class BlueColours(ColourTheme):
    """The class containing all blue colours available as a VUB theme."""
    def __init__(self):
        # The colours for this theme
        colours = [
            (1, 32, 141),  # Default
            (36, 82, 255),  # 40% lighter
            (2, 23, 100),  # 25% darker
            (97, 137, 255),  # 60% lighter
            (1, 16, 62),  # 50% darker
            (170, 194, 255),  # 80% lighter
            # Added Theme2
            (71, 103, 180),  # Default
            (137, 160, 208),  # 40% lighter
            (43, 71, 135),  # 25% darker
            (174, 191, 224),  # 60% lighter
            (29, 45, 83),  # 50% darker
            (214, 222, 239),  # 80% lighter
        ]
        # Super call
        super(BlueColours, self).__init__(colours)


class BlueColours2(ColourTheme):
    """The class containing all blue colours available as a VUB theme."""
    def __init__(self):
        # The colours for this theme
        colours = [
            (71, 103, 180),  # Default
            (137, 160, 208),  # 40% lighter
            (43, 71, 135),  # 25% darker
            (174, 191, 224),  # 60% lighter
            (29, 45, 83),  # 50% darker
            (214, 222, 239),  # 80% lighter
        ]
        # Super call
        super(BlueColours2, self).__init__(colours)


class BlueColours3(ColourTheme):
    """The class containing all blue colours available as a VUB theme."""
    def __init__(self):
        # The colours for this theme
        colours = [
            (1, 32, 141),  # Default
            (198, 196, 194),  # 80% lighter
            # (225, 224, 223),  # Default
            (253, 79, 9),  # Default
            (36, 82, 255),  # 40% lighter
            (101, 96, 91),  # 40% lighter
            (254, 144, 83),  # 40% lighter
            (2, 23, 100),  # 25% darker
            (45, 43, 41),  # 25% darker
            (176, 55, 6),  # 25% darker
            (97, 137, 255),  # 60% lighter
            (160, 156, 153),  # 60% lighter
            (253, 178, 136),  # 60% lighter
            (1, 16, 62),  # 50% darker
            (18, 18, 17),  # 50% darker
            (107, 36, 2),  # 50% darker
            (170, 194, 255),  # 80% lighter
            (198, 196, 194),  # 80% lighter
            (255, 215, 193),  # 80% lighter
        ]
        # Super call
        super(BlueColours3, self).__init__(colours)


#################
# Orange Themes #
#################
class OrangeColours(ColourTheme):
    """The class containing all orange colours available as a VUB theme."""
    def __init__(self):
        # The colours for this theme
        colours = [
            (253, 79, 9),  # Default
            (254, 144, 83),  # 40% lighter
            (176, 55, 6),  # 25% darker
            (253, 178, 136),  # 60% lighter
            (107, 36, 2),  # 50% darker
            (255, 215, 193),  # 80% lighter
            # Added Theme2
            (253, 152, 121),  # Default
            (254, 191, 171),  # 40% lighter
            (252, 73, 32),  # 25% darker
            (255, 213, 198),  # 60% lighter
            (183, 32, 5),  # 50% darker
            (255, 233, 225),  # 80% lighter
        ]
        # Super call
        super(OrangeColours, self).__init__(colours)


class OrangeColours2(ColourTheme):
    """The class containing all orange colours available as a VUB theme."""
    def __init__(self):
        # The colours for this theme
        colours = [
            (253, 152, 121),  # Default
            (254, 191, 171),  # 40% lighter
            (252, 73, 32),  # 25% darker
            (255, 213, 198),  # 60% lighter
            (183, 32, 5),  # 50% darker
            (255, 233, 225),  # 80% lighter
        ]
        # Super call
        super(OrangeColours2, self).__init__(colours)


class OrangeColours3(ColourTheme):
    """The class containing all orange colours available as a VUB theme."""
    def __init__(self):
        # The colours for this theme
        colours = [
            (253, 79, 9),  # Default
            (198, 196, 194),  # 80% lighter
            (254, 144, 83),  # 40% lighter
            (101, 96, 91),  # 40% lighter
            (176, 55, 6),  # 25% darker
            (45, 43, 41),  # 25% darker
            (253, 178, 136),  # 60% lighter
            (160, 156, 153),  # 60% lighter
            (107, 36, 2),  # 50% darker
            (18, 18, 17),  # 50% darker
            (255, 215, 193),  # 80% lighter
            (198, 196, 194),  # 80% lighter
            # Added Theme2
            (253, 152, 121),  # Default
            (254, 191, 171),  # 40% lighter
            (252, 73, 32),  # 25% darker
            (255, 213, 198),  # 60% lighter
            (183, 32, 5),  # 50% darker
            (255, 233, 225),  # 80% lighter
        ]
        # Super call
        super(OrangeColours3, self).__init__(colours)


###############
# Grey Themes #
###############
class GreyColours(ColourTheme):
    """The class containing all grey colours available as a VUB theme."""
    def __init__(self):
        # The colours for this theme
        colours = [
            (225, 224, 223),  # Default
            (101, 96, 91),  # 40% lighter
            (45, 43, 41),  # 25% darker
            (160, 156, 153),  # 60% lighter
            (18, 18, 17),  # 50% darker
            (198, 196, 194),  # 80% lighter
        ]
        # Super call
        super(GreyColours, self).__init__(colours)


class GreyColours2(ColourTheme):
    """The class containing all grey colours available as a VUB theme."""
    def __init__(self):
        # The colours for this theme
        colours = [
            (58, 58, 58),  # Default
            (129, 129, 129),  # 40% lighter
            (43, 42, 42),  # 25% darker
            (170, 168, 168),  # 60% lighter
            (29, 28, 28),  # 50% darker
            (212, 210, 210),  # 80% lighter
        ]
        # Super call
        super(GreyColours2, self).__init__(colours)
