import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from visualization.theme import *


class Visualizer(object):
    """The visualizer for experiment data."""
    def __init__(self, theme: ColourTheme, lines: LineStyles, dpi):
        self.colours = theme
        self.lines = lines
        self.markers = MarkerStyles()
        self.dpi = dpi
        self.lw = 2.5

    @staticmethod
    def _init_plot(title, x_label, y_label):
        s0 = 12
        s2 = 14
        s3 = 18
        # if True:
        #     s0 = 16
        #     s2 = 20
        #     s3 = 30
        # else:
        #     s0 = 14
        #     s2 = 16
        #     s3 = 24
        plt.title(title, fontsize=s3)
        plt.ylabel(y_label, fontsize=s2)
        plt.xlabel(x_label, fontsize=s2)
        plt.xticks(fontsize=s0)
        plt.yticks(fontsize=s0)

    def _finish_plot(self, save_path, show_labels, legend_location=None):
        if show_labels:
            plt.legend(loc=legend_location, fontsize=12)
        plt.savefig(save_path, dpi=self.dpi, bbox_inches='tight')
        plt.close()
        self.colours.reset()
        self.lines.reset()
        self.markers.reset()

    def _get_colour_and_line(self, change_colour=True, colour_idx=None, line_idx=None):
        # Colour
        if colour_idx is not None:
            color = self.colours[colour_idx]
        elif change_colour:
            color = self.colours.get_next()
        else:
            color = self.colours.get_current()
        # Lines
        if line_idx is not None:
            line = self.lines[line_idx]
            marker = self.markers[line_idx]
        elif change_colour:
            line = self.lines.get_next()
            marker = self.markers.get_next()
        else:
            line = self.lines.get_current()
            marker = self.lines.get_current()
        # Return the colour and line
        return color, line, marker

    def _plot_data(self, x, y, label=None, change_colour=True, y_lim=None, colour_idx=None, line_idx=None):
        if y_lim is not None:
            plt.ylim(y_lim)
        color, line, marker = self._get_colour_and_line(change_colour, colour_idx, line_idx)
        if len(x) == len(y) == 1:
            plt.scatter(x, y, label=label, color=color, ls=line, marker=marker, lw=self.lw)
        else:
            plt.plot(x, y, label=label, color=color, ls=line, lw=self.lw)

    def _plot_between(self, x, y_lower, y_upper, change_colour=True, colour_idx=None, line_idx=None):
        color, line, _ = self._get_colour_and_line(change_colour, colour_idx, line_idx)
        plt.fill_between(x, y_lower, y_upper, color=color, ls=line, alpha=0.3)

    def _plots(self, title, x_label, y_label, xs, ys, labels, save_path, show_labels, legend_location=None, y_lim=None):
        self._init_plot(title, x_label, y_label)
        for x, y, label in zip(xs, ys, labels):
            self._plot_data(x, y, label if show_labels else None, y_lim=y_lim)
        self._finish_plot(save_path, show_labels, legend_location)

    def _plot(self, title, x_label, y_label, x, y, save_path, legend_location=None, colour_idx=None, line_idx=None):
        self._init_plot(title, x_label, y_label)
        self._plot_data(x, y, colour_idx=colour_idx, line_idx=line_idx)
        self._finish_plot(save_path, show_labels=False, legend_location=legend_location)

    def _plot_averages(self, title, x_label, y_label, x, y_mean, y_std, save_path, colour_idx=None, line_idx=None,
                       y_lim=None):
        self._init_plot(title, x_label, y_label)
        # Plot the mean
        self._plot_data(x, y_mean, change_colour=False, colour_idx=colour_idx, line_idx=line_idx, y_lim=y_lim)
        # Plot the standard deviations
        self._plot_between(x, y_mean-y_std, y_mean+y_std, colour_idx=colour_idx, line_idx=line_idx)
        self._finish_plot(save_path, show_labels=False)

    def _plot_multiple_averages(self, title, x_label, y_label, xs, y_means, y_stds, labels, save_path, y_lim=None,
                                plot_between=True, legend_location=None, colour_idx=None, line_idx=None):
        self._init_plot(title, x_label, y_label)
        cols = colour_idx if colour_idx is not None else [None for _ in labels]
        lines = line_idx if line_idx is not None else [None for _ in labels]
        for x, y_mean, y_std, label, col, line in zip(xs, y_means, y_stds, labels, cols, lines):
            # Plot the means
            self._plot_data(x, y_mean, label, change_colour=False, y_lim=y_lim, colour_idx=col, line_idx=line)
            # Plot the standard deviations
            if plot_between:
                self._plot_between(x, y_mean-y_std, y_mean+y_std, colour_idx=col, line_idx=line)
            else:
                self.colours.get_next()
                self.lines.get_next()
        self._finish_plot(save_path, show_labels=True, legend_location=legend_location)

    @staticmethod
    def _plot_annotation_values(ax, rects, values, text_size, to_string=None):
        if to_string is None:
            to_string = str
        for rect, value in zip(rects, values):
            ax.annotate(to_string(value), xy=(rect.get_x() + rect.get_width() / 2, rect.get_height()), size=text_size,
                        xytext=(0, -10), textcoords="offset points", ha='center', va='center', color="white")

    def _bar_plot(self, title, x_label, y_label, x, y, labels, save_path, show_labels,
                  show_values=False, text_size=None, legend_location=None, to_string=None,
                  colour_idx=None, line_idx=None):
        self._init_plot(title, x_label, y_label)
        color, line, _ = self._get_colour_and_line(True, colour_idx, line_idx)
        rects = plt.bar(x, y, color=color, ls=line)
        plt.xticks(ticks=x, labels=labels)
        if show_values:
            self._plot_annotation_values(plt, rects, y, text_size, to_string)
        self._finish_plot(save_path, show_labels=show_labels, legend_location=legend_location)

    def _grouped_bar_plot(self, title, x_label, y_label, ys, labels, ys_labels, save_path,
                          show_values=False, show_labels=True, legend_location=None, to_string=None):
        # Initialise the plot
        w, h = plt.rcParams.get('figure.figsize')
        fig, ax = plt.subplots(figsize=(max(w, 3 * len(labels)), h))

        if len(labels) > 3:
            s0 = 16
            s2 = 20
            s3 = 30
        else:
            s0 = 14
            s2 = 16
            s3 = 24

        # Calculate the bar widths and positions
        num_ys = len(ys)
        bar_width = 0.80 / num_ys
        text_size = None if num_ys < 5 else 12 * bar_width * 3.5
        label_locations = np.arange(len(labels))
        ax.set_title(title, size=s3)
        ax.set_xlabel(x_label, size=s2)
        ax.set_ylabel(y_label, size=s2)
        ax.set_xticks(label_locations)
        ax.set_xticklabels(labels, size=s2)
        ax.tick_params(axis='both', which='major', labelsize=s0)
        ax.tick_params(axis='both', which='minor', labelsize=s0)

        def get_x(bar_idx):
            return label_locations + (bar_idx - (num_ys - 1) / 2) * bar_width

        # Plot the data
        for idx, (y, yl) in enumerate(zip(ys, ys_labels)):
            rects = ax.bar(get_x(idx), y, width=bar_width, label=yl,
                           color=self.colours.get_next(), ls=self.lines.get_next())
            if show_values:
                self._plot_annotation_values(ax, rects, y, text_size, to_string)
        if show_labels:
            ax.legend(loc=legend_location)
        fig.tight_layout()
        self._finish_plot(save_path, show_labels=False)

    def _averaged_grouped_bar_plot(self, title, x_label, y_label, ys_mean, ys_std_dev, labels, ys_labels, save_path,
                                   show_values=False, show_labels=True, legend_location=None, to_string=None):
        # Initialise the plot
        w, h = plt.rcParams.get('figure.figsize')
        fig, ax = plt.subplots(figsize=(max(w, 3 * len(labels)), h))

        if len(labels) > 3:
            s0 = 16
            s2 = 20
            s3 = 30
        else:
            s0 = 14
            s2 = 16
            s3 = 24
        # params = {'axes.labelsize': s1, 'axes.titlesize': s2, 'legend.fontsize': s2,
        #           'xtick.labelsize': s1, 'ytick.labelsize': s1, 'title.fontsize': s3}
        # matplotlib.rcParams.update(params)
        # plt.title(title)
        # plt.ylabel(y_label)
        # plt.xlabel(x_label)

        # Calculate the bar widths and positions
        num_ys = len(ys_mean)
        bar_width = 0.80 / num_ys
        text_size = None if num_ys < 5 else 12 * bar_width * 3.5
        label_locations = np.arange(len(labels))
        ax.set_title(title, size=s3)
        ax.set_xlabel(x_label, size=s2)
        ax.set_ylabel(y_label, size=s2)
        ax.set_xticks(label_locations)
        ax.set_xticklabels(labels, size=s2)
        ax.tick_params(axis='both', which='major', labelsize=s0)
        ax.tick_params(axis='both', which='minor', labelsize=s0)

        def get_x(bar_idx):
            return label_locations + (bar_idx - (num_ys - 1) / 2) * bar_width

        # Plot the data
        for idx, (y, y_std, yl) in enumerate(zip(ys_mean, ys_std_dev, ys_labels)):
            rects = ax.bar(get_x(idx), y, yerr=y_std, width=bar_width, label=yl, ecolor='grey', capsize=5,
                           color=self.colours.get_next(), ls=self.lines.get_next())
            if show_values:
                self._plot_annotation_values(ax, rects, y, text_size, to_string)
        if show_labels:
            ax.legend(loc=legend_location, fontsize=s0)
        fig.tight_layout()
        self._finish_plot(save_path, show_labels=False)

    #

    @staticmethod
    def _use_window(window, points, is_step_window=True):
        under_max = window < (points / 10)  # Average over at most 10% of the data
        m = 1000 if is_step_window else 100
        over_min = window >= (points / m)  # Average over at least 1/m of the data
        return under_max and over_min

    @staticmethod
    def _calculate_moving_average(window, *points):
        new_points = []
        for p in points:
            cumulative_sum = np.cumsum(np.insert(p, 0, 0))
            new_p = (cumulative_sum[window:] - cumulative_sum[:-window]) / float(window)
            new_points.append(new_p)
        return new_points

    @staticmethod
    def _calculate_statistics(ys_data):
        # Turn the data into numpy arrays
        max_len = 0
        for y in ys_data:
            max_len = max(max_len, len(y))
        full_y_datas = np.full(shape=(len(ys_data), max_len), fill_value=np.nan)
        for i, y in enumerate(ys_data):
            full_y_datas[i, :len(y)] = y
        # Calculate mean, sdt dev
        y_mean = full_y_datas.mean(axis=0)
        y_std_dev = full_y_datas.std(axis=0)
        # Return the new data
        return y_mean, y_std_dev
